/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "TPA_API.h"
#include "common.h"

void aik_usage(char * programName) {
	printf("Usage: ");

	printf("%s aik [SUBCOMMAND]\n", programName);

	printf("   create: creates an AIK");
	printf("\n\trequired parameters:\n");
	printf("\t   -l label   \tspecify the label of the AIK to create\n");
	printf("\t   -a aikpwd  \tspecify the password of the AIK to create\n");
	printf("\t   -o ownerpwd\tspecify the password of the owner of the TPM\n");
	printf("\t   -s srkpwd  \tspecify the SRK password\n");

	printf("\toptional parameters:\n");
	printf("\t   -p PCAcert \tspecify the PCA certificate file\n");
	//printf("\t   -d address \tspecify the PCA address [default localhost]\n");
	//printf("\t   -t port    \tspecify the PCA port [default 10000]\n");
	printf("\t   -u uri     \tspecify URI for TPA localstorage\n");
	printf("\t   -e ekfile  \tspecify the EK certificate file\n");

	printf("\n   list: lists the AIK saved in the local storage\n");
	printf("\toptional parameter:\n");
	printf("\t   -u uri     \tspecify URI for TPA localstorage\n");

	printf("\n   del [key uuid]: deletes an AIK from the local storage\n");
	//	printf("\toptional parameters:\n");
	//	printf("\t   -i uuid   \tspecity the UUID of the AIK\n");
}

/*
 * This function creates an AIK and registers it in the local storage
 * The user must provide few parameter to the TPA:
 * - the password of the owner
 * - the password of the TPM
 * - the password of the AIK
 * - the label of the AIK
 */
int aik_create(int argc, char **argv) {

	TPA_AIK *aik = NULL;
	TPA_TPM *tpm = NULL;
	TPA_PCA *ca = NULL;
	TPA_CONTEXT *ctx = NULL;
	TPA_RESULT result = TPA_ERROR;
	int c = 0;
	int counter = 0;
	char *cvalue = NULL;

	char *lstorage = NULL;
	char *pca_addr = NULL;
	int pca_port = 0;

	/*
	 * This function will allocate all needed objects
	 */
	if (TpaHL_CTX_allocate(&ctx) != TPA_SUCCESS)
		goto error;

	if (TpaHL_TPM_allocate(&tpm) != TPA_SUCCESS)
		goto error;

	if (TpaHL_PCA_allocate(&ca) != TPA_SUCCESS)
		goto error;

	if (TpaHL_AIK_allocate(&aik) != TPA_SUCCESS)
		goto error;

	/*
	 * Parsing parameters
	 */

	opterr = 0;

	while ((c = getopt(argc, argv, "l:a:o:s:p:e:u:d:t:")) != -1) {
		switch (c) {
		case 'l':
			/*
			 * -l passes the label of the AIK
			 * We use TpaHL_AIK_set to set the parameter in the object
			 */
			if (TpaHL_AIK_set(aik, AIK_LABEL, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 'a':
			/*
			 * -a passes the password of the AIK
			 * We use TpaHL_AIK_set to set the parameter in the object
			 */
			if (TpaHL_AIK_set(aik, AIK_AIKSECRET, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 'o':
			/*
			 * -o passes the password of the TPM
			 * We don't have a TpaHL_TPM_set yet, so we use a common strncpy
			 * A proper function for setting parameters is needed here!
			 */
			if (TpaHL_TPM_set(tpm, TPM_OWNERPWD, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 's':
			/*
			 * -s passes the password of the SRK
			 * We don't have a TpaHL_TPM_set yet, so we use a common strncpy
			 * A proper function for setting parameters is needed here!
			 */
			if (TpaHL_TPM_set(tpm, TPM_SRKPWD, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 'p':
			if (TpaHL_PCA_set(ca, PCA_PCACERT_FILENAME, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			//counter++;
			break;
		case 'e':
			if (TpaHL_TPM_set(tpm, TPM_EKCERT_FILENAME, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			break;
		case 'u':
			if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			lstorage = optarg;
			break;
		case 'd':
			if (TpaHL_PCA_set(ca, PCA_URI, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			pca_addr = optarg;
			break;
		case 't':
			pca_port = atoi(optarg);
			break;
		case '?':
			/*
			 * Unknown parameter here
			 */
			aik_usage(argv[0]);
		default:
			/*
			 * Something went wrong here
			 */
			printf("Internal error!\n");
			goto error;
		}
	}

	if (counter != 4) {
		aik_usage(argv[0]);
		goto error;
	}

	if (lstorage == NULL) {
		if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(TPA_DEFAULT_LOCAL_STORAGE),
				TPA_DEFAULT_LOCAL_STORAGE) != TPA_SUCCESS)
			goto error;
	}

	/* Set Privacy CA uri */
	if (pca_addr == NULL) {
		if (TpaHL_PCA_set(ca, PCA_URI, strlen(TPA_PCA_DEFAULT_PCA_URI),
				TPA_PCA_DEFAULT_PCA_URI) != TPA_SUCCESS)
			goto error;
	}

	/* Set Privacy CA port */
	if (pca_port == 0) {
		pca_port = atoi(TPA_PCA_DEFAULT_PCA_PORT);
	}

	if (TpaHL_PCA_set(ca, PCA_PORT, sizeof(pca_port), &pca_port) != TPA_SUCCESS)
		goto error;

	/*
	 * We call the TpaHL_AIK_create to create a new AIK
	 */
	if (TpaHL_AIK_Create(ctx, tpm, aik, ca) == TPA_SUCCESS) {
		/*
		 * OK, we have a new AIK here
		 */
		printf("AIK created. ID: %d\n", aik->aik_id);
	} else {
		/*
		 * Something went wrong
		 */
		printf("AIK not created!\n");
		goto error;
	}

	result = TPA_SUCCESS;

	error: /* Error label */
	TpaHL_AIK_freeMemory(aik);
	TpaHL_TPM_freeMemory(tpm);
	TpaHL_PCA_freeMemory(ca);
	TpaHL_CTX_freeMemory(ctx);

	if (cvalue)
		free(cvalue);

	return result;
}

int aik_list(int argc, char **argv) {
	TPA_CONTEXT *ctx = NULL;
	TPA_AIK *aikList = NULL;
	int aikList_size = 0;
	int i = 0;
	int c = 0;
	char *lstorage = NULL;

	/* Allocate Context */
	if (TpaHL_CTX_allocate(&ctx) != TPA_SUCCESS)
		goto error;

	/*
	 * Parsing parameters
	 */
	opterr = 0;

	while ((c = getopt(argc, argv, "u:")) != -1) {
		switch (c) {
		case 'u':
			if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			lstorage = optarg;
			break;
		case '?':
			/*
			 * Unknown parameter here
			 */
			aik_usage(argv[0]);
		default:
			/*
			 * Something went wrong here
			 */
			printf("Internal error!\n");
			goto error;
		}
	}

	if (lstorage == NULL) {
		/* Set Local Storage folder */
		if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(TPA_DEFAULT_LOCAL_STORAGE),
				TPA_DEFAULT_LOCAL_STORAGE) != TPA_SUCCESS)
			goto error;
	}

	if (TpaHL_AIK_List(ctx, &aikList, &aikList_size, 0) != TPA_SUCCESS)
		goto error;

	if (aikList_size < 2) { /* be careful with key 0 */
		printf("No AIK found in db\n");
	} else {
		printf("AIK in db:\n");
		for (i = 0; i < aikList_size; i++) {
			if (aikList[i].aik_id == 0)
				continue;
			printf("%d. uuid: %d\n", i, aikList[i].aik_id);
		}
	}

	TpaHL_CTX_freeMemory(ctx);
	if (aikList)
		free(aikList);

	return TPA_SUCCESS;

	error: return TPA_ERROR;
}

int aik_delete(int argc, char **argv) {
	TPA_CONTEXT *ctx = NULL;
	TPA_AIK *aikList = NULL;
	int aikList_size = 0;
	int found = 0;
	int i = 0;
	int key_id = atoi(argv[1]);

	if (argc != 2) {
		aik_usage(argv[0]);
		goto error;
	}

	/* Allocate Context */
	if (TpaHL_CTX_allocate(&ctx) != TPA_SUCCESS)
		goto error;

	/* Set Local Storage folder */
	if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(TPA_DEFAULT_LOCAL_STORAGE),
			TPA_DEFAULT_LOCAL_STORAGE) != TPA_SUCCESS)
		goto error;

	if (TpaHL_AIK_List(ctx, &aikList, &aikList_size, 0) != TPA_SUCCESS)
		goto error;

	for (i = 0; i < aikList_size; i++) {
		if (aikList[i].aik_id == 0)
			continue;

		if (aikList[i].aik_id == key_id) {
			found = 1;
			break;
		}
	}

	if (found) {
		if (TpaHL_AIK_DeleteById(ctx, atoi(argv[1])) != TPA_SUCCESS)
			goto error;
	} else {
		fprintf(stderr, "Delete failed. AIK Key not found in DB.\n");
	}

	TpaHL_CTX_freeMemory(ctx);
	if (aikList)
		free(aikList);

	return TPA_SUCCESS;

	error: return TPA_ERROR;
}

int aik(int argc, char ** argv) {
	TPA_RESULT res = TPA_ERROR;

	if (argc == 1) {
		aik_usage(argv[0]);
		return TPA_ERROR;
	}
	if (strcmp("create", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = aik_create(argc, argv);
	} else if (strcmp("list", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = aik_list(argc, argv);
	} else if (strcmp("del", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = aik_delete(argc, argv);
	} else {
		aik_usage(argv[0]);
		return TPA_ERROR;
	}

	return res;
}

