/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "TPA_API.h"
#include "TPA_Utils.h"
#include "common.h"

void data_usage(char * programName) {
	printf("Usage: ");

	printf("%s data [SUBCOMMAND]\n", programName);

	printf("   bind: bind a blob with the TPM of this machine");
	printf("\n\trequired parameters:\n");
	printf(
			"\t   -k keyid  \tspecify the key ID of a binding key saved in the storage \n");
	printf("\t   -p keysecret  \tspecify the key secret\n");
	printf("\t   -s srkpwd  \tspecify the SRK password\n");
	printf("\t   -i inFile  \tspecify the input file\n");
	printf("\t   -o outFile\tspecify the output file\n");
	printf("\toptional parameters:\n");
	printf("\t   -u uri     \tspecify URI for TPA localstorage\n");

	printf("\n   unbind: unbind a blob bound with the TPM of this machine");
	printf("\n\trequired parameters:\n");
	printf(
			"\t   -k keyid  \tspecify the key ID of a binding key saved in the storage \n");
	printf("\t   -p keysecret  \tspecify the key secret\n");
	printf("\t   -s srkpwd  \tspecify the SRK password\n");
	printf("\t   -i inFile  \tspecify the input file\n");
	printf("\t   -o outFile\tspecify the output file\n");

	printf("\toptional parameters:\n");
	printf("\t   -u uri     \tspecify URI for TPA localstorage\n");

}

int data_bind(int argc, char **argv) {

	TPA_RESULT result = TPA_ERROR;
	TPA_CONTEXT *ctx = NULL;
	TPA_TPM *tpm = NULL;
	TPA_TPMKEY *key = NULL;
	TPA_OBJ_ID keyID = -1;

	/* Filenames */
	char *input_file = NULL;
	char *output_file = NULL;

	char *keysecret = NULL;
	int keysecret_size = 0;

	unsigned char *input_blob = NULL;
	unsigned int input_blob_len = 0;
	unsigned char *output_blob = NULL;
	unsigned int output_blob_len = 0;

	char *lstorage = NULL;

	int c = 0;
	int counter = 0;

	/*
	 * This function will allocate all needed objects
	 */
	if (TpaHL_CTX_allocate(&ctx) != TPA_SUCCESS)
		goto error;

	if (TpaHL_TPM_allocate(&tpm) != TPA_SUCCESS)
		goto error;

	if (TpaHL_KEY_allocate(&key) != TPA_SUCCESS)
		goto error;

	/*
	 * Parsing parameters
	 */

	opterr = 0;

	while ((c = getopt(argc, argv, "k:o:s:i:u:p:")) != -1) {
		switch (c) {
		case 'k':
			keyID = atoi(optarg);
			counter++;
			break;
		case 'o':
			output_file = optarg;
			counter++;
			break;
		case 's':
			if (TpaHL_TPM_set(tpm, TPM_SRKPWD, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 'i':
			input_file = optarg;
			counter++;
			break;
		case 'u':
			if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			lstorage = optarg;
			break;
		case 'p':
			keysecret = optarg;
			keysecret_size = strlen(keysecret);
			counter++;
			break;
		case '?':
			data_usage(argv[0]);
		default:
			/*
			 * Something went wrong here
			 */
			printf("Internal error!\n");
			goto error;
		}
	}

	if (counter != 5) {
		data_usage(argv[0]);
		goto error;
	}

	if (lstorage == NULL) {
		if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(TPA_DEFAULT_LOCAL_STORAGE),
				TPA_DEFAULT_LOCAL_STORAGE) != TPA_SUCCESS)
			goto error;
	}

	/* Retrieve Binding Key from DB */
	if (TpaHL_KEY_RetrieveTpmBindingKeyById(ctx, &key, keyID) != TPA_SUCCESS) {
		printf("Selected key not found in DB. Quit...\n");
		goto error;
	}

	/* Set Key Secret */
	if (TpaHL_KEY_set(key, KEY_TPMKEYSECRET, keysecret_size, keysecret)
			!= TPA_SUCCESS) {
		printf(
				"It is not possible to set the secret in the key object. Quit...\n");
		goto error;
	}

	/* Read Input File */
	if (Tpa_BLOB_readFromFile(input_file, &input_blob, &input_blob_len)
			!= TPA_SUCCESS) {
		printf("It's not possible to read the input file! Quit...\n");
		goto error;
	}

	/* Perform Binding */
	if (TpaHL_DATA_BindBlob(ctx, tpm, key, input_blob, input_blob_len,
			&output_blob, &output_blob_len) != TPA_SUCCESS) {
		printf("It's not possible to bind the blob! Quit...\n");
		goto error;
	}

	/* Write Ouput into file */
	if (Tpa_BLOB_writeToFile(output_file, output_blob, output_blob_len)
			!= TPA_SUCCESS) {
		printf("It's not possible to read the input file! Quit...\n");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* free */
	TpaHL_TPM_freeMemory(tpm);
	TpaHL_CTX_freeMemory(ctx);
	TpaHL_KEY_freeMemory(key);
	//	if (bindsecret)
	//		free(bindsecret);
	//	if (keysecret)
	//		free(keysecret);

	return result;
}

int data_unbind(int argc, char **argv) {

	TPA_RESULT result = TPA_ERROR;
	TPA_CONTEXT *ctx = NULL;
	TPA_TPM *tpm = NULL;
	TPA_TPMKEY *key = NULL;
	TPA_OBJ_ID keyID = -1;

	/* Filenames */
	char *input_file = NULL;
	char *output_file = NULL;

	char *keysecret = NULL;
	int keysecret_size = 0;

	unsigned char *input_blob = NULL;
	unsigned int input_blob_len = 0;
	unsigned char *output_blob = NULL;
	unsigned int output_blob_len = 0;

	char *lstorage = NULL;

	int c = 0;
	int counter = 0;

	/*
	 * This function will allocate all needed objects
	 */
	if (TpaHL_CTX_allocate(&ctx) != TPA_SUCCESS)
		goto error;

	if (TpaHL_TPM_allocate(&tpm) != TPA_SUCCESS)
		goto error;

	if (TpaHL_KEY_allocate(&key) != TPA_SUCCESS)
		goto error;

	/*
	 * Parsing parameters
	 */

	opterr = 0;

	while ((c = getopt(argc, argv, "k:o:s:i:u:p:")) != -1) {
		switch (c) {
		case 'k':
			keyID = atoi(optarg);
			counter++;
			break;
		case 'o':
			output_file = optarg;
			counter++;
			break;
		case 's':
			if (TpaHL_TPM_set(tpm, TPM_SRKPWD, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 'i':
			input_file = optarg;
			counter++;
			break;
		case 'u':
			if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			lstorage = optarg;
			break;
		case 'p':
			keysecret = optarg;
			keysecret_size = strlen(keysecret);
			counter++;
			break;
		case '?':
			/*
			 * Unknown parameter here
			 */
			data_usage(argv[0]);
		default:
			/*
			 * Something went wrong here
			 */
			printf("Internal error!\n");
			goto error;
		}
	}

	if (counter != 5) {
		data_usage(argv[0]);
		goto error;
	}

	if (lstorage == NULL) {
		if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(TPA_DEFAULT_LOCAL_STORAGE),
				TPA_DEFAULT_LOCAL_STORAGE) != TPA_SUCCESS)
			goto error;
	}

	/* Retrieve Binding Key from DB */
	if (TpaHL_KEY_RetrieveTpmBindingKeyById(ctx, &key, keyID) != TPA_SUCCESS) {
		printf("Selected key not found in DB. Quit...\n");
		goto error;
	}

	/* Set Key Secret */
	if (TpaHL_KEY_set(key, KEY_TPMKEYSECRET, keysecret_size, keysecret)
			!= TPA_SUCCESS) {
		printf(
				"It is not possible to set the secret in the key object. Quit...\n");
		goto error;
	}

	/* Read Input File */
	if (Tpa_BLOB_readFromFile(input_file, &input_blob, &input_blob_len)
			!= TPA_SUCCESS) {
		printf("It's not possible to read the input file! Quit...\n");
		goto error;

	}

	if (TpaHL_DATA_UnbindBlob(ctx, tpm, key, input_blob, input_blob_len,
			&output_blob, &output_blob_len) != TPA_SUCCESS) {
		printf("It's not possible to unbind the blob! Quit...\n");
		goto error;
	}

	/* Write Ouput into file */
	if (Tpa_BLOB_writeToFile(output_file, output_blob, output_blob_len)
			!= TPA_SUCCESS) {
		printf("It's not possible to read the input file! Quit...\n");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* free */
	TpaHL_TPM_freeMemory(tpm);
	TpaHL_CTX_freeMemory(ctx);

	return result;
}

int data(int argc, char ** argv) {
	TPA_RESULT res = TPA_ERROR;

	if (argc == 1) {
		data_usage(argv[0]);
		return TPA_ERROR;
	}
	if (strcmp("bind", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = data_bind(argc, argv);
	} else if (strcmp("unbind", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = data_unbind(argc, argv);
	} else {
		data_usage(argv[0]);
		return TPA_ERROR;
	}

	return res;
}
