/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "TPA_API.h"
#include "common.h"

void ek_usage(char * programName) {
	printf("Usage: ");

	printf("%s ek [SUBCOMMAND]\n", programName);

	printf("   certify: generate an EK certificate through a PCA");
	printf("\n\trequired parameters:\n");
	printf("\t   -o ownerpwd\tspecify the password of the owner of the TPM\n");
	//printf("\t   -p PCAcert \tspecify the PCA certificate file\n");
	printf("\toptional parameters:\n");
	printf("\t   -d address \tspecify the PCA address [default localhost]\n");
	printf("\t   -t port    \tspecify the PCA port [default 10000]\n");
	printf(
			"\t   -e ekfile  \tspecify the EK certificate file  [default ek.cert]\n");

	printf(
			"\n   extract: extract the EK certificate stored inside the TPM (IFX 1.2.1.2 only)");

	printf("\n\toptional parameters:\n");
	printf(
			"\t   -e ekfile  \tspecify the EK certificate filename [default ek.cert]\n");

}

int ek_certify(int argc, char **argv) {

	TPA_TPM *tpm = NULL;
	TPA_PCA *ca = NULL;
	TPA_RESULT result = TPA_ERROR;
	int c = 0;
	int counter = 0;

	char *pca_addr = NULL;
	int pca_port = 0;

	/*
	 * This function will allocate all needed objects
	 */
	if (TpaHL_TPM_allocate(&tpm) != TPA_SUCCESS)
		goto error;

	if (TpaHL_PCA_allocate(&ca) != TPA_SUCCESS)
		goto error;

	/*
	 * Parsing parameters
	 */

	opterr = 0;

	while ((c = getopt(argc, argv, "o:d:t:e:")) != -1) {
		switch (c) {
		case 'o':
			/*
			 * -o passes the password of the TPM
			 * We don't have a TpaHL_TPM_set yet, so we use a common strncpy
			 * A proper function for setting parameters is needed here!
			 */
			if (TpaHL_TPM_set(tpm, TPM_OWNERPWD, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 'p':
			if (TpaHL_PCA_set(ca, PCA_PCACERT_FILENAME, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 'e':
			if (TpaHL_TPM_set(tpm, TPM_EKCERT_FILENAME, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			break;
		case 'd':
			if (TpaHL_PCA_set(ca, PCA_URI, strlen(optarg), optarg) != TPA_SUCCESS)
				goto error;
			pca_addr = optarg;
			break;
		case 't':
			pca_port = atoi(optarg);
			break;
		case '?':
			/*
			 * Unknown parameter here
			 */
			ek_usage(argv[0]);
		default:
			/*
			 * Something went wrong here
			 */
			printf("Internal error!\n");
			goto error;
		}
	}

	if (counter != 1) {
		ek_usage(argv[0]);
		goto error;
	}

	/* Set Privacy CA uri */
	if (pca_addr == NULL) {
		if (TpaHL_PCA_set(ca, PCA_URI, strlen(TPA_PCA_DEFAULT_PCA_URI),
				TPA_PCA_DEFAULT_PCA_URI) != TPA_SUCCESS)
			goto error;
	}

	/* Set Privacy CA port */
	if (pca_port == 0) {
		pca_port = atoi(TPA_PCA_DEFAULT_PCA_PORT);
	}

	if (TpaHL_PCA_set(ca, PCA_PORT, sizeof(pca_port), &pca_port) != TPA_SUCCESS)
		goto error;

	/* Set EK Certificate filename */
	if (tpm->ekCertificateFilename == NULL) {
		if (TpaHL_TPM_set(tpm, TPM_EKCERT_FILENAME, strlen(
				TPA_EK_DEFAULT_CERT_FILENAME), TPA_EK_DEFAULT_CERT_FILENAME)
				!= TPA_SUCCESS)
			goto error;
	}

	/*
	 * We call the TpaHL_EK_certify to certify the EK
	 */
	if (TpaHL_EK_CreateCert(tpm, ca) == TPA_SUCCESS) {
		printf("EK stored in %s\n", tpm->ekCertificateFilename);
	} else {
		/*
		 * Something went wrong
		 */
		printf("EK not certified!\n");
		goto error;
	}

	result = TPA_SUCCESS;

	error: /* Error label */
	TpaHL_TPM_freeMemory(tpm);
	TpaHL_PCA_freeMemory(ca);

	return result;
}

int ek_retrieve_cert(int argc, char **argv) {

	TPA_RESULT result = TPA_ERROR;
	int c = 0;
	char *ekfile = NULL;

	opterr = 0;

	while ((c = getopt(argc, argv, "e:")) != -1) {
		switch (c) {
		case 'e':
			ekfile = strdup(argv[2]);
			break;
		case '?':
			/*
			 * Unknown parameter here
			 */
			ek_usage(argv[0]);
		default:
			/*
			 * Something went wrong here
			 */
			printf("Internal error!\n");
			goto error;
		}
	}

	/* Set EK Certificate filename */
	if (ekfile == NULL) {
		ekfile = strdup(TPA_EK_DEFAULT_CERT_FILENAME);
	}

	/*
	 * We call the TpaHL_EK_certify to certify the EK
	 */
	if (TpaHL_EK_RetrieveCert(ekfile) != TPA_SUCCESS) {
		printf("Operation failed!\n");
		goto error;
	} else {
		printf("EK certificate stored in %s\n", ekfile);
	}

	result = TPA_SUCCESS;
	error: /* Error label */
	if (ekfile)
		free(ekfile);

	return result;
}

int ek(int argc, char ** argv) {
	TPA_RESULT res = TPA_ERROR;

	if (argc == 1) {
		ek_usage(argv[0]);
		return TPA_ERROR;
	}
	if (strcmp("certify", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = ek_certify(argc, argv);
	} else if (strcmp("extract", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = ek_retrieve_cert(argc, argv);
	} else {
		ek_usage(argv[0]);
		return TPA_ERROR;
	}

	return res;
}

