/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "TPA_API.h"
#include "TPA_Utils.h"
#include "common.h"

void pstore_usage(char * programName) {
	printf("Usage: ");

	printf("%s pstore [SUBCOMMAND]\n", programName);

	printf("   seal: store and seal a softKEY");
	printf("\n\trequired parameters:\n");

	printf("\t   -s srkpwd  \tspecify the SRK password\n");
	printf("\t   -i inFile  \tspecify the input file\n");
	printf("\t   -o outFile\tspecify the output file\n");

	printf("\n\toptional parameters:\n");
	printf("\t   -u uri     \tspecify URI for TPA localstorage\n");
	printf(
			"\t   -p pcrs    \tspecify pcr used for sealing, example -p 1 -p 12 ...\n\n");

	printf("   unseal: store and seal a softKEY");
	printf("\n\trequired parameters:\n");

	printf("\t   -s srkpwd  \tspecify the SRK password\n");
	printf("\t   -i inFile  \tspecify the input file\n");
	printf("\t   -o outFile\tspecify the output file\n");

	printf("\n\toptional parameters:\n");
	printf("\t   -u uri     \tspecify URI for TPA localstorage\n");
}

int pstore_seal(int argc, char **argv) {

	TPA_RESULT result = TPA_ERROR;
	TPA_CONTEXT *ctx = NULL;
	TPA_TPM *tpm = NULL;
	TPA_PCR_SET *pcrSet = NULL;

	/* Filenames */
	char *input_file = NULL;
	char *output_file = NULL;

	unsigned char *input_blob = NULL;
	unsigned int input_blob_len = 0;
	unsigned char *output_blob = NULL;
	unsigned int output_blob_len = 0;

	char *lstorage = NULL;

	int c = 0;
	int counter = 0;
	int i = 0;
	int pcrs[24] = { 0 };
	int pindex = 0;

	/*
	 * This function will allocate all needed objects
	 */
	if (TpaHL_CTX_allocate(&ctx) != TPA_SUCCESS)
		goto error;

	if (TpaHL_TPM_allocate(&tpm) != TPA_SUCCESS)
		goto error;

	/*
	 * Parsing parameters
	 */

	opterr = 0;

	while ((c = getopt(argc, argv, "o:s:i:p:u:")) != -1) {
		switch (c) {
		case 'o':
			output_file = optarg;
			counter++;
			break;
		case 's':
			/*
			 * -s passes the password of the SRK
			 * We don't have a TpaHL_TPM_set yet, so we use a common strncpy
			 * A proper function for setting parameters is needed here!
			 */
			if (TpaHL_TPM_set(tpm, TPM_SRKPWD, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 'i':
			input_file = optarg;
			counter++;
			break;
		case 'u':
			if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			lstorage = optarg;
			break;
		case 'p':
			if (pindex > 24)
				break;
			pcrs[pindex] = atoi(optarg);
			if (pcrs[pindex] > 23) {
				break;
			} else {
				pindex++;
			}
			break;
		case '?':
			/*
			 * Unknown parameter here
			 */
			pstore_usage(argv[0]);
		default:
			/*
			 * Something went wrong here
			 */
			printf("Internal error!\n");
			goto error;
		}
	}

	if (counter != 3) {
		pstore_usage(argv[0]);
		goto error;
	}

	if (lstorage == NULL) {
		if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(TPA_DEFAULT_LOCAL_STORAGE),
				TPA_DEFAULT_LOCAL_STORAGE) != TPA_SUCCESS)
			goto error;
	}

	/*
	 * PCRS for sealing
	 */
	if (TpaHL_PCRSet_initialize(&pcrSet, pindex, 20) != TPA_SUCCESS)
		goto error;

	for (i = 0; i < pindex; i++) {
		if (TpaHL_PCRSet_pcr(pcrSet, pcrs[i], NULL) != TPA_SUCCESS)
			goto error;
	}

	/* Create Storage Master Key */
	if (TpaHL_SMK_Create(ctx, tpm) != TPA_SUCCESS) {
		printf("Sealing Master Key not created! Quit...\n");
		goto error;
	}

	/* Read Input File */
	if (Tpa_BLOB_readFromFile(input_file, &input_blob, &input_blob_len)
			!= TPA_SUCCESS) {
		printf("It's not possible to read the input file! Quit...\n");
		goto error;

	}

	/* Perform Sealing */
	if (TpaHL_SealBlob(ctx, tpm, input_blob, input_blob_len, pcrSet, &output_blob,
			&output_blob_len) != TPA_SUCCESS) {
		printf("It's not possible to seal the blob! Quit...\n");
		goto error;
	}

	/* Write Ouput into file */
	if (Tpa_BLOB_writeToFile(output_file, output_blob, output_blob_len)
			!= TPA_SUCCESS) {
		printf("It's not possible to read the input file! Quit...\n");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* free */
	TpaHL_TPM_freeMemory(tpm);
	TpaHL_PCRSet_freeMemory(pcrSet);
	TpaHL_CTX_freeMemory(ctx);

	return result;
}

int pstore_unseal(int argc, char **argv) {

	TPA_RESULT result = TPA_ERROR;
	TPA_CONTEXT *ctx = NULL;
	TPA_TPM *tpm = NULL;

	/* Filenames */
	char *input_file = NULL;
	char *output_file = NULL;

	unsigned char *input_blob = NULL;
	unsigned int input_blob_len = 0;
	unsigned char *output_blob = NULL;
	unsigned int output_blob_len = 0;

	char *lstorage = NULL;

	int c = 0;
	int counter = 0;

	/*
	 * This function will allocate all needed objects
	 */
	if (TpaHL_CTX_allocate(&ctx) != TPA_SUCCESS)
		goto error;

	if (TpaHL_TPM_allocate(&tpm) != TPA_SUCCESS)
		goto error;

	/*
	 * Parsing parameters
	 */

	opterr = 0;

	while ((c = getopt(argc, argv, "o:s:i:u:")) != -1) {
		switch (c) {
		case 'o':
			output_file = optarg;
			counter++;
			break;
		case 's':
			/*
			 * -s passes the password of the SRK
			 * We don't have a TpaHL_TPM_set yet, so we use a common strncpy
			 * A proper function for setting parameters is needed here!
			 */
			if (TpaHL_TPM_set(tpm, TPM_SRKPWD, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 'i':
			input_file = optarg;
			counter++;
			break;
		case 'u':
			if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			lstorage = optarg;
			break;
		case '?':
			/*
			 * Unknown parameter here
			 */
			pstore_usage(argv[0]);
		default:
			/*
			 * Something went wrong here
			 */
			printf("Internal error!\n");
			goto error;
		}
	}

	if (counter != 3) {
		pstore_usage(argv[0]);
		goto error;
	}

	if (lstorage == NULL) {
		if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(TPA_DEFAULT_LOCAL_STORAGE),
				TPA_DEFAULT_LOCAL_STORAGE) != TPA_SUCCESS)
			goto error;
	}

	/* Create Sealing Master Key */
	if (TpaHL_SMK_Create(ctx, tpm) != TPA_SUCCESS) {
		printf("Sealing Master Key not created! Quit...\n");
		goto error;
	}

	/* Read Input File */
	if (Tpa_BLOB_readFromFile(input_file, &input_blob, &input_blob_len)
			!= TPA_SUCCESS) {
		printf("It's not possible to read the input file! Quit...\n");
		goto error;

	}

	/* Perform Sealing */
	if (TpaHL_UnsealBlob(ctx, tpm, input_blob, input_blob_len, NULL, &output_blob,
			&output_blob_len) != TPA_SUCCESS) {
		printf("It's not possible to unseal the blob! Quit...\n");
		goto error;
	}

	/* Write Ouput into file */
	if (Tpa_BLOB_writeToFile(output_file, output_blob, output_blob_len)
			!= TPA_SUCCESS) {
		printf("It's not possible to read the input file! Quit...\n");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* free */
	TpaHL_TPM_freeMemory(tpm);
	TpaHL_CTX_freeMemory(ctx);

	return result;
}



int pstore(int argc, char ** argv) {
	TPA_RESULT res = TPA_ERROR;

	if (argc == 1) {
		pstore_usage(argv[0]);
		return TPA_ERROR;
	}
	if (strcmp("seal", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = pstore_seal(argc, argv);
	} else if (strcmp("unseal", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = pstore_unseal(argc, argv);
	} else {
		pstore_usage(argv[0]);
		return TPA_ERROR;
	}

	return res;
}
