/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "TPA_API.h"
#include "TPA_Utils.h"
#include "common.h"

void ra_usage(char * programName) {
	printf("Usage: ");

	printf("%s ra [SUBCOMMAND]\n", programName);

	printf("   quote: perform a remote attestation (quote)");
	printf("\n\trequired parameters:\n");
	printf("\t   -s srkpwd  \tspecify the SRK password\n");
	printf("\t   -a aikid   \tspecify the AIK to use\n");
	printf("\t   -k aikpwd  \tspecify the password of the AIK\n");
	printf("\t   -o outfile \tspecify the output file\n");

	printf("\toptional parameters:\n");
	printf("\t   -p number  \tspecify the PCR to quote "
		"(multiple allowed, default all)\n");
	printf("\t   -i infile  \tspecify the input file (nonce)\n");
	printf("\t   -u uri     \tspecify the URI for TPA localstorage\n");
	printf("\t   -M         \tspecify to append measurement logs"
		"to the output\n");
	printf("\t   -A         \tspecify to append AIK certificate (with chain)"
		"the output\n");

	printf("\n  verify: verify a remote attestation\n");
	printf("\n\trequired parameters:\n");
	printf("\t   -i infile  \tspecify the input file (nonce)\n");
	printf("\toptional parameter:\n");
	printf("\t   -p         \tprint the quoted PCRs\n");
	printf("\t   -v         \tverbose: print all attested data\n");
}

/*
 * This function creates an AIK and registers it in the local storage
 * The user must provide few parameter to the TPA:
 * - the password of the owner
 * - the password of the TPM
 * - the password of the AIK
 * - the label of the AIK
 */
int ra_quote(int argc, char **argv) {

	TPA_CONTEXT *ctx = NULL;
	TPA_TPM *tpm = NULL;
	TPA_PCR_SET *pcrSet = NULL;
	TPA_AIK *aik = NULL;
	TPA_RA *ra = NULL;

	TPA_RESULT result = TPA_ERROR;

	char *output_file = NULL;
	unsigned char *blob = NULL;
	unsigned int blob_size = 0;

	char *input_file = NULL;
	unsigned char *input_blob = NULL;
	unsigned int input_blob_len = 0;

	int c = 0;
	int counter = 0;
	int pcrs[24] = { 0 };
	int pindex = 0;
	int i = 0;

	char *lstorage = NULL;

	/*
	 * This function will allocate all needed objects
	 */
	if (TpaHL_CTX_allocate(&ctx) != TPA_SUCCESS)
		goto error;

	if (TpaHL_TPM_allocate(&tpm) != TPA_SUCCESS)
		goto error;

	if (TpaHL_AIK_allocate(&aik) != TPA_SUCCESS)
		goto error;

	if (TpaHL_RA_allocate(&ra) != TPA_SUCCESS)
		goto error;

	/*
	 * Parsing parameters
	 */

	opterr = 0;

	while ((c = getopt(argc, argv, "a:s:p:k:i:o:u:MA")) != -1) {
		switch (c) {
		case 'o':
			output_file = optarg;
			counter++;
			break;
		case 'i':
			input_file = optarg;
			break;
		case 's':
			if (TpaHL_TPM_set(tpm, TPM_SRKPWD, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 'a':
			// TODO: dirty hack!
			aik->aik_id = atoi(optarg);
			counter++;
			break;
		case 'k':
			if (TpaHL_AIK_set(aik, AIK_AIKSECRET, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 'u':
			if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			lstorage = optarg;
			break;
		case 'p':
			if (pindex > 24)
				break;
			pcrs[pindex] = atoi(optarg);
			if (pcrs[pindex] > 23) {
				break;
			} else {
				pindex++;
			}
			break;
		case 'M':
			break;
		case 'A':
			break;
		case '?':
			/*
			 * Unknown parameter here
			 */
			ra_usage(argv[0]);
		default:
			/*
			 * Something went wrong here
			 */
			printf("Internal error!\n");
			goto error;
		}
	}

	if (counter < 4) {
		ra_usage(argv[0]);
		goto error;
	}

	if (lstorage == NULL) {
		if (TpaHL_CTX_set(ctx, CTX_LS_URI, strlen(TPA_DEFAULT_LOCAL_STORAGE),
				TPA_DEFAULT_LOCAL_STORAGE) != TPA_SUCCESS)
			goto error;
	}

	/*
	 * PCRS to quote
	 */
	if (pindex > 0) {
		if (TpaHL_PCRSet_initialize(&pcrSet, pindex, 20) != TPA_SUCCESS)
			goto error;

		for (i = 0; i < pindex; i++) {
			if (TpaHL_PCRSet_pcr(pcrSet, pcrs[i], NULL) != TPA_SUCCESS)
				goto error;
		}
	}

	if (TpaHL_RA_Quote(ctx, tpm, aik, pcrSet, input_blob, input_blob_len, ra)
			!= TPA_SUCCESS) {
		fprintf(stderr, "Something goes wrong! :-(\n");
		goto error;
	}

	if (TpaHL_RA_Serialize(ra, &blob, &blob_size) != TPA_SUCCESS) {
		goto error;
	}

	/* Write Output into file */
	if (Tpa_BLOB_writeToFile(output_file, blob, blob_size) != TPA_SUCCESS) {
		fprintf(stderr, "Error writing to output file\n");
		goto error;
	}

	printf("Remote attestation successfull!\n");
	result = TPA_SUCCESS;

	error: /* Error label */
	TpaHL_AIK_freeMemory(aik);
	TpaHL_TPM_freeMemory(tpm);
	TpaHL_RA_freeMemory(ra);
	TpaHL_CTX_freeMemory(ctx);
	if (pcrSet)
		TpaHL_PCRSet_freeMemory(pcrSet);
	if (blob)
		free(blob);

	return result;
}

int ra_verify(int argc, char **argv) {

	TPA_CONTEXT *ctx = NULL;
	TPA_RA *ra = NULL;
	TPA_AIK *aik = NULL;
	TPA_PCR_SET *pcrSet = NULL;
	TPA_MEASURE_SET *measSet = NULL;
	TPA_VERIFICATION_RESULT res = TPA_VERIFICATION_NOK;

	TPA_RESULT result = TPA_ERROR;

	char *input_file = NULL;
	unsigned char *blob = NULL;
	unsigned int blob_size = 0;

	int output_pcrs = 0;
	int output_all = 0;

	int c = 0;
	int counter = 0;
	int i = 0;
	int j = 0;

	/*
	 * This function will allocate all needed objects
	 */
	if (TpaHL_CTX_allocate(&ctx) != TPA_SUCCESS)
		goto error;

	/*
	 * Parsing parameters
	 */

	opterr = 0;

	while ((c = getopt(argc, argv, "i:pv")) != -1) {
		switch (c) {
		case 'i':
			input_file = optarg;
			counter++;
			break;
		case 'p':
			output_pcrs = 1;
			break;
		case 'v':
			output_all = 1;
			break;
		default:
			ra_usage(argv[0]);
			goto error;
		}
	}

	if (counter < 1) {
		ra_usage(argv[0]);
		goto error;
	}

	/*
	 * PCRS to quote
	 */
	if (Tpa_BLOB_readFromFile(input_file, &blob, &blob_size) != TPA_SUCCESS) {
		fprintf(stderr, "Error reading from input file\n");
		goto error;
	}

	if (TpaHL_RA_Deserialize(blob, blob_size, NULL, &ra) != TPA_SUCCESS) {
		fprintf(stderr, "Error reading the RA information\n");
		goto error;
	}

	if (TpaHL_RA_Verify(ctx, ra, &aik, NULL, NULL, &pcrSet, &measSet, &res) != TPA_SUCCESS) {
		fprintf(stderr, "Something goes wrong! :-(\n");
		goto error;
	}

	if (res == TPA_VERIFICATION_OK)
		printf("Remote attestation: verification successful!\n");
	else
		printf("Remote attestation: verification failed.\n");
	result = TPA_SUCCESS;

	/* Write output */
	if (output_pcrs || output_all) {
		printf("\nPCRS:\n");
		if (pcrSet == NULL || pcrSet->index <= 0)
			printf("   none\n");
		else {
			for (i = 0; i < pcrSet->index; i++) {
				printf("%d -> 0x", pcrSet->pcrList[i].pcr);
				for (j = 0; j < pcrSet->pcr_size; j++)
					printf("%0x", pcrSet->pcrList[i].value[j]);
				printf("\n");
			}
		}
	}

	if (output_all) {
		printf("\nMEASURES:\n");
		printf("\nAIK (Cert):\n");
		if (aik == NULL || aik->cert == NULL)
			printf("   none\n");
		else
			printf("%s\n", aik->cert);
	}

	error: /* Error label */
	TpaHL_CTX_freeMemory(ctx);
	TpaHL_RA_freeMemory(ra);
	if (aik)
		TpaHL_AIK_freeMemory(aik);
	if (pcrSet)
		TpaHL_PCRSet_freeMemory(pcrSet);
	if (measSet)
		TpaHL_MeasureSet_freeMemory(measSet);
	if (blob)
		free(blob);

	return result;
}

int ra(int argc, char ** argv) {
	TPA_RESULT res = TPA_ERROR;

	if (argc == 1) {
		ra_usage(argv[0]);
		return TPA_ERROR;
	}
	if (strcmp("quote", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = ra_quote(argc, argv);
	} else if (strcmp("verify", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = ra_verify(argc, argv);
	} else {
		ra_usage(argv[0]);
		return TPA_ERROR;
	}

	return res;
}

