/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "TPA_API.h"
#include "common.h"

void tpm_usage(char * programName) {
	printf("Usage: ");

	printf("%s tpm [SUBCOMMAND]\n", programName);

	printf("   takeownership: take the owership of the TPM");
	printf("\n\trequired parameters:\n");
	printf("\t   -o ownerpwd\tspecify the password of the owner of the TPM\n");
	printf("\t   -s srkpwd  \tspecify the SRK password\n");

	printf("\n   isowned: verify if the owership of the TPM is already taken\n");

}


int tpm_takeownership(int argc, char **argv) {

	TPA_TPM *tpm = NULL;
	TPA_CONTEXT *ctx = NULL;
	TPA_RESULT result = TPA_ERROR;
	int c = 0;
	int counter = 0;

	/*
	 * This function will allocate all needed objects
	 */
	if (TpaHL_CTX_allocate(&ctx) != TPA_SUCCESS)
		goto error;

	if (TpaHL_TPM_allocate(&tpm) != TPA_SUCCESS)
		goto error;

	/*
	 * Parsing parameters
	 */

	opterr = 0;

	while ((c = getopt(argc, argv, "o:s:")) != -1) {
		switch (c) {
		case 'o':
			/*
			 * -o passes the password of the TPM
			 * We don't have a TpaHL_TPM_set yet, so we use a common strncpy
			 * A proper function for setting parameters is needed here!
			 */
			if (TpaHL_TPM_set(tpm, TPM_OWNERPWD, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case 's':
			/*
			 * -s passes the password of the SRK
			 * We don't have a TpaHL_TPM_set yet, so we use a common strncpy
			 * A proper function for setting parameters is needed here!
			 */
			if (TpaHL_TPM_set(tpm, TPM_SRKPWD, strlen(optarg), optarg)
					!= TPA_SUCCESS)
				goto error;
			counter++;
			break;
		case '?':
			/*
			 * Unknown parameter here
			 */
			tpm_usage(argv[0]);
		default:
			/*
			 * Something went wrong here
			 */
			printf("Internal error!\n");
			goto error;
		}
	}

	if (counter != 2) {
		tpm_usage(argv[0]);
		goto error;
	}

	if(TpaHL_TPM_TakeOwnership(ctx, tpm) != TPA_SUCCESS) {
		printf("TpaHL_TPM_TakeOwnership failed\n");
		goto error;
	} else
		printf("TakeOwnership operation successful\n");

	result = TPA_SUCCESS;


	error: /* Error label */
	TpaHL_TPM_freeMemory(tpm);
	TpaHL_CTX_freeMemory(ctx);

	return result;
}


int tpm_isowned(int argc, char **argv) {

	TPA_TPM *tpm = NULL;
	TPA_CONTEXT *ctx = NULL;
	TPA_RESULT result = TPA_ERROR;
	TPA_TPM_OWNERSHIP_RESULT isTaken = TPM_IS_NOT_OWNED;

	/*
	 * This function will allocate all needed objects
	 */
	if (TpaHL_CTX_allocate(&ctx) != TPA_SUCCESS)
		goto error;

	if (TpaHL_TPM_allocate(&tpm) != TPA_SUCCESS)
		goto error;

	if(TpaHL_TPM_IsOwnershipTaken(ctx, tpm, &isTaken) != TPA_SUCCESS) {
		printf("TpaHL_TPM_IsOwnershipTaken failed\n");
		goto error;
	} else {
		if(isTaken == TPM_IS_OWNED)
			printf("The TPM is already owned\n");
		else if (isTaken == TPM_IS_NOT_OWNED)
			printf("The TPM is not owned\n");
		else
			printf("Internal error\n");
	}
	result = TPA_SUCCESS;


	error: /* Error label */
	TpaHL_TPM_freeMemory(tpm);
	TpaHL_CTX_freeMemory(ctx);

	return result;
}



int tpm(int argc, char ** argv) {
	TPA_RESULT res = TPA_ERROR;

	if (argc == 1) {
		tpm_usage(argv[0]);
		return TPA_ERROR;
	}
	if (strcmp("takeownership", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = tpm_takeownership(argc, argv);
	} else if (strcmp("isowned", argv[1]) == 0) {
		argc--;
		argv[1] = argv[0];
		argv++;
		res = tpm_isowned(argc, argv);
	} else {
		tpm_usage(argv[0]);
		return TPA_ERROR;
	}

	return res;
}



#if 0
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "TPA_API.h"
#include "TPA_TPM.h"

#define FIELD_SIZE 50
#define PCR_SIZE 20

void tpm_usage (char * programName) {
	printf ("Usage:\n");

	printf ("%s tpm [SUBCOMMAND]\n", programName);

	printf ("   quote,\t\tquote using an AIK\n");
	printf ("\toptions:\n");
	printf ("\t   -l label\tspecify the label of the AIK\n");
}

/*
 * This function will use an AIK and a set of PRCs to perform a quote operation
 * using the TPM. The function takes as input from the command line:
 * - the label of the AIK to use
 * - the password of the AIK
 * - the password of the SRK
 * Also the set of PCRs is needed, but in this example it is not taken as input,
 * but it is fixed
 */
int tpm_quote (int argc, char ** argv) {
	int c;
	TPA_AIK *aik;
	//char *cvalue = NULL;
	TPA_CONTEXT *ctx = (TPA_CONTEXT *) malloc (sizeof (TPA_CONTEXT));
	TPA_TPM tpm;

	TPA_PCR_SET_OLD pcr_set;

	int i, j;
	int counter;
	long len;
	unsigned char * quoted;

	long externalDataLength;
	unsigned char * externalData;

	/*
	 * This function will allocate a new AIK object
	 */
	TpaHL_AIK_allocate (&aik);

	/*
	 * We create a TPM object
	 * A constructor is needed here!
	 */
	tpm.ownSecret = (char *) malloc (sizeof(unsigned char) * FIELD_SIZE);
	tpm.srkSecret = (char *) malloc (sizeof(unsigned char) * FIELD_SIZE);

	if (argc == 1) {
		tpm_usage (argv[0]);
		exit (0);
	}
	while ((c = getopt (argc, argv, "o:s:l:a:")) != -1) {
		switch (c) {
			case 'l':
			/*
			 * -l passes the label of the AIK
			 * We use TpaHL_AIK_set to set the parameter in the object
			 */
			TpaHL_AIK_set (aik, AIK_LABEL, strlen(optarg), optarg);
			break;
			case 'a':
			/*
			 * -a passes the password of the AIK
			 * We use TpaHL_AIK_set to set the parameter in the object
			 */
			TpaHL_AIK_set (aik, AIK_AIKSECRET, strlen(optarg), optarg);
			break;
			case 's':
			/*
			 * -s passes the password of the SRK
			 * We don't have a TpaHL_TPM_set yet, so we use a common strncpy
			 * A proper function for setting parameters is needed here!
			 */
			strncpy (tpm.srkSecret, optarg, FIELD_SIZE);
			break;
			case '?':
			/*
			 * Unknown parameter here
			 */
			tpm_usage (argv[0]);
			exit (0);
			break;
			default:
			/*
			 * Something went wrong here
			 */
			printf ("Internal error!\n");
			exit (0);
			break;
		}
	}

	/*
	 * We reset the PCR set
	 * A TPA_PCR_SET is a vector composed of 24 TPA_PCR objects
	 *
	 * A TPA_PCR is an object that contains a int (TPA_PCR.pcr) that indicates the PCR number
	 * and a unsigned char* (TPA_PCR.value) that contains the value of the PCR.
	 *
	 * For initialization, we set the TPA_PCR.pcr equal to the object's
	 * position in the TPA_PCR_SET vector and the TPA_PCR.value = NULL
	 */
	for (i=0; i<24; i++) {
		pcr_set.pcrList[i].pcr = i;
		pcr_set.pcrList[i].value = NULL;
	}

	/*
	 * Here we indicate which PCRs we want to use for quoting
	 *
	 * The TPA will use a PCR indicated by TPA_PCR.pcr
	 * if the related TPA_PCR.value is not NULL
	 * In this example, we use the PCRs from 2 to 16
	 */
	for (i=2; i<17; i++) {
		pcr_set.pcrList[i].value = (TPA_PCR_VALUE *) malloc (sizeof (TPA_PCR_VALUE) * PCR_SIZE);
	}

	/*
	 * The TPM_Quote requires some external data to be added to the signature
	 *
	 * In this example the external data are fixed, but in real applications
	 * data should be provided by the remote entity that requests the quote
	 *
	 * External data should be 20 bytes long
	 */
	externalDataLength = 20;
	externalData = (BYTE *) malloc (sizeof (BYTE) * externalDataLength);

	for (i=0; i<externalDataLength; i++) {
		externalData[i] = i + 16;
	}

	/*
	 * We call the TpaHL_Quote here
	 */
	if (TpaHL_Quote (ctx, &tpm, aik, &pcr_set, externalDataLength, externalData, &len, &quoted) == TPA_SUCCESS) {
		/*
		 * Everything went fine:
		 * - quoted contains the signature
		 * - len contains the length of the signature
		 * - pcr_set contains the PCRs set. The TPA copied the value
		 *   of each PCR requested for the Quote in the related TPA_PCR.value.
		 * - aik is an TPA_AIK object that was filled with the certificate
		 *   of the AIK used for the Quote
		 */
		printf ("Quote\n");

		//printf ("TPM_Quote result len = %d\n", len);

		/*
		 * We print the signature here
		 */
		counter = 0;

		printf ("0000000 ");
		for (i=0; i<len; i+=2) {
			counter ++;
			printf ("%02x%02x ", quoted[i+1], quoted[i]);
			if (counter % 8 == 0) {
				printf ("\n%07x ", (i+2));
			}
		}

		/*
		 * We print the certificate of the AIK used for the Quote here
		 */
		//printf ("\n\tcert blob size: %ld\n", aik->cert_size);

		counter = 0;

		printf ("0000000 ");
		for (i=0; i<aik->cert_size; i+=2) {
			counter ++;
			printf ("%02x%02x ", aik->cert[i+1], aik->cert[i]);
			if (counter % 8 == 0) {
				printf ("\n%07x ", (i+2));
			}
		}

		printf ("\n");

		/*
		 * We print the PCRs used for the Quote here
		 */
		printf ("PCR used for the Quote:\n");

		for (i=0; i<24; i++) {
			if (pcr_set.pcrList[i].value == NULL) {
				/*
				 * Remember that only the TPA_PCR with TPA_PCR.value != NULL
				 * were used for the Quote!
				 */
				printf ("PCR %d: not used", i);
			}
			else {
				/*
				 * If TPA_PCR.value != NULL, this means that the PCR was used
				 * for the Quote; in this case we print its value
				 */
				printf ("PCR %d:", i);
				for (j = 0; j< PCR_SIZE; j++) {
					printf ("%02x ", pcr_set.pcrList[i].value[j]);
				}
			}
			printf ("\n");
		}
	}
	else {
		/*
		 * Something was wrong
		 */
		printf ("Not quote!\n");
	}
	return TPA_ERROR;
}

int tpm (int argc, char ** argv) {
	if (argc == 1) {
		tpm_usage (argv[0]);
		exit (0);
	}
	if (strcmp ("quote", argv[1]) == 0) {
		argc --;
		argv[1] = argv [0];
		argv ++;
		tpm_quote (argc, argv);
	}
	return TPA_ERROR;
}
#endif
