/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>
#include <string.h>

#include "TPA_AIK.h"
#include "TPA_TSS.h"
#include "TPA_PCA.h"
#include "TPA_EK.h"
#include "TPA_Utils.h"

TPA_RESULT Tpa_AIK_Create(TPA_TPM *tpm, TPA_AIK *aik, TPA_PCA *ca) {

	TPA_TSS_ACCESS *tss = NULL;
	unsigned char *request = NULL;
	unsigned int request_size = 0;
	unsigned char* asymBlob_bin = NULL;
	int asymBlob_bin_size = 0;
	unsigned char* symBlob_bin = NULL;
	int symBlob_bin_size = 0;

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("TPA_TPM structure is NULL");
		return TPA_ERROR;
	}
	if (aik == NULL) {
		TRACE_MSG_PARAM_ERROR("TPA_AIK structure is NULL");
		return TPA_ERROR;
	}
	if (ca == NULL) {
		TRACE_MSG_PARAM_ERROR("TPA_PCA structure is NULL");
		return TPA_ERROR;
	}
	if (tpm->ownSecret == NULL) {
		TRACE_MSG_PARAM_ERROR("The TPM owner secret is not set");
		return TPA_ERROR;
	}
	if (tpm->ownSecret_len != strlen(tpm->ownSecret)) {
		TRACE_MSG_PARAM_ERROR("ownSecret_len in TPA_TPM structure is not correct");
		return TPA_ERROR;
	}
	if (tpm->srkSecret == NULL) {
		TRACE_MSG_PARAM_ERROR("The TPM srk secret is not set");
		return TPA_ERROR;
	}
	if (tpm->srkSecret_len != strlen(tpm->srkSecret)) {
		TRACE_MSG_PARAM_ERROR("srkSecret_len in TPA_TPM structure is not correct");
		return TPA_ERROR;
	}
	if (aik->aikSecret == NULL) {
		TRACE_MSG_PARAM_ERROR("The AIK secret is not set");
		return TPA_ERROR;
	}
	if (aik->aikSecret_len != (strlen(aik->aikSecret))) {
		TRACE_MSG_PARAM_ERROR("aikSecret_len in TPA_AIK structure is not correct");
		return TPA_ERROR;
	}
	if (aik->label == NULL) {
		TRACE_MSG_PARAM_ERROR("The AIK label is not set");
		return TPA_ERROR;
	}
	if (aik->cert != NULL || aik->cert_size != 0) {
		TRACE_MSG_PARAM_ERROR("There is a certificate in the AIK structure");
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	if (ca->pcaCertificateFilename == NULL) {
		if (Tpa_PCA_Helper_DownloadPCACert(ca) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaHL_BLOB_readFromFile failed");
			goto error;
		}
	} else {
		if (ca->pcaCertificate != NULL && ca->pcaCertificate_size != 0) {
			TRACE_MSG_ERROR("pcaCertificate in CA structure is already loaded");
			goto error;
		}

		if (Tpa_BLOB_readFromFile(ca->pcaCertificateFilename,
				&(ca->pcaCertificate), &(ca->pcaCertificate_size))
				!= TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaHL_BLOB_readFromFile failed");
			goto error;
		}
	}

	/* Read EK Certificate */
	if (Tpa_EK_ReadKey(tpm) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_EK_ReadKey failed");
		goto error;
	}

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	/* Set owner secret from TPM structure */
	if (Tpa_TSS_SetOwnerSecret(tss, tpm->ownSecret, tpm->ownSecret_len)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_SetOwnerSecret failed");
		goto error;
	}

	if (Tpa_TSS_LoadSRK(tss, tpm->srkSecret, tpm->srkSecret_len) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}

	if (Tpa_TSS_CreateAIKkey(tss, aik->aikSecret, aik->aikSecret_len)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_CreateAIKkey failed");
		goto error;
	}

	if (Tpa_TSS_CreatePCAkey(tss, ca) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_CreatePCAkey failed");
		goto error;
	}

	if (Tpa_TSS_SetEKcert(tss, tpm->ekCertificate, tpm->ekCertificate_size)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_SetEKcert failed");
		goto error;
	}

	if (Tpa_TSS_CollateIdentityRequest(tss, aik->labelUTF16,
			aik->labelUTF16_len, &request, &request_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_CollateIdentityRequest failed");
		goto error;
	}

	if (Tpa_PCA_Helper_AIKRequest(ca, (char *) request, (int) request_size,
			(char **) &asymBlob_bin, (int *) &asymBlob_bin_size,
			(char **) &symBlob_bin, (int *) &symBlob_bin_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_IAIK_Helper_SendRequestToPCA failed");
		goto error;
	}

	if (Tpa_TSS_ActivateIdentity(tss, asymBlob_bin, asymBlob_bin_size,
			symBlob_bin, symBlob_bin_size, &(aik->cert), &(aik->cert_size))
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_ActivateIdentity failed");
		goto error;
	}

	//TRACE_MSG("received AIK credential");

	if (Tpa_TSS_GetKeyBlob(tss, tss->hIdentKey, &(aik->keyBlob),
			&(aik->keyBlob_size)) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetKeyBlob failed");
		goto error;
	}

	/* Extract and store AIK public key */
	if (Tpa_TSS_GetKeyModulus(tss, tss->hIdentKey, &(aik->pubKey),
			&(aik->pubKey_size)) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetKeyBlob failed");
		goto error;

	}

	result = TPA_SUCCESS;

	error: /* Free! */
	if (!tss)
		Tpa_TSS_Cleanup(tss);
	if (!asymBlob_bin)
		free(asymBlob_bin);
	if (!symBlob_bin)
		free(symBlob_bin);
	if (!request)
		free(request);

	return result;
}

TPA_RESULT Tpa_AIK_Locate(TPA_AIK *aik, TPA_PCA *ca,
		TPA_CERTIFICATE_ITEM **cert_list) {

	TPA_RESULT result = TPA_ERROR;

	if (Tpa_PCA_Helper_AIKLocate(ca, aik->label, aik->label_len, cert_list)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_IAIK_Helper_AIKLocate failed");
		goto error;
	}

	result = TPA_SUCCESS;

	error: return result;
}

TPA_RESULT Tpa_AIK_Validate(TPA_AIK *aik, TPA_PCA *ca,
		TPA_CERTIFICATE_ITEM **cert_list, TPA_VERIFICATION_RESULT *ver) {

	TPA_RESULT result = TPA_ERROR;
	TPA_VERIFICATION_RESULT vres = 0;

	*ver = TPA_VERIFICATION_NOK;

	if (Tpa_PCA_Helper_Validate(ca, aik->cert, aik->cert_size, cert_list, &vres)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_IAIK_Helper_Validate failed");
		goto error;
	}

	if (vres == 0)
		*ver = TPA_VERIFICATION_OK;

	result = TPA_SUCCESS;
	error: return result;
}

TPA_RESULT Tpa_AIK_Revoke(TPA_TPM *tpm, TPA_AIK *aik, TPA_PCA *ca) {
	return TPA_NOT_IMPLEMENTED;
}

