/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include <string.h>
#include <malloc.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>

#include <openssl/x509.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/pem.h>
#include <netinet/in.h>

#include "TPA_API.h"
#include "TPA_Core.h"
#include "TPA_Utils.h"

/****************************************************************************/
/*																			*/
/*							 Local Storage									*/
/*																			*/
/****************************************************************************/

/**
 * Creates and initializes a local storage object
 *
 * This function is used to create a file called tpa.db in a folder specified
 * in the context object. This file will contains the whole list of the keys
 * stored in the storage and the association between them
 *
 * @param context a pointer to a ::TPA_CONTEXT object
 */
TPA_RESULT TpaHL_LS_create(TPA_CONTEXT *context) {

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}

	if (context->ls_uri == NULL) {
		TRACE_MSG_PARAM_ERROR("context->ls_uri is NULL");
		return TPA_ERROR;
	}

#endif

	/* init DB*/
	if (Tpa_LS_Init(context->measure_mod->ls) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_LS_Init failed");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* free */
	return result;
}

/****************************************************************************/
/*																			*/
/*     						 Endorsement Key								*/
/*																			*/
/****************************************************************************/

/**
 * Retrieves the EK Certificate stored in the TPM NV storage (IFX only)
 *
 * @param ekfile a pointer to a buffer that contains the name of the location
 * where the certificate will be stored in the disk
 */
TPA_RESULT TpaHL_EK_RetrieveCert(char *ekfile) {

	TPA_RESULT result = TPA_ERROR;
	TPA_TPM *tpm = NULL;

#ifndef __NO_PARAM_CHECK
	if (ekfile == NULL) {
		TRACE_MSG_PARAM_ERROR("ekfile parameter cannot be null");
		goto error;
	}
#endif

	if (TpaHL_TPM_allocate(&tpm) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaHL_TPM_allocate failed");
		goto error;
	}

	if (Tpa_EK_ReadKeyTPM_ifx(tpm) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_EK_ReadKeyTPM_ifx failed");
		goto error;
	}

	if (Tpa_BLOB_writeToFile(ekfile, tpm->ekCertificate,
			tpm->ekCertificate_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaHL_BLOB_writeBlobToFile EKcert failed");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* Free */
	TpaHL_TPM_freeMemory(tpm);
	return result;
}

/**
 * Generates a certificate for the Endorsement Key
 *
 * This function generate a certificate for the TPM Endorsement Key
 * interacting with a privacy CA that use the APKI protocol
 *
 * @param tpm a pointer to the ::TPA_TPM structure
 * @param pca a pointer to the ::TPA_PCA structure
 */
TPA_RESULT TpaHL_EK_CreateCert(TPA_TPM *tpm, TPA_PCA *pca) {

#ifndef __NO_PARAM_CHECK
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm is NULL");
		return TPA_ERROR;
	}

	if (pca == NULL) {
		TRACE_MSG_PARAM_ERROR("pca is NULL");
		goto error;
	}

	if (tpm->ekCertificateFilename == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm->ekCertificateFilename is NULL");
		goto error;
	}

	if (tpm->ownSecret == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm->ownSecret is NULL");
		goto error;
	}

	if (tpm->ownSecret_len == 0) {
		TRACE_MSG_PARAM_ERROR("tpm->ownSecret_len is NULL");
		goto error;
	}

	if (pca->uri == NULL) {
		TRACE_MSG_PARAM_ERROR("pca uri is NULL");
		goto error;
	}

	if (pca->port == 0) {
		TRACE_MSG_PARAM_ERROR("pca port is 0");
		goto error;
	}
#endif

	if (Tpa_EK_Create(tpm, pca) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_EK_Create failed");
		goto error;
	}

	if (Tpa_BLOB_writeToFile(tpm->ekCertificateFilename, tpm->ekCertificate,
			tpm->ekCertificate_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaHL_BLOB_writeBlobToFile failed");
		goto error;
	}

	return TPA_SUCCESS;

	error: /* free */
	if (tpm->ekCertificate != NULL)
		free(tpm->ekCertificate);
	tpm->ekCertificate_size = 0;
	return TPA_ERROR;
}

/****************************************************************************/
/*																			*/
/*     					   Sealing Master Key								*/
/*																			*/
/****************************************************************************/
/**
 * Generates a Storage Master Key (SMK)
 *
 * This function generates a Storage Master Key, a TPM key that is used as
 * root for all storage operations inside the TPA. The SMK is unique and is
 * a child of the Storage Root Key (SRK)
 *
 * @param context a pointer to a ::TPA_CONTEXT object
 * @param tpm a pointer to a ::TPA_TPM object
 */
TPA_RESULT TpaHL_SMK_Create(TPA_CONTEXT *context, TPA_TPM *tpm) {

	TPA_RESULT result = TPA_ERROR;
	TPA_GENERIC_OBJPTR pmks = { 0 };
	TPA_GENERIC_TPMKEY mks = { 0 };
	TPA_LS lstorage = { 0 };

	unsigned char *blobKEY = NULL;
	unsigned int blobKEY_len = 0;

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		goto error;
	}
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm is NULL");
		goto error;
	}
#endif

	/* Open Local Storage */
	if (TpaLS_DB_Open(context, &lstorage) != TPA_SUCCESS)
		goto error;

	if (TpaLS_DB_LookForStorageMK(context, &pmks, &lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_LoadStorageMK failed");
		goto error;
	}

	if (pmks.saved_objid == DB_KEY_NOT_FOUND) { /* SMK is not found in the storage */
		if (Tpa_CreateMKBlob(tpm, &blobKEY, &blobKEY_len) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaHL_CreateMKBlob failed");
			goto error;
		}

		mks.blob = blobKEY;
		mks.blob_size = blobKEY_len;
		pmks.objptr = &mks;
		pmks.type = DB_STORAGEMK_TYPE;

		/* Store Key */
		if (TpaLS_DB_Store(context, &pmks, &lstorage) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaLS_DB_Store failed");
			goto error;
		}

	} else {
		TRACE_MSG_INFO("SMK already created");
	}

	/* Open Local Storage */
	if (TpaLS_DB_Close(context, &lstorage) != TPA_SUCCESS)
		goto error;

	result = TPA_SUCCESS;
	error: /* Free memory */
	if (blobKEY)
		free(blobKEY);
	return result;
}

/****************************************************************************/
/*																			*/
/*							 Generic TPM KEY								*/
/*																			*/
/****************************************************************************/

/**
 * Writes the public part of a TPM key into a file
 *
 * This function extracts the public part of a TPM key blob and save it in a
 * file which name is specified as parameter. Note that for now the exponent
 * is not extracted from the blob but is the value 65537
 *
 * @param filename a string that specifies the name of the file where the blob
 * will be stored
 * @param blob a buffer that contains the key blob
 * @param blob_len an integer that contains the length of the key blob
 */
TPA_RESULT TpaHL_GenTpmKey_writePubKeyToFile(char *filename,
		unsigned char *blob, int blob_len) {

	RSA *key = NULL;
	unsigned char exp[] = { 0x01, 0x00, 0x01 };

	BIO *out = NULL;

	out = BIO_new(BIO_s_file());
	if (out == NULL) {
		return TPA_ERROR;
	}

	key = RSA_new();
	if (key == NULL) {
		return TPA_ERROR;
	}

	/* set the public key value in the OpenSSL object */
	key->n = BN_bin2bn(blob, blob_len, 0);
	/* set the public exponent */
	key->e = BN_bin2bn(exp, sizeof(exp), 0);

	BIO_write_filename(out, filename);
	PEM_write_bio_RSA_PUBKEY(out, key);

	RSA_free(key);
	BIO_free_all(out);

	return TPA_SUCCESS;
}

/****************************************************************************/
/*																			*/
/*   	     		 Attestation Identity Key (AIK)							*/
/*																			*/
/****************************************************************************/

/**
 * Creates an AIK, requests a certificate to a privacy CA and stores the certificate in the local storage
 *
 * @param context a pointer to a ::TPA_CONTEXT object
 * @param tpm a pointer to a ::TPA_TPM that contains the information related with the TPM like owner secret,
 * and SRK secret
 * @param aik a pointer to an ::TPA_AIK object that contains all the information related with the AIK
 * @param pca a pointer to a ::TPA_CA object that contains all the information related with the PCA like the URI and any
 * authorization data related with it
 */
TPA_RESULT TpaHL_AIK_Create(TPA_CONTEXT *context, TPA_TPM *tpm, TPA_AIK *aik,
		TPA_PCA *pca) {

	TPA_RESULT result = TPA_ERROR;
	TPA_LS *lstorage = NULL;
	TPA_GENERIC_OBJPTR gk = { 0 };

	if (aik == NULL) {
		TRACE_MSG_ERROR("aik must be allocated");
	}

	if (context == NULL || tpm == NULL || pca == NULL) {
		/* TODO: FIX ME more checks here */
		TRACE_MSG_BUG("default objects are needed");
		goto error;
	}

	if (aik->aikSecret == NULL || aik->aikSecret_len == 0 || aik->label == NULL
			|| aik->label_len == 0) {
		TRACE_MSG_ERROR("aik structure must be set");
		goto error;
	}

	if (context->ls_uri == NULL || context->ls_uri_size == 0) {
		TRACE_MSG_BUG("set with default parameters");
		goto error;
	}

	if (tpm->ownSecret == NULL || tpm->ownSecret_len == 0 || tpm->srkSecret
			== NULL || tpm->srkSecret_len == 0) {
		TRACE_MSG_BUG("set with default parameters");
		goto error;
	}

	/* if (pca->pcaCertificateFilename == NULL && (pca->pcaCertificate == NULL
	 || pca->pcaCertificate_size == 0)) {
	 TRACE_MSG_ERROR("ca structure must be set");
	 goto error;
	 }*/

	/* Create AIK Key with TPA_AIK API */
	if (Tpa_AIK_Create(tpm, aik, pca) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_AIK_Create failed");
		goto error;
	}

	if (TpaLS_LS_allocate(&lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_LS_allocate failed");
		goto error;
	}

	/* Open Local Storage */
	if (TpaLS_DB_Open(context, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_Open failed");
		goto error;
	}

	/* Dump AIK Key on disk */
	gk.type = 0;
	gk.objptr = (TPA_OBJ_PTR) aik;
	if (TpaLS_DB_Store(context, &gk, lstorage) != TPA_SUCCESS)
		goto error;

	if (TpaLS_DB_Close(context, lstorage) != TPA_SUCCESS)
		goto error;

	result = TPA_SUCCESS;
	error: /* Free */
	TpaLS_LS_freeMemory(lstorage);
	return result;
}

TPA_RESULT TpaHL_AIK_Certify(TPA_CONTEXT *context, TPA_TPM *tpm, TPA_AIK *aik,
		TPA_PCA *ca) {
	/* TODO: implement TpaHL_AIK_Certify */
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT TpaHL_AIK_List(TPA_CONTEXT *context, TPA_AIK **aikList,
		int *aikList_size, unsigned int flags) {

	TPA_RESULT result = TPA_ERROR;
	TPA_LS *lstorage = NULL;
	int *idList = NULL;
	int idList_size = 0;
	int i = 0;

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}

	if (aikList == NULL) {
		TRACE_MSG_PARAM_ERROR("aikList is NULL");
		goto error;
	}

	if (aikList_size == NULL) {
		TRACE_MSG_PARAM_ERROR("aikList_size is NULL");
		goto error;
	}
#endif

	if (flags != 0) {
		TRACE_MSG_INFO("flags is not valid");
		goto error;
	}

	if (TpaLS_LS_allocate(&lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_LS_allocate failed");
		goto error;
	}

	/* Open Local Storage */
	if (TpaLS_DB_Open(context, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_Open failed");
		goto error;
	}

	if (TpaLS_DB_List(context, lstorage, DB_AIK_TYPE, &idList, &idList_size)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_List failed");
		goto error;
	}

	*aikList = (TPA_AIK *) calloc(idList_size, sizeof(TPA_AIK));
	if (*aikList == NULL) {
		TRACE_MSG_ERROR("calloc failed");
		goto error;
	}
	*aikList_size = idList_size;

	for (i = 0; i < idList_size; i++) {
		(*aikList)[i].aik_id = idList[i];
	}

	result = TPA_SUCCESS;
	goto end;

	error: /* error label */
	if (*aikList) {
		free(*aikList);
		*aikList = NULL;
	}
	*aikList_size = 0;

	end: /* end label */
	TpaLS_DB_Close(context, lstorage);
	TpaLS_LS_freeMemory(lstorage);
	if (idList)
		free(idList);

	return result;
}

TPA_RESULT TpaHL_AIK_Retrieve(TPA_CONTEXT *context, TPA_AIK *aik) {
	/* TODO: implement TpaHL_AIK_Retrieve */
	return TPA_NOT_IMPLEMENTED;
}

/**
 * Retrieves a ::TPA_AIK using its id
 *
 * @param context a pointer to a ::TPA_CONTEXT object
 * @param aik a double pointer to a ::TPA_AIK object which is allocated and filled by the function
 * @param id the id of the key to retrieve
 */
TPA_RESULT TpaHL_AIK_RetrieveById(TPA_CONTEXT *context, TPA_AIK **aik,
		TPA_OBJ_ID id) {

	TPA_RESULT result = TPA_ERROR;
	TPA_GENERIC_OBJPTR gk = { 0 };
	TPA_LS *lstorage = NULL;

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}

	if (aik == NULL) {
		TRACE_MSG_PARAM_ERROR("aik is NULL");
		return TPA_ERROR;
	}

	if (id <= 0) {
		TRACE_MSG_PARAM_ERROR("id is not valid");
		return TPA_ERROR;
	}
#endif

	if (TpaLS_LS_allocate(&lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_LS_allocate failed");
		goto error;
	}

	/* Open Local Storage */
	if (TpaLS_DB_Open(context, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_Open failed");
		goto error;
	}

	gk.saved_objid = id;
	if (TpaLS_DB_LoadById(context, &gk, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_LoadById failed");
		goto error;
	}

	*aik = (TPA_AIK *) gk.objptr;

	if (gk.type != DB_AIK_TYPE || gk.objptr == NULL) {
		TRACE_MSG_BUG("gk.objptr must be deallocated");
		TRACE_MSG_ERROR("wrong retrieved key failed");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* Free memory */
	TpaLS_DB_Close(context, lstorage);
	TpaLS_LS_freeMemory(lstorage);
	return result;
}

/**
 * Deletes a ::TPA_AIK object
 *
 * @param context a pointer to a ::TPA_CONTEXT object
 * @param aik the ::TPA_AIK object to be deleted
 */
TPA_RESULT TpaHL_AIK_Delete(TPA_CONTEXT *context, TPA_AIK *aik) {
	/* TODO: implement TpaHL_AIK_Delete */
	return TPA_NOT_IMPLEMENTED;
}

/**
 * Deletes a ::TPA_AIK using its id
 *
 * @param context a pointer to a ::TPA_CONTEXT object
 * @param aik_id the id of the key to delete
 */
TPA_RESULT TpaHL_AIK_DeleteById(TPA_CONTEXT *context, TPA_OBJ_ID aik_id) {

	TPA_RESULT result = TPA_ERROR;
	TPA_LS *lstorage = NULL;
	int *idList = NULL;

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}

	if (aik_id <= 0) {
		TRACE_MSG_PARAM_ERROR("aik_id is not valid");
		goto error;
	}
#endif

	if (TpaLS_LS_allocate(&lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_LS_allocate failed");
		goto error;
	}

	/* Open Local Storage */
	if (TpaLS_DB_Open(context, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_Open failed");
		goto error;
	}

	if (TpaLS_DB_DeleteById(context, aik_id, DB_AIK_TYPE, lstorage)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_DeleteById failed");
		goto error;
	}

	result = TPA_SUCCESS;
	goto end;

	error: /* error label */
	end: /* end label */
	TpaLS_DB_Close(context, lstorage);
	TpaLS_LS_freeMemory(lstorage);
	if (idList)
		free(idList);

	return result;
}

TPA_RESULT TpaHL_AIK_Locate(TPA_CONTEXT *context, TPA_AIK *aik, TPA_PCA *ca) {

	TPA_LS *ls = NULL;
	TPA_RESULT result = TPA_ERROR;
	TPA_CERTIFICATE_ITEM *cert_list = NULL;

	/*
	 * Parameter control
	 */
#ifndef __NO_PARAM_CHECK
	if (context == NULL || aik == NULL || ca == NULL) {
		return result;
	}

	if (aik->label == NULL) {
		return result;
	}
#endif

	/* Set the default storage directory if it is not set */
	if (context->ls_uri == NULL) {
		if (TpaHL_CTX_set(context, CTX_LS_URI, strlen(LOCALSTORAGE_DIR),
				LOCALSTORAGE_DIR) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaHL_CTX_set failed");
			goto error;
		}
		context->ls_uri_size = strlen(LOCALSTORAGE_DIR);
	}

	//	if (LS_Init_new(context, &ls) != TPA_SUCCESS) {
	//		TRACE_MSG_ERROR("LS_Init_new failed");
	//		goto error;
	//	}

	if (Tpa_AIK_Locate(aik, ca, &cert_list) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_AIK_Locate failed");
		goto error;
	}

	/*
	 * TODO: Here code for store all the certificates in the local storage
	 */

	result = TPA_SUCCESS;

	error: if (!ls)
		TpaLS_LS_freeMemory(ls);
	if (!cert_list) /* TODO: create function */
	{
		TPA_CERTIFICATE_ITEM* item;
		TPA_CERTIFICATE_ITEM* temp;

		item = cert_list;
		while (item != NULL) {
			if (item->certificate != NULL) {
				free(item->certificate);
			}
			temp = item->next;
			free(item);
			item = temp;
		}
	}

	return result;
}

TPA_RESULT TpaHL_AIK_Revoke(TPA_CONTEXT *context, TPA_AIK *aik, TPA_PCA *ca) {
	/* TODO: implement TpaHL_AIK_Revoke */
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT TpaHL_AIK_Validate(TPA_AIK *aik, TPA_PCA *ca,
		TPA_CERTIFICATE_ITEM **cert_list, TPA_VERIFICATION_RESULT *res) {

	TPA_RESULT result = TPA_ERROR;

	/*
	 * Parameter control
	 */
#ifndef __NO_PARAM_CHECK
	if (aik == NULL || ca == NULL) {
		return result;
	}
#endif

	if (Tpa_AIK_Validate(aik, ca, cert_list, res) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_AIK_Validate failed");
		goto error;
	}

	result = TPA_SUCCESS;

	error:
	//	if (!(*cert_list)) /* TODO: create function */
	//	{
	//		TPA_CERTIFICATE_ITEM* item;
	//		TPA_CERTIFICATE_ITEM* temp;
	//
	//		item = *cert_list;
	//		while (item != NULL) {
	//			if (item->certificate != NULL) {
	//				free(item->certificate);
	//			}
	//			temp = item->next;
	//			free(item);
	//			item = temp;
	//		}
	//	}

	return result;

}

/*static int TpaHL_PCRSet_search(TPA_PCR_SET *pcrs, int index) {
 int i = 0;

 if (pcrs == NULL)
 return -1;

 for (i = 0; i < pcrs->index; i++)
 if (pcrs->pcrList[i].pcr == index)
 return i;
 return -1;
 }
 */

/****************************************************************************/
/*																			*/
/*							 		  PCR									*/
/*																			*/
/****************************************************************************/

/**
 * Reads a PCR and store its value into a ::TPA_PCR object.
 * Before reading, a ::TPA_PCR object SHALL be allocated and the PCR number set.
 *
 * @param pcr a pointer to the ::TPA_PCR object to read
 */
TPA_RESULT TpaHL_PCR_Read(TPA_PCR *pcr) {
	return Tpa_PCR_Read(pcr);
}

/**
 * Extends a PCR described by a ::TPA_PCR object.
 *
 * @param pcr a pointer to the ::TPA_PCR object to extend
 * @param blob a buffer containing the data to use for the extension
 * @param blob_size an integer containing the length of the blob
 */
TPA_RESULT TpaHL_PCR_Extend(TPA_PCR *pcr, unsigned char *blob,
		unsigned int blob_size) {
	return Tpa_PCR_Extend(pcr, blob, blob_size);
}

/****************************************************************************/
/*																			*/
/*							 		Sealing									*/
/*																			*/
/****************************************************************************/
TPA_RESULT TpaHL_SealBlob(TPA_CONTEXT *context, TPA_TPM *tpm,
		unsigned char *plainblob, unsigned int plainblob_size,
		TPA_PCR_SET *selectedPcrs, unsigned char **sealedblob,
		unsigned int *sealedblob_size) {

	TPA_RESULT result = TPA_ERROR;
	TPA_LS *lstorage = NULL;
	TPA_GENERIC_OBJPTR pMasterKey = { 0 };
	TPA_GENERIC_TPMKEY storageMK = { 0 };

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		goto error;
	}
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm is NULL");
		goto error;
	}
	if (plainblob == NULL || plainblob_size == 0) {
		TRACE_MSG_PARAM_ERROR("plainblob is NULL");
		goto error;
	}
	if (selectedPcrs == NULL) {
		TRACE_MSG_PARAM_ERROR("selectedPcrs is NULL");
		goto error;
	}
	if (sealedblob == NULL || sealedblob_size == NULL) {
		TRACE_MSG_PARAM_ERROR("sealedblob is NULL");
		goto error;
	}

#endif

	if (TpaLS_LS_allocate(&lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_LS_allocate failed");
		goto error;
	}

	/* Open DB*/
	if (TpaLS_DB_Open(context, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_Open failed");
		goto error;
	}

	/* Check for Storage Master Key into DB */
	if (TpaLS_DB_LoadStorageMK(context, &pMasterKey, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_LoadStorageMK failed");
		goto error;
	}

	if (pMasterKey.saved_objid == DB_KEY_NOT_FOUND) {
		//Create Storage Master Key
		TRACE_MSG_ERROR("Storage Master Key not found in DB");
		goto error;
	}

	storageMK.blob = ((TPA_GENERIC_TPMKEY *) pMasterKey.objptr)->blob;
	storageMK.blob_size = ((TPA_GENERIC_TPMKEY *) pMasterKey.objptr)->blob_size;

	/* Seal Blob */
	if (Tpa_SealBlob(tpm, plainblob, plainblob_size, selectedPcrs,
			storageMK.blob, storageMK.blob_size, sealedblob, sealedblob_size)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_SealBlob failed");
		goto error;
	}

	result = TPA_SUCCESS;

	error: /* Free memory */
	TpaLS_DB_Close(context, lstorage);
	TpaLS_LS_freeMemory(lstorage);
	if (pMasterKey.objptr)
		TpaHL_GenTpmKey_freeMemory((TPA_GENERIC_TPMKEY *) pMasterKey.objptr);

	return result;
}

/****************************************************************************/
/*																			*/
/*							 	   Unsealing								*/
/*																			*/
/****************************************************************************/
TPA_RESULT TpaHL_UnsealBlob(TPA_CONTEXT *context, TPA_TPM *tpm,
		unsigned char *sealedblob, unsigned int sealedblob_size,
		TPA_PCR_SET **selectedPcrs, unsigned char **unsealedblob,
		unsigned int *unsealedblob_size) {

	TPA_RESULT result = TPA_ERROR;
	TPA_LS *lstorage = NULL;
	TPA_GENERIC_OBJPTR pMasterKey = { 0 };
	TPA_GENERIC_TPMKEY storageMK = { 0 };

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		goto error;
	}
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm is NULL");
		goto error;
	}
	if (sealedblob == NULL || sealedblob_size == 0) {
		TRACE_MSG_PARAM_ERROR("sealedblob is NULL");
		goto error;
	}
	if (unsealedblob == NULL || unsealedblob_size == NULL) {
		TRACE_MSG_PARAM_ERROR("unsealedblob is NULL");
		goto error;
	}

#endif

	if (TpaLS_LS_allocate(&lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_LS_allocate failed");
		goto error;
	}

	/* Open DB*/
	if (TpaLS_DB_Open(context, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_Open failed");
		goto error;
	}

	/* Check for Storage Master Key into DB */
	if (TpaLS_DB_LoadStorageMK(context, &pMasterKey, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_LoadStorageMK failed");
		goto error;
	}

	if (pMasterKey.saved_objid == DB_KEY_NOT_FOUND) {
		TRACE_MSG_ERROR("Storage Master Key not found in DB");
		goto error;
	}

	storageMK.blob = ((TPA_GENERIC_TPMKEY *) pMasterKey.objptr)->blob;
	storageMK.blob_size = ((TPA_GENERIC_TPMKEY *) pMasterKey.objptr)->blob_size;

	/* Unseal Blob */
	if (Tpa_UnsealBlob(tpm, sealedblob, sealedblob_size, /*selectedPcrs,*/
	storageMK.blob, storageMK.blob_size, unsealedblob, unsealedblob_size)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_UnsealBlob failed");
		goto error;
	}

	result = TPA_SUCCESS;

	error: /* Free memory */
	TpaLS_DB_Close(context, lstorage);
	TpaLS_LS_freeMemory(lstorage);
	if (pMasterKey.objptr)
		TpaHL_GenTpmKey_freeMemory((TPA_GENERIC_TPMKEY *) pMasterKey.objptr);

	return result;
}

TPA_RESULT TpaHL_KEY_CreateTpmBindingKey(TPA_CONTEXT *context, TPA_TPM *tpm,
		TPA_TPMKEY *key) {

	TPA_RESULT result = TPA_ERROR;
	TPA_LS *lstorage = NULL;
	TPA_GENERIC_OBJPTR pMasterKey = { 0 };
	TPA_GENERIC_OBJPTR gk = { 0 };

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm is NULL");
		return TPA_ERROR;
	}
	if (key == NULL) {
		TRACE_MSG_PARAM_ERROR("skae is NULL");
		return TPA_ERROR;
	}
#endif

	if (TpaLS_LS_allocate(&lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_LS_allocate failed");
		goto error;
	}

	if (TpaLS_DB_Open(context, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_Open failed");
		goto error;
	}

	key->keyType = TPA_KEY_TYPE_BINDING;

	if (key->parentKey == NULL) {
		if (TpaLS_DB_LoadStorageMK(context, &pMasterKey, lstorage)
				!= TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaLS_DB_LoadStorageMK failed");
			goto error;
		}

		if (pMasterKey.saved_objid == DB_KEY_NOT_FOUND) {
			TRACE_MSG_ERROR("Storage Master Key not found in DB");
			goto error;
		}

		key->parentKey = (TPA_TPMKEY *) calloc(1, sizeof(TPA_TPMKEY));
		if (key->parentKey == NULL) {
			TRACE_MSG_ERROR("calloc failed");
			goto error;
		}
		key->parentKey->keyType = TPA_KEY_TYPE_SMK;
		key->parentKey->blob = ((TPA_GENERIC_TPMKEY *) pMasterKey.objptr)->blob;
		key->parentKey->blob_size
				= ((TPA_GENERIC_TPMKEY *) pMasterKey.objptr)->blob_size;
	} else {
		//Load from database key hierarchy
		return TPA_NOT_IMPLEMENTED;
	}

	if (Tpa_KEY_CreateTpmBindingKey(tpm, key) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_KEY_CreateTpmBindingKey failed");
		goto error;
	}

	gk.type = DB_BINDINGKEY_TYPE;
	gk.objptr = (TPA_OBJ_PTR) key;
	if (TpaLS_DB_Store(context, &gk, lstorage) != TPA_SUCCESS)
		goto error;

	if (TpaLS_DB_Close(context, lstorage) != TPA_SUCCESS)
		goto error;

	result = TPA_SUCCESS;
	error: if (pMasterKey.objptr)
		free(pMasterKey.objptr);
	return result;
}

TPA_RESULT TpaHL_KEY_ListTpmBindingKey(TPA_CONTEXT *context,
		TPA_TPMKEY **keyList, int *keyList_size, unsigned int flags) {

	TPA_RESULT result = TPA_ERROR;
	TPA_LS *lstorage = NULL;
	int *idList = NULL;
	int idList_size = 0;
	int i = 0;

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}

	if (keyList == NULL) {
		TRACE_MSG_PARAM_ERROR("keyList is NULL");
		goto error;
	}

	if (keyList_size == NULL) {
		TRACE_MSG_PARAM_ERROR("keyList_size is NULL");
		goto error;
	}
#endif

	if (flags != 0) {
		TRACE_MSG_INFO("flags is not valid");
		goto error;
	}

	if (TpaLS_LS_allocate(&lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_LS_allocate failed");
		goto error;
	}

	/* Open Local Storage */
	if (TpaLS_DB_Open(context, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_Open failed");
		goto error;
	}

	if (TpaLS_DB_List(context, lstorage, DB_BINDINGKEY_TYPE, &idList,
			&idList_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_List failed");
		goto error;
	}

	*keyList = (TPA_TPMKEY *) calloc(idList_size, sizeof(TPA_TPMKEY));
	if (*keyList == NULL) {
		TRACE_MSG_ERROR("calloc failed");
		goto error;
	}
	*keyList_size = idList_size;

	for (i = 0; i < idList_size; i++) {
		(*keyList)[i].tpmkey_id = idList[i];
	}

	result = TPA_SUCCESS;
	goto end;

	error: /* error label */
	if (*keyList) {
		free(*keyList);
		*keyList = NULL;
	}
	*keyList_size = 0;

	end: /* end label */
	TpaLS_DB_Close(context, lstorage);
	TpaLS_LS_freeMemory(lstorage);
	if (idList)
		free(idList);

	return result;
}

TPA_RESULT TpaHL_KEY_RetrieveTpmBindingKeyById(TPA_CONTEXT *context,
		TPA_TPMKEY **key, TPA_OBJ_ID id) {

	TPA_RESULT result = TPA_ERROR;
	TPA_LS *lstorage = NULL;
	TPA_GENERIC_OBJPTR gk = { 0 };

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}

	if (key == NULL) {
		TRACE_MSG_PARAM_ERROR("key is NULL");
		return TPA_ERROR;
	}

	if (id <= 0) {
		TRACE_MSG_PARAM_ERROR("id is not valid");
		return TPA_ERROR;
	}
#endif

	if (TpaLS_LS_allocate(&lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_LS_allocate failed");
		goto error;
	}

	/* Open Local Storage */
	if (TpaLS_DB_Open(context, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_Open failed");
		goto error;
	}

	/* Read SoftKey */
	gk.saved_objid = id;
	if (TpaLS_DB_LoadById(context, &gk, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_LoadById failed");
		goto error;
	}

	*key = (TPA_TPMKEY *) gk.objptr;

	if (gk.type != DB_BINDINGKEY_TYPE || gk.objptr == NULL) {
		TRACE_MSG_BUG("gk.objptr must be deallocated");
		TRACE_MSG_ERROR("wrong retrieved key failed");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* Free memory */
	TpaLS_DB_Close(context, lstorage);
	TpaLS_LS_freeMemory(lstorage);
	return result;
}

TPA_RESULT TpaHL_KEY_DeleteTpmBindingKeyById(TPA_CONTEXT *context,
		TPA_OBJ_ID key_id) {
	return TPA_NOT_IMPLEMENTED;
}

/****************************************************************************/
/*																			*/
/*							 	 Binding									*/
/*																			*/
/****************************************************************************/
TPA_RESULT TpaHL_DATA_BindBlob(TPA_CONTEXT *context, TPA_TPM *tpm,
		TPA_TPMKEY *key, unsigned char *plainblob, unsigned int plainblob_size,
		unsigned char **boundblob, unsigned int *boundblob_size) {

	TPA_RESULT result = TPA_ERROR;
	TPA_LS *lstorage = NULL;
	TPA_GENERIC_OBJPTR pMasterKey = { 0 };

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		goto error;
	}
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm is NULL");
		goto error;
	}
	if (key == NULL) {
		TRACE_MSG_PARAM_ERROR("key is NULL");
		goto error;
	}
	if (plainblob == NULL || plainblob_size == 0) {
		TRACE_MSG_PARAM_ERROR("plainblob is NULL");
		goto error;
	}
	if (boundblob == NULL || boundblob_size == NULL) {
		TRACE_MSG_PARAM_ERROR("boundblob is NULL");
		goto error;
	}
	if (key->keyType != TPA_KEY_TYPE_BINDING) {
		TRACE_MSG_PARAM_ERROR("key must be a TPA_KEY_TYPE_BINDING");
		goto error;
	}
#endif

	if (TpaLS_LS_allocate(&lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_LS_allocate failed");
		goto error;
	}

	/* Open DB*/
	if (TpaLS_DB_Open(context, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_Open failed");
		goto error;
	}

	if (key->parentKey == NULL) {
		if (TpaLS_DB_LoadStorageMK(context, &pMasterKey, lstorage)
				!= TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaLS_DB_LoadStorageMK failed");
			goto error;
		}

		if (pMasterKey.saved_objid == DB_KEY_NOT_FOUND) {
			TRACE_MSG_ERROR("Storage Master Key not found in DB");
			goto error;
		}

		key->parentKey = (TPA_TPMKEY *) calloc(1, sizeof(TPA_TPMKEY));
		if (key->parentKey == NULL) {
			TRACE_MSG_ERROR("calloc failed");
			goto error;
		}
		key->parentKey->keyType = TPA_KEY_TYPE_SMK;
		key->parentKey->blob = ((TPA_GENERIC_TPMKEY *) pMasterKey.objptr)->blob;
		key->parentKey->blob_size
				= ((TPA_GENERIC_TPMKEY *) pMasterKey.objptr)->blob_size;
	} else if (key->parentKey->keyType != TPA_KEY_TYPE_SMK) {
		//Load from database key hierarchy
		return TPA_NOT_IMPLEMENTED;
	}

	/* Bind Blob */
	if (Tpa_DATA_BindBlob(tpm, key, plainblob, plainblob_size, boundblob,
			boundblob_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_DATA_BindBlob failed");
		goto error;
	}

	result = TPA_SUCCESS;

	error: /* Free memory */
	TpaLS_DB_Close(context, lstorage);
	TpaLS_LS_freeMemory(lstorage);
	if (pMasterKey.objptr)
		TpaHL_GenTpmKey_freeMemory((TPA_GENERIC_TPMKEY *) pMasterKey.objptr);

	//	TRACE_MSG_BUG("FIX HERE: do we have to free the key here")
	//	if (key->parentKey)
	//		TpaHL_KEY_freeMemory(key->parentKey);
	//	key->parentKey = NULL;

	return result;
}

/****************************************************************************/
/*																			*/
/*							 	   Unbinding								*/
/*																			*/
/****************************************************************************/
TPA_RESULT TpaHL_DATA_UnbindBlob(TPA_CONTEXT *context, TPA_TPM *tpm,
		TPA_TPMKEY *key, unsigned char *boundblob, unsigned int boundblob_size,
		unsigned char **unboundblob, unsigned int *unboundblob_size) {

	TPA_RESULT result = TPA_ERROR;
	TPA_LS *lstorage = NULL;
	TPA_GENERIC_OBJPTR pMasterKey = { 0 };

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		goto error;
	}
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm is NULL");
		goto error;
	}
	if (boundblob == NULL || boundblob_size == 0) {
		TRACE_MSG_PARAM_ERROR("sealedblob is NULL");
		goto error;
	}
	if (unboundblob == NULL || unboundblob_size == NULL) {
		TRACE_MSG_PARAM_ERROR("unsealedblob is NULL");
		goto error;
	}
	if (key->keyType != TPA_KEY_TYPE_BINDING) {
		TRACE_MSG_PARAM_ERROR("key must be a TPA_KEY_TYPE_BINDING");
		goto error;
	}
#endif

	if (TpaLS_LS_allocate(&lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_LS_allocate failed");
		goto error;
	}

	/* Open DB*/
	if (TpaLS_DB_Open(context, lstorage) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaLS_DB_Open failed");
		goto error;
	}

	if (key->parentKey == NULL) {
		if (TpaLS_DB_LoadStorageMK(context, &pMasterKey, lstorage)
				!= TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaLS_DB_LoadStorageMK failed");
			goto error;
		}

		if (pMasterKey.saved_objid == DB_KEY_NOT_FOUND) {
			TRACE_MSG_ERROR("Storage Master Key not found in DB");
			goto error;
		}

		key->parentKey = (TPA_TPMKEY *) calloc(1, sizeof(TPA_TPMKEY));
		if (key->parentKey == NULL) {
			TRACE_MSG_ERROR("calloc failed");
			goto error;
		}
		key->parentKey->keyType = TPA_KEY_TYPE_SMK;
		key->parentKey->blob = ((TPA_GENERIC_TPMKEY *) pMasterKey.objptr)->blob;
		key->parentKey->blob_size
				= ((TPA_GENERIC_TPMKEY *) pMasterKey.objptr)->blob_size;
	} else if (key->parentKey->keyType != TPA_KEY_TYPE_SMK) {
		//Load from database key hierarchy
		return TPA_NOT_IMPLEMENTED;
	}

	/* Unseal Blob */
	if (Tpa_DATA_UnbindBlob(tpm, boundblob, boundblob_size, key, unboundblob,
			unboundblob_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_DATA_UnbindBlob failed");
		goto error;
	}

	result = TPA_SUCCESS;

	error: /* Free memory */
	TpaLS_DB_Close(context, lstorage);
	TpaLS_LS_freeMemory(lstorage);
	if (pMasterKey.objptr)
		TpaHL_GenTpmKey_freeMemory((TPA_GENERIC_TPMKEY *) pMasterKey.objptr);

	return result;
}

TPA_RESULT TpaHL_TPM_IsOwnershipTaken(TPA_CONTEXT *context, TPA_TPM *tpm,
		TPA_TPM_OWNERSHIP_RESULT *isTaken) {

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (context == NULL || tpm == NULL) {
		/* TODO: FIX ME more checks here */
		TRACE_MSG_PARAM_ERROR("default objects are needed");
		goto error;
	}

	if (isTaken == NULL) {
		/* TODO: FIX ME more checks here */
		TRACE_MSG_PARAM_ERROR("isTaken parameter is NULL");
		goto error;
	}

#endif

	if (Tpa_TPM_IsOwnershipTaken(tpm, isTaken) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TPM_IsOwnershipTaken failed");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* Free */
	return result;
}

TPA_RESULT TpaHL_TPM_TakeOwnership(TPA_CONTEXT *context, TPA_TPM *tpm) {

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (context == NULL || tpm == NULL) {
		/* TODO: FIX ME more checks here */
		TRACE_MSG_PARAM_ERROR("default objects are needed");
		goto error;
	}

	if (tpm->ownSecret == NULL || tpm->ownSecret_len == 0 || tpm->srkSecret
			== NULL || tpm->srkSecret_len == 0) {
		TRACE_MSG_PARAM_ERROR("set with default parameters");
		goto error;
	}

#endif

	if (Tpa_TPM_TakeOwnership(tpm) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TPM_TakeOwnership failed");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* Free */
	return result;
}

/*********************************************************************/
/* REMOTE ATTESTATION                                                */
/*********************************************************************/

/**
 * Performs consistency checks, builds the ::TPA_RA object and calls
 * Tpa_RA_Quote().
 */
TPA_RESULT TpaHL_RA_Quote(TPA_CONTEXT *context, TPA_TPM *tpm, TPA_AIK *aik,
		TPA_PCR_SET *selectedPcrs, unsigned char *raNonce,
		unsigned int raNonce_size, TPA_RA *raResult) {

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		goto error;
	}
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm is NULL");
		goto error;
	}
	if (raResult == NULL) {
		TRACE_MSG_PARAM_ERROR("raResult is NULL");
		goto error;
	}
	if (raNonce == NULL && raNonce_size != 0) {
		TRACE_MSG_PARAM_ERROR("raNonce is NULL but raNonce_size is not 0");
		goto error;
	}
	/*
	 * aik = NULL allowed: already in raResult
	 * selectedPcrs = NULL allowed: all
	 * raNonce = NULL allowed: generated by TSS
	 */
#endif

	/* raResult->aik = NULL: must be copied from aik
	 * In this case, aik can't be NULL.
	 */
	if (raResult->aik == NULL) {
		if (aik == NULL) {
			TRACE_MSG_PARAM_ERROR("aik is NULL: can't copy it in raResult");
			goto error;
		}

		if (TpaHL_AIK_RetrieveById(context, &(raResult->aik), aik->aik_id)
				!= TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaHL_AIK_RetrieveById failed");
			goto error;
		}

		/* copy AIK password to raResult->aik */
		if (aik->aikSecret != NULL) {
			if (TpaHL_AIK_set(raResult->aik, AIK_AIKSECRET, aik->aikSecret_len,
					aik->aikSecret) != TPA_SUCCESS) {
				TRACE_MSG_ERROR("TpaHL_AIK_RetrieveById failed");
				goto error;
			}
		}

	} else {
		//TODO
		// check ID and secret
		if (TpaHL_AIK_Retrieve(context, raResult->aik) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaHL_AIK_Retrieve failed");
			goto error;
		}
	}

	/* raResult->pcrs = NULL: copied from pcrs if available
	 * or retrieve all (currently only 15, i.e. TPMv1.1) PCRs
	 */
	if (raResult->pcrs == NULL) {
		int pindex = 0;
		int i = 0;
		//TODO: define a PCRSet_Copy
		//TODO: perhaps a "set index" is better... no need to read PCRs here
		if (selectedPcrs != NULL) {
			pindex = selectedPcrs->index;
			if (TpaHL_PCRSet_initialize(&raResult->pcrs, pindex, 20)
					!= TPA_SUCCESS) {
				TRACE_MSG_ERROR("TpaHL_PCRSet_initialize failed");
				goto error;
			}
			for (i = 0; i < pindex; i++) {
				if (TpaHL_PCRSet_pcr(raResult->pcrs,
						selectedPcrs->pcrList[i].pcr, NULL) != TPA_SUCCESS) {
					TRACE_MSG_ERROR("TpaHL_PCRSet_pcr failed");
					goto error;
				}
			}
		} else {
			//TODO: define macros for 15/24 and 20?
			pindex = 15;
			if (TpaHL_PCRSet_initialize(&raResult->pcrs, pindex, 20)
					!= TPA_SUCCESS) {
				TRACE_MSG_ERROR("TpaHL_PCRSet_initialize failed");
				goto error;
			}
			for (i = 0; i < pindex; i++) {
				if (TpaHL_PCRSet_pcr(raResult->pcrs, i, NULL) != TPA_SUCCESS) {
					TRACE_MSG_ERROR("TpaHL_PCRSet_pcr failed");
					goto error;
				}
			}
		}
	}

	/* raResult->raNonce = NULL: copied from raNonce if available
	 * if, in the end, raResult->raNonce is NULL, then it will be
	 * generated by the TSS.
	 */
	if (raResult->raNonce == NULL) {
		if (raNonce != NULL) {
			if (Tpa_BLOB_copy(raNonce, raNonce_size, &raResult->raNonce,
					&raResult->raNonce_size) != TPA_SUCCESS) {
				TRACE_MSG_ERROR("Tpa_BLOB_setValue failed");
				goto error;
			}
		}
	}

	/* invoke Tpa_RA_Quote() */
	if (Tpa_RA_Quote(tpm, raResult) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_RA_Quote failed");
		goto error;
	}

	/* attach measurements to the ::TPA_RA object */
	if (raResult->flags & TPA_RA_FLAG_MEASURES) {
		if (TpaHL_MeasureSet_RetrieveByPcrs(context, raResult->pcrs,
				&(raResult->measures)) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaHL_MeasureSet_RetrieveByPcrs failed");
			goto error;
		}
	}

	result = TPA_SUCCESS;
	error: /* Free */
	return result;
}

/**
 * Performs consistency checks, maybe fix the ::TPA_RA object (AIK) and
 * finally calls Tpa_RA_Verify().
 */
TPA_RESULT TpaHL_RA_Verify(TPA_CONTEXT *context, TPA_RA *ra, TPA_AIK **aik,
		unsigned char **raNonce, unsigned int *raNonce_size,
		TPA_PCR_SET **pcrs, TPA_MEASURE_SET **measures,
		TPA_VERIFICATION_RESULT *res) {

	TPA_RESULT result = TPA_ERROR;
	TPA_VERIFICATION_RESULT measureRes = TPA_VERIFICATION_NOK;

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		goto error;
	}
	if (ra == NULL) {
		TRACE_MSG_PARAM_ERROR("ra is NULL");
		goto error;
	}
	if (res == NULL) {
		TRACE_MSG_PARAM_ERROR("res is NULL");
		goto error;
	}
	if ((raNonce == NULL) ^ (raNonce_size == NULL)) {
		TRACE_MSG_PARAM_ERROR("raNonce and raNonce_size have different NULL-ness");
		goto error;
	}
#endif

	/* invoke Tpa_RA_Verify() */
	if (Tpa_RA_Verify(ra, res) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_RA_Verify failed");
		goto error;
	}

	/* copy AIK */
	if (aik != NULL) {
		if (*aik == NULL) {
			if (TpaHL_AIK_allocate(aik) != TPA_SUCCESS) {
				TRACE_MSG_ERROR("TpaHL_AIK_allocate failed");
				goto error;
			}
		}
		if (ra->aik->pubKey) {
			//TODO
		}
		if (ra->aik->cert) {
			(*aik)->cert_size = ra->aik->cert_size;
			(*aik)->cert = malloc((*aik)->cert_size);
			if ((*aik)->cert == NULL) {
				TRACE_MSG_ERROR("malloc failed (aik->cert)");
				goto error;
			}
			memcpy((*aik)->cert, ra->aik->cert, (*aik)->cert_size);
		}
		//TODO: PCA
	}

	/* copy nonce */
	if (raNonce != NULL) {
		*raNonce_size = ra->raNonce_size;
		if (*raNonce_size > 0) {
			*raNonce = malloc(*raNonce_size);
			if (*raNonce == NULL) {
				TRACE_MSG_ERROR("malloc failed (raNonce)");
				goto error;
			}
			memcpy(*raNonce, ra->raNonce, *raNonce_size);
		} else {
			*raNonce = NULL;
		}
	}

	/* copy PCRs */
	if (pcrs != NULL) {
		int pindex = ra->pcrs->index;
		int i = 0;
		//TODO: define a PCRSet_Copy
		//TODO: check if pcrs if already initialized, but with smaller size than pindex
		if (*pcrs == NULL) {
			if (TpaHL_PCRSet_initialize(pcrs, pindex, 20) != TPA_SUCCESS) {
				TRACE_MSG_ERROR("TpaHL_PCRSet_initialize failed");
				goto error;
			}
		}
		for (i = 0; i < pindex; i++) {
			if (TpaHL_PCRSet_pcr(*pcrs, ra->pcrs->pcrList[i].pcr,
					ra->pcrs->pcrList[i].value) != TPA_SUCCESS) {
				TRACE_MSG_ERROR("TpaHL_PCRSet_pcr failed");
				goto error;
			}
		}
	}

	/* if res is not TPA_VERIFICATION_OK, we already know RA verification failed,
	 * and we don't need to go through the measure set.
	 * Note the function TpaHL_RA_Verify() returns success.
	 */
	if (*res != TPA_VERIFICATION_OK)
		return TPA_SUCCESS;

	/* check measures */
	if ((measures != NULL) && (ra->flags & TPA_RA_FLAG_MEASURES)) {
		if (Tpa_MeasureSet_VerifyPcrs(ra->measures, *pcrs, measures,
				&measureRes) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("Tpa_MeasureSet_VerifyPcrs failed");
			goto error;
		}

		/* adjust res */
		if (measureRes != TPA_VERIFICATION_OK) {
			*res = measureRes;
		}
	}

	result = TPA_SUCCESS;
	error: /* Free */
	return result;
}

/*********************************************************************/
/* MEASURE                                                           */
/*********************************************************************/

TPA_RESULT TpaHL_Measure_TargetPCR(TPA_CONTEXT *context, TPA_PCR_NUM pcr) {
#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}
	if (context->measure_mod == NULL) {
		TRACE_MSG_PARAM_ERROR("context->measure_mod is NULL");
		return TPA_ERROR;
	}
	if (context->measure_mod->targetPcr == NULL) {
		TRACE_MSG_PARAM_ERROR("context->measure_mod->targetPcr is NULL");
		return TPA_ERROR;
	}
	if (pcr < TPA_PCR_NUM_MIN || pcr > TPA_PCR_NUM_MAX) {
		TRACE_MSG_PARAM_ERROR("pcr is not a valid PCR number");
		return TPA_ERROR;
	}
#endif
	context->measure_mod->targetPcr->pcr = pcr;
	return TpaHL_PCR_Read(context->measure_mod->targetPcr);
}

TPA_RESULT TpaHL_Measure_dlopen(TPA_CONTEXT *context, const char *filename,
		int flag, void **handle) {

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}
	if (context->measure_mod == NULL) {
		TRACE_MSG_PARAM_ERROR("context->measure_mod is NULL");
		return TPA_ERROR;
	}
	if (filename == NULL) {
		TRACE_MSG_PARAM_ERROR("filename is NULL");
		return TPA_ERROR;
	}
	if (handle == NULL) {
		TRACE_MSG_PARAM_ERROR("handle is NULL");
		return TPA_ERROR;
	}
#endif
	context->measure_mod->preprocessing_callback = NULL;
	context->measure_mod->postprocessing_callback = Tpa_Measure_Callback_dlopen;
	context->measure_mod->ctx_data = (void *) &flag;
	*handle = NULL;

	if (Tpa_Measure_File(context->measure_mod, TPA_MEASURE_TYPE_LIBRARY,
			filename, NULL, 0, NULL) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_Measure_File failed");
		goto error;
	}

	/* return dlopen's handle */
	*handle = context->measure_mod->ctx_data;

	result = TPA_SUCCESS;

	error: /* free */
	/* cleanup callbacks and ctx_data */
	context->measure_mod->postprocessing_callback = NULL;
	context->measure_mod->ctx_data = NULL;
	return result;
}

TPA_RESULT TpaHL_Measure_fopen(TPA_CONTEXT *context, const char *path,
		const char *mode, FILE **fp) {
	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}
	if (context->measure_mod == NULL) {
		TRACE_MSG_PARAM_ERROR("context->measure_mod is NULL");
		return TPA_ERROR;
	}
	if (path == NULL) {
		TRACE_MSG_PARAM_ERROR("path is NULL");
		return TPA_ERROR;
	}
	if (fp == NULL) {
		TRACE_MSG_PARAM_ERROR("fp is NULL");
		return TPA_ERROR;
	}
#endif
	context->measure_mod->preprocessing_callback = NULL;
	context->measure_mod->postprocessing_callback = Tpa_Measure_Callback_fopen;
	context->measure_mod->ctx_data = (void *) mode;
	*fp = NULL;

	if (Tpa_Measure_File(context->measure_mod, TPA_MEASURE_TYPE_FILE,
			path, NULL, 0, NULL) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_Measure_File failed");
		goto error;
	}

	/* return dlopen's handle */
	*fp = context->measure_mod->ctx_data;

	result = TPA_SUCCESS;

	error: /* free */
	/* cleanup callbacks and ctx_data */
	context->measure_mod->postprocessing_callback = NULL;
	context->measure_mod->ctx_data = NULL;
	return result;
}

TPA_RESULT TpaHL_Measure_File(TPA_CONTEXT *context, const char *path,
		char *label, unsigned int label_len, TPA_MEASURE **measure) {

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}
	if (context->measure_mod == NULL) {
		TRACE_MSG_PARAM_ERROR("context->measure_mod is NULL");
		return TPA_ERROR;
	}
	if (path == NULL) {
		TRACE_MSG_PARAM_ERROR("path is NULL");
		return TPA_ERROR;
	}
	if (label == NULL && label_len > 0) {
		TRACE_MSG_PARAM_ERROR("label_len > 0 but label is NULL");
		return TPA_ERROR;
	}
#endif
	context->measure_mod->preprocessing_callback = NULL;
	context->measure_mod->postprocessing_callback = NULL;

	return Tpa_Measure_File(context->measure_mod, TPA_MEASURE_TYPE_FILE, path,
			label, label_len, measure);
}

TPA_RESULT TpaHL_Measure_Blob(TPA_CONTEXT *context, unsigned char *blob,
		unsigned int blob_size, char *label, unsigned int label_len,
		TPA_MEASURE **measure) {

#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}
	if (context->measure_mod == NULL) {
		TRACE_MSG_PARAM_ERROR("context->measure_mod is NULL");
		return TPA_ERROR;
	}
	if (blob == NULL && blob_size > 0) {
		TRACE_MSG_PARAM_ERROR("blob_size > 0 but blob is NULL");
		return TPA_ERROR;
	}
	if (label == NULL && label_len > 0) {
		TRACE_MSG_PARAM_ERROR("label_len > 0 but label is NULL");
		return TPA_ERROR;
	}
#endif
	context->measure_mod->preprocessing_callback = NULL;
	context->measure_mod->postprocessing_callback = NULL;

	return Tpa_Measure_Blob(context->measure_mod, TPA_MEASURE_TYPE_GENERIC,
			blob, blob_size, label, label_len, measure);
}

TPA_RESULT TpaHL_Measure_DeleteAll(TPA_CONTEXT *context) {
#ifndef __NO_PARAM_CHECK
	if (context == NULL) {
		TRACE_MSG_PARAM_ERROR("context is NULL");
		return TPA_ERROR;
	}
	if (context->measure_mod == NULL) {
		TRACE_MSG_PARAM_ERROR("context->measure_mod is NULL");
		return TPA_ERROR;
	}
#endif

	return Tpa_Measure_DeleteAll(context->measure_mod);
}

/*********************************************************************/
/* MEASURE SET                                                       */
/*********************************************************************/

TPA_RESULT TpaHL_MeasureSet_RetrieveByPcrs(TPA_CONTEXT *context,
		TPA_PCR_SET *selectedPcrs, TPA_MEASURE_SET **measures) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT TpaHL_MeasureSet_Serialize(TPA_MEASURE_SET *measures,
		unsigned char **blob, unsigned int *blob_size) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT TpaHL_MeasureSet_Deserialize(unsigned char *blob,
		unsigned int blob_size, unsigned char **blob_end,
		TPA_MEASURE_SET **measures) {
	return TPA_NOT_IMPLEMENTED;
}
