/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include "TPA_API.h"
#include "TPA_Core.h"
#include "TPA_Utils.h"

#include <stdlib.h>
#include <string.h>

/****************************************************************************/
/*																			*/
/*							 Context Helper									*/
/*																			*/
/****************************************************************************/

/**
 * Allocates and initializes a ::TPA_CONTEXT object
 *
 * @param ctx a double pointer to a ::TPA_CONTEXT object
 */
TPA_RESULT TpaHL_CTX_allocate(TPA_CONTEXT **ctx) {

#ifndef __NO_PARAM_CHECK
	if (ctx == NULL) {
		return TPA_ERROR;
	}
#endif

	*ctx = (TPA_CONTEXT *) malloc(sizeof(TPA_CONTEXT));
	if (*ctx == NULL) {
		return TPA_ERROR;
	}

	(*ctx)->ls_uri = NULL;
	(*ctx)->ls_uri_size = 0;

	/* allocate common local storage */
	if (Tpa_LS_allocate(&(*ctx)->ls) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_LS_allocate failed");
		goto error;
	}

	/* allocate KEY module */
	if (Tpa_KEY_allocateModule(&(*ctx)->key_mod, (*ctx)->ls) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_KEY_allocateModule failed");
		goto error;
	}

	/* allocate MEASURE module, sharing the local storage */
	if (Tpa_Measure_allocateModule(&(*ctx)->measure_mod, (*ctx)->ls)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_Measure_allocateModule failed");
		goto error;
	}

	/* set default */
	(*ctx)->ls->driver_lib = strdup("libtpa-sqlite.so");
	(*ctx)->ls->path = strdup(TPA_DEFAULT_LOCAL_STORAGE);
	(*ctx)->ls->filename = strdup(TPA_DEFAULT_LOCAL_STORAGE_FILE);
	(*ctx)->ls->complete_path = calloc(1, strlen((*ctx)->ls->path) + strlen(
			(*ctx)->ls->filename) + 2);
	sprintf((*ctx)->ls->complete_path, "%s/%s", (*ctx)->ls->path,
			(*ctx)->ls->filename);

	/* hack */
	(*ctx)->ls_uri = (*ctx)->ls->path;
	(*ctx)->ls_uri_size = strlen((*ctx)->ls_uri);

	return TPA_SUCCESS;
	error: /* error */
	if (*ctx != NULL) {
		TpaHL_CTX_freeMemory(*ctx);
		*ctx = NULL;
	}
	return TPA_ERROR;
}

/**
 * Assigns a member of a ::TPA_CONTEXT object
 *
 * @param ctx a pointer to the ::TPA_CONTEXT object
 * @param idx a flag indicating the member to assign. It can be: CTX_LS_URI
 * @param len size of the data to be assigned to the member
 * @param data a pointer to the memory area that contains the data to be assigned to the member
 */
TPA_RESULT TpaHL_CTX_set(TPA_CONTEXT *ctx, int idx, long len, void * data) {

	TPA_RESULT res = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (ctx == NULL) {
		return TPA_ERROR;
	}
	if (idx < 0) {
		return TPA_ERROR;
	}
	if (len <= 0) {
		return TPA_ERROR;
	}
	if (data == NULL) {
		return TPA_ERROR;
	}
#endif

	switch (idx) {
	case CTX_LS_URI: /* TODO rename in PATH */
		if (ctx->ls != NULL) {
			if (ctx->ls->path != NULL)
				free(ctx->ls->path);
			if (ctx->ls->complete_path != NULL)
				free(ctx->ls->complete_path);

			res = Tpa_setString((void *) &(ctx->ls->path), len, data);

			ctx->ls->complete_path = calloc(1, strlen(ctx->ls->path)
					+ strlen(ctx->ls->filename) + 2);
			sprintf(ctx->ls->complete_path, "%s/%s", ctx->ls->path,
					ctx->ls->filename);

			/* hack */
			ctx->ls_uri = ctx->ls->path;
			ctx->ls_uri_size = strlen(ctx->ls_uri);
		}
		return res;
		break;
	default:
		return TPA_ERROR;
	}

	return TPA_ERROR;
}

/**
 * Frees a ::TPA_CONTEXT object
 *
 * @param ctx a pointer to a ::TPA_CONTEXT object
 */
void TpaHL_CTX_freeMemory(TPA_CONTEXT *ctx) {

	if (!ctx)
		return;

	/* prevent multi-free is local storage is shared */
	if (ctx->ls != NULL) {
		ctx->key_mod->ls = NULL;
		ctx->measure_mod->ls = NULL;
	}

	Tpa_Measure_freeModuleMemory(ctx->measure_mod);
	ctx->measure_mod = NULL;

	Tpa_KEY_freeModuleMemory(ctx->key_mod);
	ctx->key_mod = NULL;

	Tpa_LS_freeMemory(ctx->ls);
	ctx->ls = NULL;

	free(ctx);
}

/****************************************************************************/
/*																			*/
/*								TPM Helper									*/
/*																			*/
/****************************************************************************/

/**
 * Allocates and initializes a ::TPA_TPM object
 *
 * @param tpm a double pointer to a ::TPA_TPM object
 */
TPA_RESULT TpaHL_TPM_allocate(TPA_TPM **tpm) {
#ifndef __NO_PARAM_CHECK
	if (tpm == NULL) {
		return TPA_ERROR;
	}
#endif

	*tpm = (TPA_TPM *) malloc(sizeof(TPA_TPM));
	if (*tpm == NULL) {
		return TPA_ERROR;
	}

	(*tpm)->ownSecret = NULL;
	(*tpm)->srkSecret = NULL;
	(*tpm)->ekCertificate = NULL;
	(*tpm)->ekCertificateFilename = NULL;

	(*tpm)->ownSecret_len = 0;
	(*tpm)->srkSecret_len = 0;
	(*tpm)->ekCertificate_size = 0;

	return TPA_SUCCESS;
}

/**
 * Assigns a member of a ::TPA_TPM object
 *
 * @param tpm a pointer to the ::TPA_TPM object
 * @param idx a flag indicating the member to assign. It can be: TPM_OWNERPWD, TPM_SRKPWD, TPM_EKCERT_FILENAME, TPM_EKCERT,
 * @param len size of the data to be assigned to the member
 * @param data a pointer to the memory area that contains the data to be assigned to the member
 */
TPA_RESULT TpaHL_TPM_set(TPA_TPM *tpm, int idx, long len, void *data) {
	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (tpm == NULL) {
		return TPA_ERROR;
	}
	if (idx < 0) {
		return TPA_ERROR;
	}
	if (len <= 0) {
		return TPA_ERROR;
	}
	if (data == NULL) {
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	switch (idx) {
	case TPM_OWNERPWD:
		tpm->ownSecret_len = len;
		result = Tpa_setString((void *) &(tpm->ownSecret), len, data);
		return result;

	case TPM_SRKPWD:
		tpm->srkSecret_len = len;
		result = Tpa_setString((void *) &(tpm->srkSecret), len, data);
		return result;

	case TPM_EKCERT_FILENAME:
		result = Tpa_setString((void *) &(tpm->ekCertificateFilename), len,
				data);
		return result;

	case TPM_EKCERT:
		tpm->ekCertificate_size = len;
		result = Tpa_setMember((void *) &(tpm->ekCertificate), len, data);
		return result;

	default:
		return result;
	}

	/*
	 * If it reaches here, something went wrong
	 */
	return TPA_ERROR;
}

/**
 * Frees a ::TPA_TPM object
 *
 * @param tpm a pointer to a ::TPA_TPM object
 */
void TpaHL_TPM_freeMemory(TPA_TPM *tpm) {

	if (!tpm)
		return;

	if (tpm->ekCertificate)
		free(tpm->ekCertificate);

	if (tpm->ekCertificateFilename)
		free(tpm->ekCertificateFilename);

	if (tpm->ownSecret)
		free(tpm->ownSecret);

	if (tpm->srkSecret)
		free(tpm->srkSecret);

	free(tpm);
}

/****************************************************************************/
/*																			*/
/*							 Generic TPM KEY								*/
/*																			*/
/****************************************************************************/

/**
 * Allocates and initializes a ::TPA_GENERIC_TPMKEY object
 *
 * ::TPA_GENERIC_TPMKEY contains a blob which represents a key that can be used
 * by a TPM. Such key can be any node of the tree of keys managed by the TPM.
 *
 * @param gTpmKey a double pointer to a ::TPA_GENERIC_TPMKEY object
 */
TPA_RESULT TpaHL_GenTpmKey_allocate(TPA_GENERIC_TPMKEY **gTpmKey) {

#ifndef __NO_PARAM_CHECK
	if (gTpmKey == NULL) {
		return TPA_ERROR;
	}
#endif

	*gTpmKey = (TPA_GENERIC_TPMKEY *) malloc(sizeof(TPA_GENERIC_TPMKEY));
	if (*gTpmKey == NULL) {
		return TPA_ERROR;
	}

	(*gTpmKey)->blob = NULL;
	(*gTpmKey)->blob_size = 0;
	(*gTpmKey)->tpmkey_id = -1;

	return TPA_SUCCESS;
}

/**
 * Assigns a member of a ::TPA_GENERIC_TPMKEY object. Currently this is a dummy function
 *
 * @param gTpmKey a pointer to the ::TPA_GENERIC_TPMKEY object
 * @param idx a flag indicating the member to assign
 * @param len size of the data to be assigned to the member
 * @param data a pointer to the memory area that contains the data to be assigned to the member
 */
TPA_RESULT TpaHL_GenTpmKey_set(TPA_GENERIC_TPMKEY *gTpmKey, int idx, long len,
		void *data) {
	/* TODO: implement TpaHL_GenTpmKey_set */
	return TPA_NOT_IMPLEMENTED;
}

/**
 * Frees a ::TPA_GENERIC_TPMKEY object
 *
 * @param gTpmKey a pointer to a ::TPA_GENERIC_TPMKEY object
 */
void TpaHL_GenTpmKey_freeMemory(TPA_GENERIC_TPMKEY *gTpmKey) {

	if (!gTpmKey)
		return;

	if (gTpmKey->blob)
		free(gTpmKey->blob);

	gTpmKey->blob = NULL;
	gTpmKey->blob_size = 0;
	gTpmKey->tpmkey_id = -1;

	free(gTpmKey);
}

/****************************************************************************/
/*																			*/
/*							    PCA Helper									*/
/*																			*/
/****************************************************************************/
/**
 * Allocates and initializes a ::TPA_CONTEXT object
 *
 * @param pca a double pointer to a ::TPA_CONTEXT object
 */
TPA_RESULT TpaHL_PCA_allocate(TPA_PCA **pca) {
#ifndef __NO_PARAM_CHECK
	/*
	 * Parameter check
	 */
	if (pca == NULL) {
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	*pca = (TPA_PCA *) malloc(sizeof(TPA_PCA));
	if (*pca == NULL) {
		return TPA_ERROR;
	}

	(*pca)->uri = NULL;
	//	(*pca)->auth = NULL;
	(*pca)->pcaCertificateFilename = NULL;
	(*pca)->pcaCertificate = NULL;

	(*pca)->type = 0;
	(*pca)->port = 0;
	(*pca)->pcaCertificate_size = 0;

	return TPA_SUCCESS;
}

/**
 * Assigns a member of a ::TPA_PCA object
 *
 * @param pca a pointer to the ::TPA_PCA object
 * @param idx a flag indicating the member to assign. It can be: PCA_URI, PCA_AUTH, PCA_PCACERT, PCA_PCACERT_FILENAME, PCA_PORT
 * @param len size of the data to be assigned to the member
 * @param data a pointer to the memory area that contains the data to be assigned to the member
 */
TPA_RESULT TpaHL_PCA_set(TPA_PCA *pca, int idx, long len, void *data) {
	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (pca == NULL) {
		return TPA_ERROR;
	}
	if (idx < 0) {
		return TPA_ERROR;
	}
	if (len <= 0) {
		return TPA_ERROR;
	}
	if (data == NULL) {
		return TPA_ERROR;
	}
#endif

	switch (idx) {
	case PCA_URI:
		result = Tpa_setString((void *) &(pca->uri), len, data);
		return result;

		//	case PCA_AUTH:
		//		result = Tpa_setString((void *) &(pca->auth), len, data);
		//		return result;

	case PCA_PCACERT:
		pca->pcaCertificate_size = len;
		result = Tpa_setMember((void *) &(pca->pcaCertificate), len, data);
		return result;

	case PCA_PCACERT_FILENAME:
		result = Tpa_setString((void *) &(pca->pcaCertificateFilename), len,
				data);
		return result;

	case PCA_PORT:
		if (pca) {
			pca->port = *((int*) data);
			result = TPA_SUCCESS;
		}
		return result;

	default:
		return result;
	}

	/*
	 * If it reaches here, something went wrong
	 */
	return TPA_ERROR;
}

/**
 * Frees a ::TPA_PCA context
 *
 * @param pca a pointer to a ::TPA_PCA object
 */
void TpaHL_PCA_freeMemory(TPA_PCA *pca) {

	if (!pca)
		return;

	if (pca->uri)
		free(pca->uri);

	//	if (pca->auth)
	//		free(pca->auth);

	if (pca->pcaCertificateFilename)
		free(pca->pcaCertificateFilename);

	if (pca->pcaCertificate)
		free(pca->pcaCertificate);

	free(pca);
}

/****************************************************************************/
/*																			*/
/*   	     		 Attestation Identity Key (AIK)							*/
/*																			*/
/****************************************************************************/

/**
 * Allocates and initializes a ::TPA_AIK object
 *
 * @param aik a double pointer to a ::TPA_AIK object
 */
TPA_RESULT TpaHL_AIK_allocate(TPA_AIK **aik) {

#ifndef __NO_PARAM_CHECK
	if (aik == NULL) {
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	*aik = (TPA_AIK *) malloc(sizeof(TPA_AIK));
	if (*aik == NULL) {
		return TPA_ERROR;
	}

	(*aik)->pubKey = NULL;
	(*aik)->privKey = NULL;
	(*aik)->cert = NULL;
	(*aik)->aikSecret = NULL;
	(*aik)->label = NULL;
	(*aik)->uuid = NULL;
	(*aik)->keyBlob = NULL;
	(*aik)->labelUTF16 = NULL;

	(*aik)->aikSecret_len = 0;
	(*aik)->pubKey_size = 0;
	(*aik)->privKey_size = 0;
	(*aik)->cert_size = 0;
	(*aik)->keyBlob_size = 0;
	(*aik)->label_len = 0;
	(*aik)->labelUTF16_len = 0;

	(*aik)->aik_id = -1;

	return TPA_SUCCESS;
}

static char* convert_ASCII_to_UTF_16LE(char* secret, int secret_length,
		int* new_length) {

	int conv_length = secret_length * 2;
	char* new_secret = calloc(conv_length, sizeof(char));
	if (new_secret == NULL) {
		return NULL;
	}

	int idx = 0;
	int i = 0;
	while (idx < conv_length - 1) {
		new_secret[idx] = secret[i];
		idx++;
		new_secret[idx] = 0;
		idx++;
		i++;
	}
	*new_length = conv_length;

	return new_secret;
}

/**
 * Assigns a member of a ::TPA_AIK object
 *
 * @param aik a pointer to the ::TPA_AIK object
 * @param idx a flag indicating the member to assign. It can be: AIK_PUBKEY, AIK_PRIVKEY, AIK_CERT, AIK_LABEL, AIK_UUID, AIK_AIKSECRET
 * @param len size of the data to be assigned to the member
 * @param data a pointer to the memory area that contains the data to be assigned to the member
 */
TPA_RESULT TpaHL_AIK_set(TPA_AIK *aik, int idx, long len, void * data) {
#ifndef __NO_PARAM_CHECK
	/*
	 * Parameter check
	 */
	if (aik == NULL) {
		return TPA_ERROR;
	}
	if (idx < 0) {
		return TPA_ERROR;
	}
	if (len <= 0) {
		return TPA_ERROR;
	}
	if (data == NULL) {
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	switch (idx) {
	case AIK_PUBKEY:
		aik->pubKey_size = len;
		return Tpa_setMember((void *) &(aik->pubKey), len, data);

	case AIK_PRIVKEY:
		aik->privKey_size = len;
		return Tpa_setMember((void *) &(aik->privKey), len, data);

	case AIK_CERT:
		aik->cert_size = len;
		return Tpa_setMember((void *) &(aik->cert), len, data);

	case AIK_LABEL:
		aik->labelUTF16 = convert_ASCII_to_UTF_16LE(data, strlen(data),
				(int *) &(aik->labelUTF16_len));
		if (aik->labelUTF16 == NULL) {
			return TPA_ERROR;
		}
		aik->label_len = len;
		return Tpa_setString((void *) &(aik->label), len, data);

	case AIK_UUID:
		return Tpa_setString((void *) &(aik->uuid), len, data);

	case AIK_AIKSECRET:
		aik->aikSecret_len = len;
		return Tpa_setString((void *) &(aik->aikSecret), len, data);

	default:
		return TPA_ERROR;
	}

	/*
	 * If it reaches here, something went wrong
	 */
	return TPA_ERROR;
}

/**
 * Frees a ::TPA_AIK object
 *
 * @param aik a pointer to a ::TPA_AIK object
 */
void TpaHL_AIK_freeMemory(TPA_AIK *aik) {

	if (aik == NULL) {
		return;
	}
	aik->aik_id = -1;
	if (aik->aikSecret != NULL) {
		free(aik->aikSecret);
		aik->aikSecret = NULL;
	}
	aik->aikSecret_len = 0;
	if (aik->label != NULL) {
		free(aik->label);
		aik->label = NULL;
	}
	aik->label_len = 0;
	if (aik->labelUTF16 != NULL) {
		free(aik->labelUTF16);
		aik->labelUTF16 = NULL;
	}
	aik->labelUTF16_len = 0;
	if (aik->uuid != NULL) {
		free(aik->uuid);
		aik->uuid = NULL;
	}
	if (aik->cert != NULL) {
		free(aik->cert);
		aik->cert = NULL;
	}
	aik->cert_size = 0;
	if (aik->pubKey != NULL) {
		free(aik->pubKey);
		aik->pubKey = NULL;
	}
	aik->pubKey_size = 0;
	if (aik->privKey != NULL) {
		free(aik->privKey);
		aik->privKey = NULL;
	}
	aik->privKey_size = 0;
	if (aik->keyBlob != NULL) {
		free(aik->keyBlob);
		aik->keyBlob = NULL;
	}
	aik->keyBlob_size = 0;
	free(aik);
	return;
}

/****************************************************************************/
/*																			*/
/*							 	PCR Helper									*/
/*																			*/
/****************************************************************************/
/**
 * Allocates and initializes a ::TPA_PCR object. Currently this is a dummy function
 *
 * @param pcr a double pointer to a ::TPA_PCR object
 */
TPA_RESULT TpaHL_PCR_allocate(TPA_PCR **pcr) {
#ifndef __NO_PARAM_CHECK
	/*
	 * Parameter check
	 */
	if (pcr == NULL) {
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	*pcr = (TPA_PCR *) malloc(sizeof(TPA_PCR));
	if (*pcr == NULL) {
		return TPA_ERROR;
	}

	(*pcr)->pcr = 0;
	(*pcr)->pcr_size = 20;
	(*pcr)->value = NULL;

	return TPA_SUCCESS;
}

/**
 * Assigns a member of a ::TPA_PCR object. Currently this is a dummy function
 *
 * @param pcr a pointer to the ::TPA_PCR object
 * @param idx a flag indicating the member to assign
 * @param len size of the data to be assigned to the member
 * @param data a pointer to the memory area that contains the data to be assigned to the member
 */
TPA_RESULT TpaHL_PCR_set(TPA_PCR *pcr, int idx, long len, void *data) {
	/* TODO: implement TpaHL_PCR_set */
	return TPA_NOT_IMPLEMENTED;
}

/**
 * Frees a ::TPA_PCR object. Currently this is a dummy function
 *
 * @param pcr a pointer to a ::TPA_PCR object
 */
void TpaHL_PCR_freeMemory(TPA_PCR *pcr) {
	/* TODO: implement TpaHL_PCR_freeMemory */
}

/****************************************************************************/
/*																			*/
/*							 PCRSet Helper									*/
/*																			*/
/****************************************************************************/

/**
 * Allocates and initializes a ::TPA_PCR_SET object
 *
 * @param pcrSet a double pointer to a ::TPA_PCR_SET object
 */
TPA_RESULT TpaHL_PCRSet_allocate(TPA_PCR_SET **pcrSet) {

#ifndef __NO_PARAM_CHECK
	/*
	 * Parameter check
	 */
	if (pcrSet == NULL) {
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	*pcrSet = (TPA_PCR_SET *) malloc(sizeof(TPA_PCR_SET));
	if (*pcrSet == NULL) {
		return TPA_ERROR;
	}

	(*pcrSet)->pcrList = NULL;
	(*pcrSet)->index = 0;
	(*pcrSet)->pcrList_size = 0;
	(*pcrSet)->pcr_size = 0;

	return TPA_SUCCESS;
}

/**
 * Frees a ::TPA_PCR_SET object
 *
 * @param pcrSet a pointer to a ::TPA_PCR_SET object
 */
void TpaHL_PCRSet_freeMemory(TPA_PCR_SET *pcrSet) {
	int i = 0;

	if (pcrSet == NULL)
		return;

	for (i = 0; i < pcrSet->index; i++)
		Tpa_PCR_freeMemory(&pcrSet->pcrList[i], 0);

	memset(pcrSet->pcrList, 0, sizeof(TPA_PCR) * pcrSet->pcrList_size);
	pcrSet->index = 0;
	pcrSet->pcrList_size = 0;
	free(pcrSet->pcrList);
	pcrSet->pcrList = NULL;

	free(pcrSet);
}

/**
 * Initializes a ::TPA_PCR_SET object
 *
 * @param pcrSet a double pointer to a ::TPA_PCR_SET object
 * @param setDim number of PCRs. In a normal TPM 1.2 this is set to 24
 * @param pcrSize size in bytes of each PCR. In a normal TPM 1.2 this is set to 20
 */
TPA_RESULT TpaHL_PCRSet_initialize(TPA_PCR_SET **pcrSet, int setDim,
		int pcrSize) {

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (pcrSet == NULL) {
		TRACE_MSG_PARAM_ERROR("pcrSet cannot be NULL");
		goto error;
	}
	if (setDim < 1) {
		TRACE_MSG_PARAM_ERROR("setDim cannot be zero");
		goto error;
	}
#endif

	if (*pcrSet == NULL) {
		if (TpaHL_PCRSet_allocate(pcrSet) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaHL_PCRSet_allocate failed");
			goto error;
		}

	}

	(*pcrSet)->pcrList_size = setDim;

	(*pcrSet)->pcrList = (TPA_PCR *) calloc((*pcrSet)->pcrList_size,
			sizeof(TPA_PCR));

	(*pcrSet)->pcr_size = pcrSize;

	if ((*pcrSet)->pcrList == NULL) {
		TRACE_MSG_ERROR("calloc failed");
		return result;
	}

	result = TPA_SUCCESS;
	error: return result;
}

/**
 * Sets a member of a ::TPA_PCR_SET object
 *
 * This funcion sets a value into one of the PCRs that compose a ::TPA_PCR_SET object. Note that these PCR are not the
 * physical PCRs of the TPM, but only memory areas that belong to the ::TPA_PCR_SET object.
 * If the buffer passed as third parameter is NULL, the value is read from the actual PCR of the TPM and is set into the ::TPA_PCR_SET object
 *
 * @param pcrSet a pointer to a ::TPA_PCR_SET object
 * @param index the number of the PCR to set
 * @param value a buffer containing the value to set into the PCR. If the buffer is NULL, the value is read from the actual PCR of the TPM and is set into the ::TPA_PCR_SET object
 */
TPA_RESULT TpaHL_PCRSet_pcr(TPA_PCR_SET *pcrSet, int index,
		unsigned char* value) {

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (pcrSet == NULL) {
		TRACE_MSG_PARAM_ERROR("pcrSet cannot be NULL");
		goto error;
	}
	if (index < 0 || index > 23) {
		TRACE_MSG_PARAM_ERROR("index is not valid");
		goto error;
	}

	if (pcrSet->index >= pcrSet->pcrList_size) {
		TRACE_MSG_PARAM_ERROR("pcrSet is full");
		goto error;
	}

#endif

	pcrSet->pcrList[pcrSet->index].pcr = index;

	if (value == NULL) {
		TpaHL_PCR_Read(&pcrSet->pcrList[pcrSet->index++]);
	} else {
		pcrSet->pcrList[pcrSet->index].value = (unsigned char *) calloc(
				pcrSet->pcr_size, sizeof(unsigned char));
		if (pcrSet->pcrList[pcrSet->index].value == NULL) {
			TRACE_MSG_ERROR("calloc failed");
			goto error;
		}
		memcpy(pcrSet->pcrList[pcrSet->index++].value, value, pcrSet->pcr_size);

	}

	result = TPA_SUCCESS;
	error: return result;
}

/****************************************************************************/
/*																			*/
/*							 Generic Object									*/
/*																			*/
/****************************************************************************/

/**
 * Allocates and initializes a ::TPA_GENERIC_OBJPTR object
 *
 * @param objPtr a double pointer to a ::TPA_GENERIC_OBJPTR object
 */
TPA_RESULT TpaHL_OBJPTR_allocate(TPA_GENERIC_OBJPTR **objPtr) {

#ifndef __NO_PARAM_CHECK
	/*
	 * Parameter check
	 */
	if (objPtr == NULL) {
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	*objPtr = (TPA_GENERIC_OBJPTR *) malloc(sizeof(TPA_GENERIC_OBJPTR));
	if (*objPtr == NULL) {
		return TPA_ERROR;
	}

	(*objPtr)->objptr = NULL;
	(*objPtr)->saved_objid = 0;
	(*objPtr)->type = 0;

	return TPA_SUCCESS;
}

/**
 * Frees a ::TPA_GENERIC_OBJPTR object
 *
 * @param objPtr a pointer to a ::TPA_GENERIC_OBJPTR object
 */
void TpaHL_OBJPTR_freeMemory(TPA_GENERIC_OBJPTR *objPtr) {

	if (!objPtr)
		return;

	/* TODO : fix me objPtr inside TPA_GENERIC_OBJPTR MUST be deallocated */

	free(objPtr);
}

/****************************************************************************/
/*																			*/
/*							 KEY Objects									*/
/*																			*/
/****************************************************************************/
TPA_RESULT TpaHL_KEY_allocate(TPA_TPMKEY **key) {
#ifndef __NO_PARAM_CHECK
	/*
	 * Parameter check
	 */
	if (key == NULL) {
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	*key = (TPA_TPMKEY *) calloc(1, sizeof(TPA_TPMKEY));
	if (*key == NULL) {
		return TPA_ERROR;
	}

	return TPA_SUCCESS;
}

TPA_RESULT TpaHL_KEY_set(TPA_TPMKEY *key, int idx, long len, void *data) {
#ifndef __NO_PARAM_CHECK
	/*
	 * Parameter check
	 */
	if (key == NULL) {
		return TPA_ERROR;
	}
	if (idx < 0) {
		return TPA_ERROR;
	}
	if (len <= 0) {
		return TPA_ERROR;
	}
	if (data == NULL) {
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	switch (idx) {
	case KEY_PUBKEY:
		key->pubKey_size = len;
		return Tpa_setMember((void *) &(key->pubKey), len, data);

	case KEY_PRIVKEY:
		key->privKey_size = len;
		return Tpa_setMember((void *) &(key->privKey), len, data);

	case KEY_CERT:
		key->cert_size = len;
		return Tpa_setMember((void *) &(key->cert), len, data);

	case KEY_LABEL:
		key->label_len = len;
		return Tpa_setString((void *) &(key->label), len, data);

	case KEY_TPMKEYSECRET:
		key->keySecret_len = len;
		return Tpa_setString((void *) &(key->keySecret), len, data);

	default:
		return TPA_ERROR;
		break;
	}

	/*
	 * If it reaches here, something went wrong
	 */
	return TPA_ERROR;
}

void TpaHL_KEY_freeMemory(TPA_TPMKEY *key) {
	if (!key)
		return;
	if (key->keySecret)
		free(key->keySecret);
	key->keySecret = NULL;
	key->keySecret_len = 0;
	if (key->label)
		free(key->label);
	key->label = NULL;
	key->label_len = 0;
	if (key->cert)
		free(key->cert);
	key->cert = NULL;
	key->cert_size = 0;
	if (key->pubKey)
		free(key->pubKey);
	key->pubKey = NULL;
	key->pubKey_size = 0;
	if (key->privKey)
		free(key->privKey);
	key->privKey = NULL;
	key->privKey_size = 0;
	if (key->blob)
		free(key->blob);
	key->blob = NULL;
	key->blob_size = 0;
	key->tpmkey_id = 0;
	key->keyType = 0;
	key->keyFlags = 0;
	key->parentKey = NULL;
	key->childKey = NULL;
	key->parent_id = 0;
	//key->sealingPcrs;
	if (key->tss_info)
		free(key->tss_info);
	free(key);
}

/*********************************************************************/
/* CODE TOBE CHECKED                                                 */
/*********************************************************************/

/**
 * Allocates a ::TPA_RA object
 *
 * @param ra a double pointer to a ::TPA_RA object
 */
TPA_RESULT TpaHL_RA_allocate(TPA_RA **ra) {

#ifndef __NO_PARAM_CHECK
	if (ra == NULL) {
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	*ra = (TPA_RA *) malloc(sizeof(TPA_RA));
	if (*ra == NULL) {
		return TPA_ERROR;
	}

	(*ra)->raNonce = NULL;
	(*ra)->raNonce_size = 0;
	(*ra)->aik = NULL;
	(*ra)->pcrs = NULL;
	(*ra)->raSign = NULL;
	(*ra)->raSign_size = 0;
	(*ra)->measures = NULL;
	(*ra)->flags = TPA_RA_FLAG_DEFAULT;

	return TPA_SUCCESS;
}

/**
 * Sets a member of a ::TPA_RA object
 *
 * @param ra a pointer to a ::TPA_RA object
 * @param idx the parameter to set
 * @param len the length of the value to set
 * @param data a buffer containing the value to set
 */
TPA_RESULT TpaHL_RA_set(TPA_RA *ra, int idx, long len, void * data) {
#ifndef __NO_PARAM_CHECK
	/*
	 * Parameter check
	 */
	if (ra == NULL) {
		return TPA_ERROR;
	}
	if (idx < 0) {
		return TPA_ERROR;
	}
	if (len <= 0) {
		return TPA_ERROR;
	}
	if (data == NULL) {
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	switch (idx) {
	case RA_NONCE:
		ra->raNonce_size = len;
		return Tpa_setMember((void *) &(ra->raNonce), len, data);
	default:
		return TPA_ERROR;
	}

	/*
	 * If it reaches here, something went wrong
	 */
	return TPA_ERROR;
}

/**
 * Frees a ::TPA_RA object
 *
 * @param ra a pointer to a ::TPA_RA object
 */
void TpaHL_RA_freeMemory(TPA_RA *ra) {

	if (ra == NULL) {
		return;
	}
	if (ra->raNonce != NULL) {
		free(ra->raNonce);
		ra->raNonce = NULL;
	}
	ra->raNonce_size = 0;
	if (ra->aik != NULL) {
		TpaHL_AIK_freeMemory(ra->aik);
		ra->aik = NULL;
	}
	if (ra->pcrs != NULL) {
		TpaHL_PCRSet_freeMemory(ra->pcrs);
		ra->pcrs = NULL;
	}
	if (ra->raSign != NULL) {
		free(ra->raSign);
		ra->raSign = NULL;
	}
	ra->raSign_size = 0;
	if (ra->measures != NULL) {
		TpaHL_MeasureSet_freeMemory(ra->measures);
		ra->measures = NULL;
	}
	ra->flags = 0;

	free(ra);
	return;
}

/**
 * Frees a ::TPA_MEASURE object. Relies on Tpa_Measure_freeMemory().
 *
 * @param measure a pointer to a ::TPA_MEASURE object
 */
void TpaHL_Measure_freeMemory(TPA_MEASURE *measure) {
	Tpa_Measure_freeMemory(measure, 1);
}

/**
 * Frees a ::TPA_MEASURE_SET object
 *
 * @param measures a pointer to a ::TPA_MEASURE_SET object
 */
void TpaHL_MeasureSet_freeMemory(TPA_MEASURE_SET *measures) {
	int i = 0;

	if (measures == NULL)
		return;

	for (i = 0; i < measures->index; i++)
		Tpa_Measure_freeMemory(&measures->measureList[i], 0);

	memset(measures->measureList, 0, sizeof(TPA_MEASURE)
			* measures->measureList_size);
	measures->index = 0;
	measures->measureList_size = 0;
	free(measures->measureList);
	measures->measureList = NULL;

	free(measures);
}
