/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include "TPA_API.h"
#include "TPA_Core.h"
#include "TPA_Utils.h"

#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <arpa/inet.h>

/****************************************************************************/
/*							 		PCRSet									*/
/****************************************************************************/

TPA_RESULT TpaHL_PCRSet_Serialize(TPA_PCR_SET *selectedPcrs,
		unsigned char **blob, unsigned int *blob_size) {

	return Tpa_PCRSet_Serialize(selectedPcrs, blob, blob_size);
}

TPA_RESULT TpaHL_PCRSet_Deserialize(unsigned char *blob,
		unsigned int blob_size, unsigned char **blob_end,
		TPA_PCR_SET **selectedPcrs) {
#ifdef __TSS_HL
	if (Tpa_TSS_GetPcrSet(buf, buf_size, pcrs) != TPA_SUCCESS)
	return TPA_ERROR;
	return TPA_SUCCESS;
#else
	TPA_RESULT result = TPA_ERROR;
	TPA_PCR_INFO selection = { 0 };
	uint8_t mask = 0x01;
	uint8_t tmask = 0x00;
	unsigned char *p = NULL;
	int i = 0;
	int j = 0;
	int counter = 0;

	if (blob == NULL || blob_size == 0) {
		TRACE_MSG_ERROR("blob or blob_size cannot be null");
		goto error;
	}

	p = blob;
	//	memcpy(&selection.sizeOfSelect, p, sizeof(selection.sizeOfSelect));
	selection.sizeOfSelect = ntohs(*(uint16_t*) p);
	selection.pcrSelect = (unsigned char *) calloc(selection.sizeOfSelect,
			sizeof(unsigned char));
	if (selection.pcrSelect == NULL) {
		TRACE_MSG_ERROR("calloc failed");
		goto error;
	}

	p += sizeof(uint16_t);
	memcpy(selection.pcrSelect, p, selection.sizeOfSelect);
	p += selection.sizeOfSelect;

	// skip size
	p += sizeof(uint32_t);

	for (i = 0; i < selection.sizeOfSelect; i++) {
		tmask = selection.pcrSelect[i];
		for (j = 0; j < 8; j++) {
			if (tmask & mask)
				counter++;
			tmask >>= 1;
		}
	}

	if (TpaHL_PCRSet_initialize(selectedPcrs, counter, 20) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaHL_PCRSet_initialize failed");
		goto error;
	}

	int current_pcr = 0;
	for (i = 0; i < selection.sizeOfSelect; i++) {
		tmask = selection.pcrSelect[i];
		for (j = 0; j < 8; j++) {
			if (tmask & mask) {
				if (TpaHL_PCRSet_pcr(*selectedPcrs, current_pcr, p)
						!= TPA_SUCCESS) {
					TRACE_MSG_ERROR("TpaHL_PCRSet_pcr failed");
					goto error;
				}
				p += (*selectedPcrs)->pcr_size;
			}
			tmask >>= 1;
			current_pcr++;
		}
	}

	if (blob_end != NULL)
		*blob_end = p;

	result = TPA_SUCCESS;
	error: if (selection.pcrSelect)
		free(selection.pcrSelect);
	return result;
#endif
}

/****************************************************************************/
/*							 		RA										*/
/****************************************************************************/

/**
 * Serializes a ::TPA_RA object.
 */
TPA_RESULT TpaHL_RA_Serialize(TPA_RA *ra, unsigned char **blob,
		unsigned int *blob_size) {

	TPA_RESULT result = TPA_ERROR;
	unsigned int blobPcrs_size = 0;
	unsigned int blobNonce_size = 0;
	unsigned int blobMeasures_size = 0;
	unsigned int blobAIKPub_size = 0;
	unsigned int blobAIKCert_size = 0;
	unsigned int blobPCA_size = 0;
	unsigned char *blobPcrs = NULL;
	unsigned char *blobMeasures = NULL;
	unsigned char *tmp = NULL;
	unsigned int tmp_size = 0;
	const unsigned int tlvHeader = 1 + sizeof(uint32_t);

#ifndef __NO_PARAM_CHECK
	if (ra == NULL) {
		TRACE_MSG_PARAM_ERROR("ra is NULL");
		goto error;
	}
	if (ra->pcrs == NULL) {
		TRACE_MSG_PARAM_ERROR("ra->pcrs is NULL");
		goto error;
	}
	if (ra->raSign == NULL) {
		TRACE_MSG_PARAM_ERROR("ra->raSign is NULL");
		goto error;
	}

	if (blob == NULL) {
		TRACE_MSG_PARAM_ERROR("blob is NULL");
		goto error;
	}
	if (blob_size == NULL) {
		TRACE_MSG_PARAM_ERROR("blob_size is NULL");
		goto error;
	}
#endif

	*blob = NULL;
	*blob_size = 0;

	if (TpaHL_PCRSet_Serialize(ra->pcrs, &blobPcrs, &blobPcrs_size)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaHL_PCRSet_Serialize failed");
		goto error;
	}

	/* serialize also nonce */
	if ((ra->flags & TPA_RA_FLAG_NONCE) && (ra->raNonce != NULL)) {
		blobNonce_size = ra->raNonce_size + tlvHeader;
	}

	/* serialize also measures */
	if ((ra->flags & TPA_RA_FLAG_MEASURES) && (ra->measures != NULL)) {
		if (TpaHL_MeasureSet_Serialize(ra->measures, &blobMeasures,
				&blobMeasures_size) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaHL_MeasureSet_Serialize failed");
			goto error;
		}
		blobMeasures_size += tlvHeader;
	}

	/* serialize also AIK cert */
	if ((ra->flags & TPA_RA_FLAG_AIK_CERT) && (ra->aik != NULL)) {
		if (ra->aik->cert != NULL)
			blobAIKCert_size = ra->aik->cert_size + tlvHeader;

		/* serialize also PCA cert (only available if AIK is certified) */
		//TODO
	} else
	/* serialize also AIK pub - useless if AIK cert is serialized */
	if ((ra->flags & TPA_RA_FLAG_AIK_PUB) && (ra->aik != NULL)) {
		if (ra->aik->pubKey != NULL)
			blobAIKPub_size = ra->aik->pubKey_size + tlvHeader;
	}

	*blob_size = blobPcrs_size + (sizeof(uint32_t) + ra->raSign_size)
			+ blobNonce_size + blobMeasures_size + blobAIKPub_size
			+ blobAIKCert_size + blobPCA_size;

	*blob = malloc(*blob_size);
	if (*blob == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	tmp = *blob;

	memcpy(tmp, blobPcrs, blobPcrs_size);
	tmp += blobPcrs_size;

	*(uint32_t *) tmp = (uint32_t) htonl(ra->raSign_size);
	tmp += sizeof(uint32_t);
	memcpy(tmp, ra->raSign, ra->raSign_size);
	tmp += ra->raSign_size;

	if (blobNonce_size) {
		tmp_size = blobNonce_size - tlvHeader;
		*tmp = TPA_RA_SERIAL_NONCE;
		tmp++;
		*(uint32_t *) tmp = (uint32_t) htonl(tmp_size);
		tmp += sizeof(uint32_t);
		memcpy(tmp, ra->raNonce, tmp_size);
		tmp += tmp_size;
	}

	if (blobMeasures_size) {
		tmp_size = blobMeasures_size - tlvHeader;
		*tmp = TPA_RA_SERIAL_MEASURES;
		tmp++;
		*(uint32_t *) tmp = (uint32_t) htonl(tmp_size);
		tmp += sizeof(uint32_t);
		memcpy(tmp, blobMeasures, tmp_size);
		tmp += tmp_size;
	}

	if (blobAIKPub_size) {
		tmp_size = blobAIKPub_size - tlvHeader;
		*tmp = TPA_RA_SERIAL_AIK_PUB;
		tmp++;
		*(uint32_t *) tmp = (uint32_t) htonl(tmp_size);
		tmp += sizeof(uint32_t);
		memcpy(tmp, ra->aik->pubKey, tmp_size);
		tmp += tmp_size;
	}

	if (blobAIKCert_size) {
		tmp_size = blobAIKCert_size - tlvHeader;
		*tmp = TPA_RA_SERIAL_AIK_CERT;
		tmp++;
		*(uint32_t *) tmp = (uint32_t) htonl(tmp_size);
		tmp += sizeof(uint32_t);
		memcpy(tmp, ra->aik->cert, tmp_size);
		tmp += tmp_size;
	}

	if (blobPCA_size) {
		tmp_size = blobPCA_size - tlvHeader;
		*tmp = TPA_RA_SERIAL_AIK_PCA;
		tmp++;
		*(uint32_t *) tmp = (uint32_t) htonl(tmp_size);
		tmp += sizeof(uint32_t);
		//TODO
		// memcpy(tmp, ..., tmp_size);
		tmp += tmp_size;
	}

	free(blobPcrs);
	free(blobMeasures);

	return TPA_SUCCESS;

	error: /* Free */
	if (blobPcrs)
		free(blobPcrs);
	if (blobMeasures)
		free(blobMeasures);
	if (blob != NULL && *blob != NULL) {
		free(*blob);
		*blob = NULL;
		*blob_size = 0;
	}
	return result;
}

/**
 * Deserializes a ::TPA_RA object.
 *
 * TODO:
 * - PCA
 * - what if some param of ra are not NULL? Should we... free 'em!?
 */
TPA_RESULT TpaHL_RA_Deserialize(unsigned char *blob, unsigned int blob_size,
		unsigned char **blob_end, TPA_RA **ra) {

	TPA_RESULT result = TPA_ERROR;
	unsigned char *tmp = NULL;
	unsigned char *tmp_test = NULL;
	unsigned char type = 0;
	unsigned int length = 0;

#ifndef __NO_PARAM_CHECK
	if (blob == NULL) {
		TRACE_MSG_PARAM_ERROR("blob is NULL");
		goto error;
	}
	if (blob_size == 0) {
		TRACE_MSG_PARAM_ERROR("blob_size is 0");
		goto error;
	}

	if (ra == NULL) {
		TRACE_MSG_PARAM_ERROR("ra is NULL");
		goto error;
	}
#endif
	if (*ra == NULL) {
		if (TpaHL_RA_allocate(ra) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("TpaHL_RA_allocate failed");
			goto error;
		}
	}

	/* PCR */
	if (TpaHL_PCRSet_Deserialize(blob, blob_size, &tmp, &((*ra)->pcrs))
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaHL_PCRSet_Deserialize failed");
		goto error;
	}

	/* raSign_size */
	if (tmp - blob + sizeof(uint32_t) > blob_size) {
		TRACE_MSG_ERROR("blob_size insufficient for raSign_size");
		goto error;
	}
	(*ra)->raSign_size = ntohl(*((uint32_t *) tmp));
	tmp += sizeof(uint32_t);

	/* raSign */
	if (tmp - blob + (*ra)->raSign_size > blob_size) {
		TRACE_MSG_ERROR("blob_size insufficient for raSign");
		goto error;
	}

	(*ra)->raSign = malloc((*ra)->raSign_size);
	if ((*ra)->raSign == NULL) {
		TRACE_MSG_ERROR("malloc failed (raSign)");
		goto error;
	}
	memcpy((*ra)->raSign, tmp, (*ra)->raSign_size);
	tmp += (*ra)->raSign_size;

	while (tmp - blob < blob_size) {
		type = *tmp;
		tmp++;

		if (tmp - blob + sizeof(uint32_t) > blob_size) {
			TRACE_MSG_ERROR("blob_size insufficient for length");
			goto error;
		}
		length = ntohl(*((uint32_t *) tmp));
		tmp += sizeof(uint32_t);

		if (tmp - blob + length > blob_size) {
			TRACE_MSG_ERROR("blob_size insufficient for value");
			goto error;
		}

		switch (type) {
		case TPA_RA_SERIAL_NONCE:
			(*ra)->raNonce_size = length;
			(*ra)->raNonce = malloc(length);
			if ((*ra)->raNonce == NULL) {
				TRACE_MSG_ERROR("malloc failed (raNonce)");
				goto error;
			}
			memcpy((*ra)->raNonce, tmp, length);
			break;
		case TPA_RA_SERIAL_MEASURES:
			if (TpaHL_MeasureSet_Deserialize(tmp, length, &tmp_test,
					&((*ra)->measures)) != TPA_SUCCESS) {
				TRACE_MSG_ERROR("TpaHL_MeasureSet_Deserialize failed");
				goto error;
			}
			if (tmp_test - tmp != length) {
				TRACE_MSG_ERROR("TpaHL_MeasureSet_Deserialize: something goes wrong");
				goto error;
			}
			break;
		case TPA_RA_SERIAL_AIK_PUB:
			if ((*ra)->aik == NULL) {
				if (TpaHL_AIK_allocate(&((*ra)->aik)) != TPA_SUCCESS) {
					TRACE_MSG_ERROR("TpaHL_AIK_allocate failed");
					goto error;
				}
			}
			(*ra)->aik->pubKey_size = length;
			(*ra)->aik->pubKey = malloc(length);
			if ((*ra)->aik->pubKey == NULL) {
				TRACE_MSG_ERROR("malloc failed (aik->pubKey)");
				goto error;
			}
			memcpy((*ra)->aik->pubKey, tmp, length);
			break;
		case TPA_RA_SERIAL_AIK_CERT:
			if ((*ra)->aik == NULL) {
				if (TpaHL_AIK_allocate(&((*ra)->aik)) != TPA_SUCCESS) {
					TRACE_MSG_ERROR("TpaHL_AIK_allocate failed");
					goto error;
				}
			}
			(*ra)->aik->cert_size = length;
			(*ra)->aik->cert = malloc(length);
			if ((*ra)->aik->cert == NULL) {
				TRACE_MSG_ERROR("malloc failed (aik->cert)");
				goto error;
			}
			memcpy((*ra)->aik->cert, tmp, length);
			break;
		case TPA_RA_SERIAL_AIK_PCA:
			//TODO
			break;
		default:
			break;
		}

		tmp += length;
	}

	if (blob_end != NULL)
		*blob_end = tmp;

	return TPA_SUCCESS;

	error: /* Free */
	return result;
}
