/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <openssl/evp.h>
#include <openssl/buffer.h>

#include "TPA_Data.h"
#include "TPA_TSS.h"
#include "TPA_Utils.h"

#define TPMSEAL_HDR_STRING "-----BEGIN TSS-----\n"
#define TPMSEAL_FTR_STRING "-----END TSS-----\n"
#define TPMSEAL_TSS_STRING "-----TSS KEY-----\n"
#define TPMSEAL_EVP_STRING "-----ENC KEY-----\n"
#define TPMSEAL_ENC_STRING "-----ENC DAT-----\n"
#define TPMSEAL_KEYTYPE_SYM "Symmetric Key: "
#define TPMSEAL_CIPHER_AES256CBC "AES-256-CBC\n"
#define EVPKEY_DEFAULT_SIZE 512

/* WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING
 *
 * Dear programmer it's highly recommended to change the IV
 * each time you encrypt something. Support for that will be given
 * soon. Now DIY (Do It Yourself).
 *
 * WARNING WARNING WARNING WARNING WARNING WARNING WARNING WARNING */
#define TPMBIND_IV "IVBINDINGIVBINDI"

TPA_RESULT Tpa_DATA_BindBlob(TPA_TPM *tpm, TPA_TPMKEY *key,
		unsigned char *plainblob, unsigned int plainblob_size,
		unsigned char **boundblob, unsigned int *boundblob_size) {

	TPA_RESULT result = TPA_ERROR;
	TPA_TSS_ACCESS *tss = NULL;
	BIO *bin = NULL;
	BIO *b64 = NULL;
	BIO *bdata = NULL;
	BUF_MEM *pp = NULL;
	EVP_CIPHER_CTX ctx;
	unsigned char *randKey = NULL;
	unsigned char *encKey = NULL;
	unsigned int randKey_size = 0;
	unsigned int encKey_size = 0;
	int lineLen = 0;
	int encDataLen = 0;
	int bioRc = 0;
	unsigned char line[EVP_CIPHER_block_size(EVP_aes_256_cbc()) * 16];
	unsigned char encData[sizeof(line) + EVP_CIPHER_block_size(
			EVP_aes_256_cbc())];

	/* Check parameters */
	if (tpm == NULL) {
		TRACE_MSG_ERROR("tpm object is null");
	}

	if (plainblob == NULL || plainblob_size < 1) {
		TRACE_MSG_ERROR("plainblob pointer is null or size is not correct");
	}

	if (boundblob == NULL || boundblob_size == NULL) {
		TRACE_MSG_ERROR("boundblob pointer or size pointer is not valid");
	}

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection to TSS */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	/* Retrieve random data used as symmetric key for the encryption */
	randKey_size = EVP_CIPHER_key_length(EVP_aes_256_cbc());
	if (Tpa_TSS_GetRandom(tss, randKey_size, &randKey) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetRandom failed");
		goto error;
	}

	/******************************************************************************
	 *******************************************************************************
	 *******************************************************************************
	 *				 THIS IS A COMMON PART - TODO: REMOVE						  *
	 *******************************************************************************
	 *******************************************************************************
	 ******************************************************************************/
	/* Load SRK Key */
	if (Tpa_TSS_LoadSRK(tss, tpm->srkSecret, tpm->srkSecret_len) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}

	if (Tpa_TSS_LoadStorageMasterKey(tss, key->parentKey->blob,
			key->parentKey->blob_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}

	if (Tpa_TSS_LoadBindingKey(tss, key->blob, key->blob_size, key->keySecret,
			key->keySecret_len, NULL, (TPA_TSS_KEY **) &key->tss_info)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_CreateBindingKey failed");
		goto error;
	}
	/******************************************************************************
	 *******************************************************************************
	 *******************************************************************************
	 *									END										  *
	 *******************************************************************************
	 *******************************************************************************
	 ******************************************************************************/

	/* Seal the symmetric key that will be used for encrypt the plain text */
	if (Tpa_TSS_Bind(tss, (TPA_TSS_KEY *) key->tss_info, randKey, randKey_size,
			&encKey, &encKey_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Bind failed");
		goto error;
	}

	/* Create a BIO to perform base64 encoding */
	if ((b64 = BIO_new(BIO_f_base64())) == NULL) {
		TRACE_MSG_ERROR("Unable to open base64 BIO");
		goto error;
	}

	/* Create a BIO for the output buffer */
	if ((bdata = BIO_new(BIO_s_mem())) == NULL) {
		TRACE_MSG_ERROR("Unable to open memory BIO");
		goto error;
	}

	/* Output the sealed data header string */
	BIO_puts(bdata, TPMSEAL_HDR_STRING);

	/* Sealed EVP Symmetric Key */
	BIO_puts(bdata, TPMSEAL_EVP_STRING);
	BIO_puts(bdata, TPMSEAL_KEYTYPE_SYM);
	BIO_puts(bdata, TPMSEAL_CIPHER_AES256CBC);
	bdata = BIO_push(b64, bdata);
	BIO_write(bdata, encKey, encKey_size);
	if (BIO_flush(bdata) != 1) {
		TRACE_MSG_ERROR("Unable to flush output\n");
		goto error;
	}
	bdata = BIO_pop(b64);

	/* Create a read BIO */
	if ((bin = BIO_new_mem_buf(plainblob, plainblob_size)) == NULL) {
		TRACE_MSG_ERROR("BIO_new_mem_buf failed");
		goto error;
	}

	EVP_CIPHER_CTX_init(&ctx);

	/* Encrypted Data */
	BIO_puts(bdata, TPMSEAL_ENC_STRING);
	bdata = BIO_push(b64, bdata);

	EVP_EncryptInit(&ctx, EVP_aes_256_cbc(), randKey,
			(unsigned char *) TPMBIND_IV);

	while ((lineLen = BIO_read(bin, line, sizeof(line))) > 0) {
		EVP_EncryptUpdate(&ctx, encData, &encDataLen, line, lineLen);
		BIO_write(bdata, encData, encDataLen);
	}

	EVP_EncryptFinal(&ctx, encData, &encDataLen);

	EVP_CIPHER_CTX_cleanup(&ctx);

	BIO_write(bdata, encData, encDataLen);
	if (BIO_flush(bdata) != 1) {
		TRACE_MSG_ERROR("Unable to flush output");
		goto error;
	}
	bdata = BIO_pop(b64);

	BIO_puts(bdata, TPMSEAL_FTR_STRING);

	/* Copy data */
	BIO_get_mem_ptr(bdata, &pp);
	*boundblob_size = pp->length;
	*boundblob = (unsigned char *) calloc(*boundblob_size,
			sizeof(unsigned char));
	if (*boundblob == NULL) {
		TRACE_MSG_ERROR("calloc failed");
		goto error;
	}

	memcpy(*boundblob, pp->data, *boundblob_size);

	result = TPA_SUCCESS;

	error: /* Free */
	if (key->tss_info) {
		Tpa_TSS_KeyFreeMemory(tss, (TPA_TSS_KEY *) key->tss_info);
		key->tss_info = NULL;
	}
	if (tss)
		Tpa_TSS_Cleanup(tss);
	if (bin) {
		bioRc = BIO_set_close(bin, BIO_CLOSE);
		BIO_free(bin);
	}
	if (b64)
		BIO_free(b64);
	if (bdata)
		BIO_free(bdata);
	if (randKey)
		free(randKey);
	if (encKey)
		free(encKey);

	return result;
}

TPA_RESULT Tpa_DATA_UnbindBlob(TPA_TPM *tpm, unsigned char *boundblob,
		unsigned int boundblob_size, TPA_TPMKEY *key,
		unsigned char **unboundblob, unsigned int *unboundblob_size) {

	TPA_RESULT result = TPA_ERROR;
	TPA_TSS_ACCESS *tss = NULL;
	BIO *bdata = NULL;
	BIO *bmem = NULL;
	BIO *b64 = NULL;
	BYTE *evpKeyData = NULL;
	BYTE *rcPtr = NULL;
	BYTE *symKey = NULL;
	EVP_CIPHER_CTX ctx;
	UINT32 symKeyLen = 0;
	unsigned char *res_data = NULL;
	int res_size = 0;
	int evpKeyDataSize = 0;
	int evpLen = 0;
	int bioRc = 0;
	int rcLen = 0;
	char data[EVP_CIPHER_block_size(EVP_aes_256_cbc()) * 16];

#ifndef __NO_PARAM_CHECK
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm is NULL");
		goto error;
	}
	if (tpm->srkSecret == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm->srkSecret is NULL");
		goto error;
	}
	if (boundblob == NULL) {
		TRACE_MSG_PARAM_ERROR("sealedblob is NULL");
		goto error;
	}

	if (unboundblob == NULL) {
		TRACE_MSG_PARAM_ERROR("unsealedblob is NULL");
		goto error;
	}
	if (unboundblob_size == NULL) {
		TRACE_MSG_PARAM_ERROR("unsealedblob_size is NULL");
		goto error;
	}
#endif

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection to TSS*/
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	/******************************************************************************
	 *******************************************************************************
	 *******************************************************************************
	 *				 THIS IS A COMMON PART - TODO: REMOVE						  *
	 *******************************************************************************
	 *******************************************************************************
	 ******************************************************************************/
	/* Load SRK Key */
	if (Tpa_TSS_LoadSRK(tss, tpm->srkSecret, tpm->srkSecret_len) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}

	if (Tpa_TSS_LoadStorageMasterKey(tss, key->parentKey->blob,
			key->parentKey->blob_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}

	if (Tpa_TSS_LoadBindingKey(tss, key->blob, key->blob_size, key->keySecret,
			key->keySecret_len, NULL, (TPA_TSS_KEY **) &key->tss_info)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_CreateBindingKey failed");
		goto error;
	}
	/******************************************************************************
	 *******************************************************************************
	 *******************************************************************************
	 *									END										  *
	 *******************************************************************************
	 *******************************************************************************
	 ******************************************************************************/

	/* Create a read BIO */
	if ((bdata = BIO_new_mem_buf(boundblob, boundblob_size)) == NULL) {
		TRACE_MSG_ERROR("BIO_new_mem_buf failed");
		goto error;
	}

	/* Test file header for TSS */
	BIO_gets(bdata, data, sizeof(data));
	if (strncmp(data, TPMSEAL_HDR_STRING, strlen(TPMSEAL_HDR_STRING)) != 0) {
		TRACE_MSG_ERROR("Boundary BEGIN TSS in sealed blob is missing");
		goto error;
	}

	/* Test file header for ENC KEY */
	BIO_gets(bdata, data, sizeof(data));
	if (strncmp(data, TPMSEAL_EVP_STRING, strlen(TPMSEAL_EVP_STRING)) != 0) {
		TRACE_MSG_ERROR("Boundary BEGIN TSS in sealed blob is missing");
		goto error;
	}

	/* Make sure it's a supported cipher (currently only AES 256 CBC) */
	BIO_gets(bdata, data, sizeof(data));
	if (strncmp(data + strlen(TPMSEAL_KEYTYPE_SYM), TPMSEAL_CIPHER_AES256CBC,
			strlen(TPMSEAL_CIPHER_AES256CBC)) != 0) {
		TRACE_MSG_ERROR("Unsupported cipher");
		goto error;
	}

	/* Create a memory BIO to hold the base64 TSS key */
	if ((bmem = BIO_new(BIO_s_mem())) == NULL) {
		TRACE_MSG_ERROR("BIO creation failed");
		goto error;
	}
	BIO_set_mem_eof_return(bmem, 0);
	bioRc = BIO_set_close(bmem, BIO_CLOSE);

	/* Read the base64 Symmetric key into the memory BIO */
	while ((rcLen = BIO_gets(bdata, data, sizeof(data))) > 0) {
		/* Look for Encrypted Data Header (end of key) */
		if (strncmp(data, TPMSEAL_ENC_STRING, strlen(TPMSEAL_ENC_STRING)) == 0)
			break;
		if (BIO_write(bmem, data, rcLen) <= 0) {
			TRACE_MSG_ERROR("BIO_write error");
			goto error;
		}
	}

	/* Test file header for ENC DAT */
	if (strncmp(data, TPMSEAL_ENC_STRING, strlen(TPMSEAL_ENC_STRING)) != 0) {
		TRACE_MSG_ERROR("Boundary ENC DAT in sealed blob is missing");
		goto error;
	}

	/* Create a base64 BIO to decode the Symmetric key */
	if ((b64 = BIO_new(BIO_f_base64())) == NULL) {
		TRACE_MSG_ERROR("BIO base64 creation failed");
		goto error;
	}

	/* Decode the Symmetric key */
	bmem = BIO_push(b64, bmem);
	while ((rcLen = BIO_read(bmem, data, sizeof(data))) > 0) {
		if ((evpLen + rcLen) > evpKeyDataSize) {
			evpKeyDataSize += EVPKEY_DEFAULT_SIZE;
			rcPtr = realloc(evpKeyData, evpKeyDataSize);
			if (rcPtr == NULL) {
				TRACE_MSG_ERROR("realloc failed during symmetric key decoding");
				goto error;
			}
			evpKeyData = rcPtr;
		}
		memcpy(evpKeyData + evpLen, data, rcLen);
		evpLen += rcLen;
	}
	bmem = BIO_pop(b64);
	BIO_free(b64);
	b64 = NULL;
	bioRc = BIO_reset(bmem);

	/* Read the base64 encrypted data into the memory BIO */
	while ((rcLen = BIO_gets(bdata, data, sizeof(data))) > 0) {
		/* Look for TSS Footer (end of data) */
		if (strncmp(data, TPMSEAL_FTR_STRING, strlen(TPMSEAL_FTR_STRING)) == 0)
			break;

		if (BIO_write(bmem, data, rcLen) <= 0) {
			TRACE_MSG_ERROR("error in reading b64 encrypted data");
			goto error;
		}
	}
	if (strncmp(data, TPMSEAL_FTR_STRING, strlen(TPMSEAL_FTR_STRING)) != 0) {
		TRACE_MSG_ERROR("error in reading b64 encrypted data");
		goto error;
	}

	/* Unseal blob */
	if (Tpa_TSS_Unbind(tss, (TPA_TSS_KEY *) key->tss_info, evpKeyData, evpLen,
			&symKey, &symKeyLen) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_UnsealBlobWithMK failed");
		goto error;
	}

	/* Decrypt blob with symmetric key */
	/* Malloc a block of storage to hold the decrypted data
	 Using the size of the mem BIO is more than enough
	 (plus an extra cipher block size) */
	res_data = malloc(BIO_pending(bmem) + EVP_CIPHER_block_size(
			EVP_aes_256_cbc()));
	if (res_data == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}

	EVP_CIPHER_CTX_init(&ctx);

	/* Decode and decrypt the encrypted data */
	EVP_DecryptInit(&ctx, EVP_aes_256_cbc(), symKey,
			(unsigned char *) TPMBIND_IV);

	/* Create a base64 BIO to decode the encrypted data */
	if ((b64 = BIO_new(BIO_f_base64())) == NULL) {
		TRACE_MSG_ERROR("BIO_new b64 creation failed");
		goto error;
	}

	bmem = BIO_push(b64, bmem);
	while ((rcLen = BIO_read(bmem, data, sizeof(data))) > 0) {
		EVP_DecryptUpdate(&ctx, res_data + res_size, &rcLen,
				(unsigned char *) data, rcLen);
		res_size += rcLen;
	}
	EVP_DecryptFinal(&ctx, res_data + res_size, &rcLen);
	EVP_CIPHER_CTX_cleanup(&ctx);

	res_size += rcLen;
	bmem = BIO_pop(b64);
	BIO_free(b64);
	b64 = NULL;
	bioRc = BIO_reset(bmem);

	*unboundblob = res_data;
	*unboundblob_size = res_size;

	result = TPA_SUCCESS;
	error: /* Free data */
	if (key->tss_info) {
		Tpa_TSS_KeyFreeMemory(tss, (TPA_TSS_KEY *) key->tss_info);
		key->tss_info = NULL;
	}
	if (tss)
		Tpa_TSS_Cleanup(tss);
	if (bdata)
		BIO_free(bdata);
	if (bmem) {
		BIO_free(bmem);
	}
	if (evpKeyData)
		free(evpKeyData);
	if (symKey)
		free(symKey);

	return result;
}

TPA_RESULT Tpa_DATA_SealBlob(TPA_TPM *tpm, unsigned char *plainblob,
		unsigned int plainblob_size, TPA_PCR_SET *selectedPcrs,
		unsigned char *mkBlob, unsigned int mkBlob_size,
		unsigned char **sealedblob, unsigned int *sealedblob_size) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT Tpa_DATA_UnsealBlob(TPA_TPM *tpm, unsigned char *sealedblob,
		unsigned int sealedblob_size, unsigned char *mkBlob,
		unsigned int mkBlob_size, unsigned char **unsealedblob,
		unsigned int *unsealedblob_size) {
	return TPA_NOT_IMPLEMENTED;
}

/* old iface: sealing */

#define TPMSEAL_HDR_STRING "-----BEGIN TSS-----\n"
#define TPMSEAL_FTR_STRING "-----END TSS-----\n"
#define TPMSEAL_TSS_STRING "-----TSS KEY-----\n"
#define TPMSEAL_EVP_STRING "-----ENC KEY-----\n"
#define TPMSEAL_ENC_STRING "-----ENC DAT-----\n"
#define TPMSEAL_KEYTYPE_SYM "Symmetric Key: "
#define TPMSEAL_CIPHER_AES256CBC "AES-256-CBC\n"
#define EVPKEY_DEFAULT_SIZE 512

/* *
 * TODO: fix me,
 * Emanuele: if the IV is static it's possible to make some attacks, right?
 * */
#define TPMSEAL_IV "IVSEALINGIVSEALI"

TPA_RESULT Tpa_CreateMKBlob(TPA_TPM *tpm, unsigned char **mkBlob,
		unsigned int *mkBlob_size) {
	TPA_RESULT result = TPA_ERROR;
	TPA_TSS_ACCESS *tss = NULL;

	/* Check parameters */
	if (mkBlob == NULL || mkBlob_size == NULL) {
		TRACE_MSG_ERROR("sealedblob pointer or size pointer is not valid");
	}

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection to TSS */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	/* Load SRK Key */
	if (Tpa_TSS_LoadSRK(tss, tpm->srkSecret, tpm->srkSecret_len) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}

	/* Create master key */
	if (Tpa_TSS_CreateStorageMasterKey(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_CreateSealingMasterKey failed");
		goto error;
	}

	/* Retrieve master key blob */
	if (Tpa_TSS_GetKeyBlob(tss, tss->hStorageMK, mkBlob, mkBlob_size)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetKeyBlob failed");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* Free */
	if (tss)
		Tpa_TSS_Cleanup(tss);
	return result;
}

TPA_RESULT Tpa_SealBlob(TPA_TPM *tpm, unsigned char *plainblob,
		unsigned int plainblob_size, TPA_PCR_SET *selectedPcrs,
		unsigned char *mkBlob, unsigned int mkBlob_size,
		unsigned char **sealedblob, unsigned int *sealedblob_size) {

	TPA_RESULT result = TPA_ERROR;
	TPA_TSS_ACCESS *tss = NULL;
	BIO *bin = NULL;
	BIO *b64 = NULL;
	BIO *bdata = NULL;
	BUF_MEM *pp = NULL;
	EVP_CIPHER_CTX ctx;
	unsigned char *randKey = NULL;
	unsigned char *encKey = NULL;
	unsigned int randKey_size = 0;
	unsigned int encKey_size = 0;
	int lineLen = 0;
	int encDataLen = 0;
	int bioRc = 0;
	unsigned char line[EVP_CIPHER_block_size(EVP_aes_256_cbc()) * 16];
	unsigned char encData[sizeof(line) + EVP_CIPHER_block_size(
			EVP_aes_256_cbc())];

	/* Check parameters */
	if (tpm == NULL) {
		TRACE_MSG_ERROR("tpm object is null");
	}

	if (plainblob == NULL || plainblob_size < 1) {
		TRACE_MSG_ERROR("plainblob pointer is null or size is not correct");
	}

	if (selectedPcrs == NULL) {
		TRACE_MSG_ERROR("selectedPcrs vector is invalid");
	}

	if (sealedblob == NULL || sealedblob_size == NULL) {
		TRACE_MSG_ERROR("sealedblob pointer or size pointer is not valid");
	}

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection to TSS */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	/* Retrieve random data used as symmetric key for the encryption */
	randKey_size = EVP_CIPHER_key_length(EVP_aes_256_cbc());
	if (Tpa_TSS_GetRandom(tss, randKey_size, &randKey) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetRandom failed");
		goto error;
	}

	/* Load SRK Key */
	if (Tpa_TSS_LoadSRK(tss, tpm->srkSecret, tpm->srkSecret_len) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}

	/* Load master key */
	if (Tpa_TSS_LoadStorageMasterKey(tss, mkBlob, mkBlob_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSealingMasterKey failed");
		goto error;
	}

	/* Seal the symmetric key that will be used for encrypt the plain text */
	if (Tpa_TSS_SealBlobWithMK(tss, selectedPcrs, randKey, randKey_size,
			&encKey, &encKey_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_SealBlobWithSRK failed");
		goto error;
	}

	/* Create a BIO to perform base64 encoding */
	if ((b64 = BIO_new(BIO_f_base64())) == NULL) {
		TRACE_MSG_ERROR("Unable to open base64 BIO");
		goto error;
	}

	/* Create a BIO for the output buffer */
	if ((bdata = BIO_new(BIO_s_mem())) == NULL) {
		TRACE_MSG_ERROR("Unable to open memory BIO");
		goto error;
	}

	/* Output the sealed data header string */
	BIO_puts(bdata, TPMSEAL_HDR_STRING);

	/* Sealed EVP Symmetric Key */
	BIO_puts(bdata, TPMSEAL_EVP_STRING);
	BIO_puts(bdata, TPMSEAL_KEYTYPE_SYM);
	BIO_puts(bdata, TPMSEAL_CIPHER_AES256CBC);
	bdata = BIO_push(b64, bdata);
	BIO_write(bdata, encKey, encKey_size);
	if (BIO_flush(bdata) != 1) {
		TRACE_MSG_ERROR("Unable to flush output\n");
		goto error;
	}
	bdata = BIO_pop(b64);

	/* Create a read BIO */
	if ((bin = BIO_new_mem_buf(plainblob, plainblob_size)) == NULL) {
		TRACE_MSG_ERROR("BIO_new_mem_buf failed");
		goto error;
	}

	EVP_CIPHER_CTX_init(&ctx);

	/* Encrypted Data */
	BIO_puts(bdata, TPMSEAL_ENC_STRING);
	bdata = BIO_push(b64, bdata);

	EVP_EncryptInit(&ctx, EVP_aes_256_cbc(), randKey,
			(unsigned char *) TPMSEAL_IV);

	while ((lineLen = BIO_read(bin, line, sizeof(line))) > 0) {
		EVP_EncryptUpdate(&ctx, encData, &encDataLen, line, lineLen);
		BIO_write(bdata, encData, encDataLen);
	}

	EVP_EncryptFinal(&ctx, encData, &encDataLen);

	EVP_CIPHER_CTX_cleanup(&ctx);

	BIO_write(bdata, encData, encDataLen);
	if (BIO_flush(bdata) != 1) {
		TRACE_MSG_ERROR("Unable to flush output");
		goto error;
	}
	bdata = BIO_pop(b64);

	BIO_puts(bdata, TPMSEAL_FTR_STRING);

	/* Copy data */
	BIO_get_mem_ptr(bdata, &pp);
	*sealedblob_size = pp->length;
	*sealedblob = (unsigned char *) calloc(*sealedblob_size,
			sizeof(unsigned char));
	if (*sealedblob == NULL) {
		TRACE_MSG_ERROR("calloc failed");
		goto error;
	}

	memcpy(*sealedblob, pp->data, *sealedblob_size);

	result = TPA_SUCCESS;

	error: /* Free */
	if (tss)
		Tpa_TSS_Cleanup(tss);
	if (bin) {
		bioRc = BIO_set_close(bin, BIO_CLOSE);
		BIO_free(bin);
	}
	if (b64)
		BIO_free(b64);
	if (bdata)
		BIO_free(bdata);
	if (randKey)
		free(randKey);
	if (encKey)
		free(encKey);

	return result;
}

TPA_RESULT Tpa_UnsealBlob(TPA_TPM *tpm, unsigned char *sealedblob,
		unsigned int sealedblob_size, unsigned char *mkBlob,
		unsigned int mkBlob_size, unsigned char **unsealedblob,
		unsigned int *unsealedblob_size) {

	TPA_RESULT result = TPA_ERROR;
	TPA_TSS_ACCESS *tss = NULL;
	BIO *bdata = NULL;
	BIO *bmem = NULL;
	BIO *b64 = NULL;
	BYTE *evpKeyData = NULL;
	BYTE *rcPtr = NULL;
	BYTE *symKey = NULL;
	EVP_CIPHER_CTX ctx;
	UINT32 symKeyLen = 0;
	unsigned char *res_data = NULL;
	int res_size = 0;
	int evpKeyDataSize = 0;
	int evpLen = 0;
	int bioRc = 0;
	int rcLen = 0;
	char data[EVP_CIPHER_block_size(EVP_aes_256_cbc()) * 16];

#ifndef __NO_PARAM_CHECK
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm is NULL");
		goto error;
	}
	if (tpm->srkSecret == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm->srkSecret is NULL");
		goto error;
	}
	if (sealedblob == NULL) {
		TRACE_MSG_PARAM_ERROR("sealedblob is NULL");
		goto error;
	}
	if (mkBlob == NULL) {
		TRACE_MSG_PARAM_ERROR("mkBlob is NULL");
		goto error;
	}
	if (unsealedblob == NULL) {
		TRACE_MSG_PARAM_ERROR("unsealedblob is NULL");
		goto error;
	}
	if (unsealedblob_size == NULL) {
		TRACE_MSG_PARAM_ERROR("unsealedblob_size is NULL");
		goto error;
	}
#endif

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection to TSS*/
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	/* Load SRK Key */
	if (Tpa_TSS_LoadSRK(tss, tpm->srkSecret, tpm->srkSecret_len) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}

	/* Load master key */
	if (Tpa_TSS_LoadStorageMasterKey(tss, mkBlob, mkBlob_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSealingMasterKey failed");
		goto error;
	}

	/* Create a read BIO */
	if ((bdata = BIO_new_mem_buf(sealedblob, sealedblob_size)) == NULL) {
		TRACE_MSG_ERROR("BIO_new_mem_buf failed");
		goto error;
	}

	/* Test file header for TSS */
	BIO_gets(bdata, data, sizeof(data));
	if (strncmp(data, TPMSEAL_HDR_STRING, strlen(TPMSEAL_HDR_STRING)) != 0) {
		TRACE_MSG_ERROR("Boundary BEGIN TSS in sealed blob is missing");
		goto error;
	}

	/* Test file header for ENC KEY */
	BIO_gets(bdata, data, sizeof(data));
	if (strncmp(data, TPMSEAL_EVP_STRING, strlen(TPMSEAL_EVP_STRING)) != 0) {
		TRACE_MSG_ERROR("Boundary BEGIN TSS in sealed blob is missing");
		goto error;
	}

	/* Make sure it's a supported cipher (currently only AES 256 CBC) */
	BIO_gets(bdata, data, sizeof(data));
	if (strncmp(data + strlen(TPMSEAL_KEYTYPE_SYM), TPMSEAL_CIPHER_AES256CBC,
			strlen(TPMSEAL_CIPHER_AES256CBC)) != 0) {
		TRACE_MSG_ERROR("Unsupported cipher");
		goto error;
	}

	/* Create a memory BIO to hold the base64 TSS key */
	if ((bmem = BIO_new(BIO_s_mem())) == NULL) {
		TRACE_MSG_ERROR("BIO creation failed");
		goto error;
	}
	BIO_set_mem_eof_return(bmem, 0);
	bioRc = BIO_set_close(bmem, BIO_CLOSE);

	/* Read the base64 Symmetric key into the memory BIO */
	while ((rcLen = BIO_gets(bdata, data, sizeof(data))) > 0) {
		/* Look for Encrypted Data Header (end of key) */
		if (strncmp(data, TPMSEAL_ENC_STRING, strlen(TPMSEAL_ENC_STRING)) == 0)
			break;
		if (BIO_write(bmem, data, rcLen) <= 0) {
			TRACE_MSG_ERROR("BIO_write error");
			goto error;
		}
	}

	/* Test file header for ENC DAT */
	if (strncmp(data, TPMSEAL_ENC_STRING, strlen(TPMSEAL_ENC_STRING)) != 0) {
		TRACE_MSG_ERROR("Boundary ENC DAT in sealed blob is missing");
		goto error;
	}

	/* Create a base64 BIO to decode the Symmetric key */
	if ((b64 = BIO_new(BIO_f_base64())) == NULL) {
		TRACE_MSG_ERROR("BIO base64 creation failed");
		goto error;
	}

	/* Decode the Symmetric key */
	bmem = BIO_push(b64, bmem);
	while ((rcLen = BIO_read(bmem, data, sizeof(data))) > 0) {
		if ((evpLen + rcLen) > evpKeyDataSize) {
			evpKeyDataSize += EVPKEY_DEFAULT_SIZE;
			rcPtr = realloc(evpKeyData, evpKeyDataSize);
			if (rcPtr == NULL) {
				TRACE_MSG_ERROR("realloc failed during symmetric key decoding");
				goto error;
			}
			evpKeyData = rcPtr;
		}
		memcpy(evpKeyData + evpLen, data, rcLen);
		evpLen += rcLen;
	}
	bmem = BIO_pop(b64);
	BIO_free(b64);
	b64 = NULL;
	bioRc = BIO_reset(bmem);

	/* Read the base64 encrypted data into the memory BIO */
	while ((rcLen = BIO_gets(bdata, data, sizeof(data))) > 0) {
		/* Look for TSS Footer (end of data) */
		if (strncmp(data, TPMSEAL_FTR_STRING, strlen(TPMSEAL_FTR_STRING)) == 0)
			break;

		if (BIO_write(bmem, data, rcLen) <= 0) {
			TRACE_MSG_ERROR("error in reading b64 encrypted data");
			goto error;
		}
	}
	if (strncmp(data, TPMSEAL_FTR_STRING, strlen(TPMSEAL_FTR_STRING)) != 0) {
		TRACE_MSG_ERROR("error in reading b64 encrypted data");
		goto error;
	}

	/* Unseal blob */
	if (Tpa_TSS_UnsealBlobWithMK(tss, evpKeyData, evpLen, &symKey, &symKeyLen)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_UnsealBlobWithMK failed");
		goto error;
	}

	/* Decrypt blob with symmetric key */
	/* Malloc a block of storage to hold the decrypted data
	 Using the size of the mem BIO is more than enough
	 (plus an extra cipher block size) */
	res_data = malloc(BIO_pending(bmem) + EVP_CIPHER_block_size(
			EVP_aes_256_cbc()));
	if (res_data == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}

	EVP_CIPHER_CTX_init(&ctx);

	/* Decode and decrypt the encrypted data */
	EVP_DecryptInit(&ctx, EVP_aes_256_cbc(), symKey,
			(unsigned char *) TPMSEAL_IV);

	/* Create a base64 BIO to decode the encrypted data */
	if ((b64 = BIO_new(BIO_f_base64())) == NULL) {
		TRACE_MSG_ERROR("BIO_new b64 creation failed");
		goto error;
	}

	bmem = BIO_push(b64, bmem);
	while ((rcLen = BIO_read(bmem, data, sizeof(data))) > 0) {
		EVP_DecryptUpdate(&ctx, res_data + res_size, &rcLen,
				(unsigned char *) data, rcLen);
		res_size += rcLen;
	}
	EVP_DecryptFinal(&ctx, res_data + res_size, &rcLen);
	EVP_CIPHER_CTX_cleanup(&ctx);

	res_size += rcLen;
	bmem = BIO_pop(b64);
	BIO_free(b64);
	b64 = NULL;
	bioRc = BIO_reset(bmem);

	*unsealedblob = res_data;
	*unsealedblob_size = res_size;

	result = TPA_SUCCESS;
	error: /* Free data */
	if (tss)
		Tpa_TSS_Cleanup(tss);
	if (bdata)
		BIO_free(bdata);
	if (bmem) {
		BIO_free(bmem);
	}
	if (evpKeyData)
		free(evpKeyData);
	if (symKey)
		free(symKey);

	return result;
}


