/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include "TPA_EK.h"
#include "TPA_PCA.h"
#include "TPA_TSS.h"

TPA_RESULT Tpa_EK_ReadKeyTPM_ifx(TPA_TPM *tpm) {

	TPA_RESULT result = TPA_ERROR;
	TPA_TSS_ACCESS *tss;

	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	if (Tpa_TSS_ReadEKCert(tss, (&(tpm->ekCertificate)),
			(&(tpm->ekCertificate_size))) != 0) {
		TRACE_MSG_ERROR("getEkCert failed");
		goto error;
	}

	result = TPA_SUCCESS;
	error: Tpa_TSS_Cleanup(tss);

	return result;
}

TPA_RESULT Tpa_EK_ReadKeyFile(TPA_TPM *tpm) {

#ifndef __NO_PARAM_CHECK
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("TPA_TPM structure is NULL");
		return TPA_ERROR;
	}

	if (tpm->ekCertificate != NULL) {
		TRACE_MSG_PARAM_ERROR("ekCertificate in TPA_TPM structure is not NULL");
		return TPA_ERROR;
	}

	if (tpm->ekCertificateFilename == NULL) {
		TRACE_MSG_PARAM_ERROR(
				"EK Certificate filename is not set in TPA_TPM structure");
		return TPA_ERROR;
	}
#endif

	if (Tpa_BLOB_readFromFile(tpm->ekCertificateFilename,
			&(tpm->ekCertificate), &(tpm->ekCertificate_size)) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaHL_BLOB_readFromFile failed");
		return TPA_ERROR;
	}

	return TPA_SUCCESS;
}

TPA_RESULT Tpa_EK_ReadKey(TPA_TPM *tpm) {

#ifndef __NO_PARAM_CHECK
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("TPA_TPM structure is NULL");
		return TPA_ERROR;
	}
#endif

	if (tpm->ekCertificateFilename != NULL) {
		if (Tpa_EK_ReadKeyFile(tpm) != TPA_SUCCESS)
			return TPA_ERROR;
	} else if (tpm->ekCertificateFilename == NULL) {
		if (Tpa_EK_ReadKeyTPM_ifx(tpm) != TPA_SUCCESS)
			return TPA_ERROR;
	} else
		return TPA_ERROR;

	return TPA_SUCCESS;
}

TPA_RESULT Tpa_EK_Create(TPA_TPM *tpm, TPA_PCA *pca) {

	TPA_TSS_ACCESS *tss = NULL;
	TPA_RESULT result = TPA_ERROR;
	unsigned char *modulus = NULL;
	unsigned int modulus_size = 0;
	unsigned char *exponent = NULL;
	unsigned int exponent_size = 0;

	TRACE_MSG_BUG("Check parameters");

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	/* Set owner secret from TPM structure */
	if (Tpa_TSS_SetOwnerSecret(tss, tpm->ownSecret, tpm->ownSecret_len)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_SetOwnerSecret failed");
		goto error;
	}

	if (Tpa_TSS_GetTPMPublicEK(tss, &modulus, &modulus_size, &exponent,
			&exponent_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetTPMPublicEK failed");
		goto error;
	}

	if (Tpa_PCA_Helper_EKCertify(pca, modulus, modulus_size, exponent,
			exponent_size, &tpm->ekCertificate, &tpm->ekCertificate_size)
			!= TPA_SUCCESS) {
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* free */
	if (!tss)
		Tpa_TSS_Cleanup(tss);
	if (modulus)
		free(modulus);
	if (exponent)
		free(exponent);

	return result;
}

