/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include "TPA_KEY.h"
#include "TPA_TSS.h"

/**
 * Allocates a ::TPA_MODULE_KEY object.
 *
 * Optionally uses as local storage a ::TPA_LS object given in input, or
 * allocate a new one (useful to share the local storage with other modules).
 *
 * @param mod a double pointer to a ::TPA_MODULE_KEY object
 * @param ls a pointer to a ::TPA_LS object, set as local storage if not NULL
 */
TPA_RESULT Tpa_KEY_allocateModule(TPA_MODULE_KEY **mod, TPA_LS *ls) {
	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (mod == NULL) {
		TRACE_MSG_PARAM_ERROR("mod is NULL");
		return TPA_ERROR;
	}
#endif

	*mod = (TPA_MODULE_KEY *) calloc(1, sizeof(TPA_MODULE_KEY));
	if (*mod == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}

	if (ls != NULL) {
		(*mod)->ls = ls;
	} else {
		if ((result = Tpa_LS_allocate(&(*mod)->ls)) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("Tpa_LS_allocate failed");
			goto error;
		}
	}

	return TPA_SUCCESS;
	error: /* error */
	if (*mod != NULL) {
		Tpa_KEY_freeModuleMemory(*mod);
		*mod = NULL;
	}

	return TPA_ERROR;
}

/**
 * Frees a ::TPA_MODULE_KEY object
 *
 * @param mod a pointer to a ::TPA_MODULE_KEY object
 */
void Tpa_KEY_freeModuleMemory(TPA_MODULE_KEY *mod) {
	if (mod == NULL)
		return;

	if (mod->ls) {
		Tpa_LS_freeMemory(mod->ls);
		mod->ls = NULL;
	}

	free(mod);
}


/*
 * keys
 */

TPA_RESULT Tpa_KEY_CreateTpmBindingKey(TPA_TPM *tpm, TPA_TPMKEY *key) {

	TPA_RESULT result = TPA_ERROR;
	TPA_TSS_ACCESS *tss = NULL;

#ifndef __NO_PARAM_CHECK
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm is NULL");
		return TPA_ERROR;
	}
	if (key == NULL) {
		TRACE_MSG_PARAM_ERROR("key is NULL");
		return TPA_ERROR;
	}
	if (key->keySecret == NULL) {
		TRACE_MSG_PARAM_ERROR("keySecret is NULL");
		return TPA_ERROR;
	}
	if (key->keySecret_len == 0) {
		TRACE_MSG_PARAM_ERROR("keySecret_size is 0");
		return TPA_ERROR;
	}
	if (key->parentKey == NULL) {
		TRACE_MSG_PARAM_ERROR("key->parentKey is NULL");
		return TPA_ERROR;
	}
#endif

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	/******************************************************************************
	 *******************************************************************************
	 *******************************************************************************
	 *				 THIS IS A COMMON PART - TODO: REMOVE						  *
	 *******************************************************************************
	 *******************************************************************************
	 ******************************************************************************/
	/* Load SRK Key */
	if (Tpa_TSS_LoadSRK(tss, tpm->srkSecret, tpm->srkSecret_len) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}

	if (key->parentKey->keyType != TPA_KEY_TYPE_SMK) {
		TRACE_MSG_ERROR("root in the tree is not SMK");
		goto error;
	}

	if (Tpa_TSS_LoadStorageMasterKey(tss, key->parentKey->blob,
			key->parentKey->blob_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}
	/******************************************************************************
	 *******************************************************************************
	 *******************************************************************************
	 *									END										  *
	 *******************************************************************************
	 *******************************************************************************
	 ******************************************************************************/

	/*
	 * TODO: implement tree here
	 */

	if (Tpa_TSS_CreateBindingKey(tss, NULL, key->keySecret,
			key->keySecret_len, (TPA_TSS_KEY **) &key->tss_info)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_CreateBindingKey failed");
		goto error;
	}

	/* Extract key blob */
	if (Tpa_TSS_GetKeyBlob(tss, ((TPA_TSS_KEY *) key->tss_info)->hKey,
			&(key->blob), &(key->blob_size)) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetKeyBlob failed");
		goto error;
	}

	/* Extract public key */
	if (Tpa_TSS_GetKeyModulus(tss, ((TPA_TSS_KEY *) key->tss_info)->hKey,
			&(key->pubKey), &(key->pubKey_size)) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetKeyBlob failed");
		goto error;

	}

	result = TPA_SUCCESS;

	error: /* Free! */
	if (key->tss_info) {
		Tpa_TSS_KeyFreeMemory(tss, (TPA_TSS_KEY *) key->tss_info);
		key->tss_info = NULL;
	}
	if (tss)
		Tpa_TSS_Cleanup(tss);

	return result;

	/*
	 * Time is running out, so this is not the definitive solution
	 * Two level tree is implemented
	 */
#if 0
	TPA_TPMKEY *tail = NULL;
	TPA_TPMKEY *p = NULL;

	tail = key;
	p = key;
	while (!p->parentKey)
	p = p->parentKey;

	if (p->keyType != TPA_KEY_TYPE_SMK) {
		TRACE_MSG_ERROR("root in the tree is not SMK");
		goto error;
	} else {
		if (Tpa_KEY_LoadStorageMasterKey(tpm, p) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("Tpa_KEY_LoadStorageMasterKey failed");
			goto error;
		}
		p = p->childKey;
	}

	while (p != tail) {
		if (p->keyType != TPA_KEY_TYPE_STORAGE) {
			TRACE_MSG_ERROR("one node in the tree is not a storage key");
			goto error;
		} else {
			if (Tpa_KEY_LoadStorageStorageKey(tpm, p) != TPA_SUCCESS) {
				TRACE_MSG_ERROR("Tpa_KEY_LoadStorageMasterKey failed");
				goto error;
			}
			if ((p = p->childKey) == NULL) {
				TRACE_MSG_ERROR("tree is malformed");
				goto error;
			}
		}
	}
# endif
}

TPA_RESULT Tpa_KEY_LoadTpmBindingKey(TPA_TPM *tpm, TPA_TPMKEY *key) {
	TPA_RESULT result = TPA_ERROR;
	TPA_TSS_ACCESS *tss = NULL;

#ifndef __NO_PARAM_CHECK
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("tpm is NULL");
		return TPA_ERROR;
	}
	if (key == NULL) {
		TRACE_MSG_PARAM_ERROR("key is NULL");
		return TPA_ERROR;
	}
	if (key->keySecret == NULL) {
		TRACE_MSG_PARAM_ERROR("keySecret is NULL");
		return TPA_ERROR;
	}
	if (key->keySecret_len == 0) {
		TRACE_MSG_PARAM_ERROR("keySecret_size is 0");
		return TPA_ERROR;
	}
#endif

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	/******************************************************************************
	 *******************************************************************************
	 *******************************************************************************
	 *				 THIS IS A COMMON PART - TODO: REMOVE						  *
	 *******************************************************************************
	 *******************************************************************************
	 ******************************************************************************/
	/* Load SRK Key */
	if (Tpa_TSS_LoadSRK(tss, tpm->srkSecret, tpm->srkSecret_len) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}

	if (key->parentKey->keyType != TPA_KEY_TYPE_SMK) {
		TRACE_MSG_ERROR("root in the tree is not SMK");
		goto error;
	}

	if (Tpa_TSS_LoadStorageMasterKey(tss, key->parentKey->blob,
			key->parentKey->blob_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}
	/******************************************************************************
	 *******************************************************************************
	 *******************************************************************************
	 *									END										  *
	 *******************************************************************************
	 *******************************************************************************
	 ******************************************************************************/

	if (Tpa_TSS_LoadBindingKey(tss, key->blob, key->blob_size, key->keySecret,
			key->keySecret_len, NULL, (TPA_TSS_KEY **) &key->tss_info)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_CreateBindingKey failed");
		goto error;
	}

	/* Extract key blob */
	if (Tpa_TSS_GetKeyBlob(tss, ((TPA_TSS_KEY *) key->tss_info)->hKey,
			&(key->blob), &(key->blob_size)) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetKeyBlob failed");
		goto error;
	}

	/* Extract public key */
	if (Tpa_TSS_GetKeyModulus(tss, ((TPA_TSS_KEY *) key->tss_info)->hKey,
			&(key->pubKey), &(key->pubKey_size)) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetKeyBlob failed");
		goto error;

	}

	result = TPA_SUCCESS;

	error: /* Free! */
	if (key->tss_info) {
		Tpa_TSS_KeyFreeMemory(tss, (TPA_TSS_KEY *) key->tss_info);
		key->tss_info = NULL;
	}
	if (tss)
		Tpa_TSS_Cleanup(tss);

	return result;
}

TPA_RESULT Tpa_KEY_CreateTpmSealingKey(TPA_TPM *tpm, TPA_TPMKEY *key) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT Tpa_KEY_LoadTpmSealingKey(TPA_TPM *tpm, TPA_TPMKEY *key) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT Tpa_KEY_CreateTpmStorageKey(TPA_TPM *tpm, TPA_TPMKEY *key) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT Tpa_KEY_LoadTpmStorageKey(TPA_TPM *tpm, TPA_TPMKEY *key) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT Tpa_KEY_CreateTpmSignKey(TPA_TPM *tpm, TPA_TPMKEY *key) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT Tpa_KEY_LoadTpmSignKey(TPA_TPM *tpm, TPA_TPMKEY *key) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT Tpa_KEY_CreateTpmAttestationKey(TPA_TPM *tpm, TPA_TPMKEY *key) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT Tpa_KEY_LoadTpmAttestationKey(TPA_TPM *tpm, TPA_TPMKEY *key) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT Tpa_KEY_CreateTpmWrappingKey(TPA_TPM *tpm, TPA_TPMKEY *key) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT Tpa_KEY_LoadTpmWrappingKey(TPA_TPM *tpm, TPA_TPMKEY *key) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT Tpa_KEY_CreateTpmStorageMasterKey(TPA_TPM *tpm,
		unsigned char **mkBlob, unsigned int *mkBlob_size) {
	return TPA_NOT_IMPLEMENTED;
}

TPA_RESULT Tpa_KEY_LoadTpmStorageMasterKey(TPA_TPM *tpm, TPA_TPMKEY *key) {
	return TPA_NOT_IMPLEMENTED;
}
