/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politectnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>
#include <string.h>
#include <dlfcn.h>
#include <stdint.h>
#include <arpa/inet.h>

#include <openssl/evp.h>

#include "TPA_Measure.h"
#include "TPA_LS.h"
#include "TPA_PCR.h"

/**
 * Allocates a ::TPA_MODULE_MEASURE object.
 *
 * Optionally uses as local storage a ::TPA_LS object given in input, or
 * allocate a new one (useful to share the local storage with other modules).
 *
 * @param mod a double pointer to a ::TPA_MODULE_MEASURE object
 * @param ls a pointer to a ::TPA_LS object, set as local storage if not NULL
 */
TPA_RESULT Tpa_Measure_allocateModule(TPA_MODULE_MEASURE **mod, TPA_LS *ls) {
	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (mod == NULL) {
		TRACE_MSG_PARAM_ERROR("mod is NULL");
		return TPA_ERROR;
	}
#endif

	*mod = (TPA_MODULE_MEASURE *) calloc(1, sizeof(TPA_MODULE_MEASURE));
	if (*mod == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}

	if (ls != NULL) {
		(*mod)->ls = ls;
	} else {
		if ((result = Tpa_LS_allocate(&(*mod)->ls)) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("Tpa_LS_allocate failed");
			goto error;
		}
	}

	if ((result = Tpa_PCR_allocate(&(*mod)->targetPcr)) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_PCR_allocate failed");
		goto error;
	}

	return TPA_SUCCESS;
	error: /* error */
	if (*mod != NULL) {
		Tpa_Measure_freeModuleMemory(*mod);
		*mod = NULL;
	}

	return TPA_ERROR;
}

/**
 * Frees a ::TPA_MODULE_MEASURE object
 *
 * @param mod a pointer to a ::TPA_MODULE_MEASURE object
 */
void Tpa_Measure_freeModuleMemory(TPA_MODULE_MEASURE *mod) {
	if (mod == NULL)
		return;

	if (mod->ls) {
		Tpa_LS_freeMemory(mod->ls);
		mod->ls = NULL;
	}

	if (mod->targetPcr) {
		Tpa_PCR_freeMemory(mod->targetPcr, 1);
		mod->targetPcr = NULL;
	}

	free(mod);
}

/*
 * Measure
 */

/**
 * Allocates a ::TPA_MEASURE object.
 *
 * @param measure a double pointer to a ::TPA_MEASURE object
 */
TPA_RESULT Tpa_Measure_allocate(TPA_MEASURE **measure) {

#ifndef __NO_PARAM_CHECK
	if (measure == NULL) {
		TRACE_MSG_PARAM_ERROR("measure is NULL");
		return TPA_ERROR;
	}
	if (*measure != NULL) {
		TRACE_MSG_PARAM_ERROR("*measure is not NULL");
		return TPA_ERROR;
	}
#endif

	*measure = malloc(sizeof(TPA_MEASURE));
	if (*measure == NULL) {
		return TPA_ERROR;
	}

	(*measure)->type = 0;
	(*measure)->value = NULL;
	(*measure)->value_size = 0;
	(*measure)->label = NULL;
	(*measure)->label_len = 0;
	(*measure)->pcrBefore = NULL;
	(*measure)->timestamp = 0;

	return TPA_SUCCESS;
}

/**
 * Frees a ::TPA_MEASURE object.
 *
 * @param measure a pointer to a ::TPA_MEASURE object
 * @param do_free a boolean to specify if invoke the free() on the object or not
 */
void Tpa_Measure_freeMemory(TPA_MEASURE *measure, int do_free) {
#ifndef __NO_PARAM_CHECK
	if (measure == NULL) {
		TRACE_MSG_PARAM_ERROR("measure is NULL");
		return;
	}
#endif
	if (measure->value)
		free(measure->value);
	measure->value = NULL;
	if (measure->label)
		free(measure->label);
	measure->label = NULL;
	if (measure->pcrBefore)
		Tpa_PCR_freeMemory(measure->pcrBefore, 1);
	measure->pcrBefore = NULL;

	measure->type = 0;
	measure->value_size = 0;
	measure->label_len = 0;
	measure->timestamp = 0;

	if (do_free)
		free(measure);
}

/*
 * TPA
 * distinguish between for BIG or SMALL files
 * SMALL files maybe pre and/or post-processed - relying on Tpa_Measure_Blob()
 *
 * IMA
 * WILL distinguish between files that need pre-processing or not.
 * If not needed, set mod->targetPcr = NULL such that no additional (unnecessary)
 * PCR extension will be done, prepare the measure and call Tpa_Measure_Create().
 * Such logs are stored in the LS, perhaps reported, their integrity is only
 * guaranteed by TPA if the additional PCR extension is done.
 * If pre-processing is needed, then call Tpa_Measure_Blob().
 * Measure done by IMA is probably meaningless, thus we have to store our measure
 * in a PCR.
 */
TPA_RESULT Tpa_Measure_File(TPA_MODULE_MEASURE *mod, int type,
		const char *filename, char *label, unsigned int label_len,
		TPA_MEASURE **measure) {

	TPA_RESULT result = TPA_ERROR;

	FILE *fp = NULL;
	int file_size = 0;
	const int MAXSIZE = 4 * 1024 * 1024;

	unsigned char *blob = NULL;
	unsigned int blob_size = 0;

	EVP_MD_CTX mdctx = { 0 };
	unsigned char value[EVP_MAX_MD_SIZE] = { 0 };
	unsigned int value_size = 0;

	const EVP_MD *md = EVP_sha1();

#ifndef __NO_PARAM_CHECK
	if (mod == NULL) {
		TRACE_MSG_PARAM_ERROR("drv is NULL");
		return TPA_ERROR;
	}
	if (filename == NULL) {
		TRACE_MSG_PARAM_ERROR("filename is NULL");
		return TPA_ERROR;
	}
	/* all other checks are left to Tpa_Measure_Create() */
#endif

	/* open file */
	fp = fopen(filename, "rb");
	if (fp == NULL) {
		TRACE_MSG_ERROR("fopen failed");
		goto error;
	}

	/* label == NULL -> filename */
	if (label == NULL) {
		label = (char *) filename;
		label_len = strlen(filename);
	}

	/* read file size */
	fseek(fp, 0, SEEK_END);
	file_size = ftell(fp);
	fseek(fp, 0, SEEK_SET);

	/* distinguish between small and big files */
	if (file_size < MAXSIZE) {
		/* small files:
		 * - read the full content
		 * - create the measure
		 * - postprocess the file, e.g. execute it
		 */

		/* read file */
		blob = malloc(file_size);
		blob_size = file_size;
		if (blob == NULL) {
			TRACE_MSG_ERROR("malloc failed");
			goto error;
		}
		if ((fread(blob, blob_size, 1, fp)) != 1) {
			TRACE_MSG_ERROR("fread failed");
			goto error;
		}

		/* create measure */
		if (Tpa_Measure_Blob(mod, type, blob, blob_size, label, label_len,
				measure) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("Tpa_Measure_Blob failed");
			goto error;
		}

		/* postprocessing */
		fclose(fp); /* close file to avoid multiple open */
		fp = NULL;
		if (mod->postprocessing_callback != NULL) {
			if (mod->postprocessing_callback(mod, (char *) filename)
					!= TPA_SUCCESS) {
				TRACE_MSG_ERROR("postprocessing_callback failed");
				goto error;
			}
		}

	} else {
		/* big files: just hash the content and create the measure */
		blob = malloc(MAXSIZE);
		if (blob == NULL) {
			TRACE_MSG_ERROR("malloc failed");
			goto error;
		}

		/* init hash */
		EVP_MD_CTX_init(&mdctx);
		if (EVP_DigestInit(&mdctx, md) != 1) {
			TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestInit' failed");
			goto error;
		}

		/* loop reading+hashing */
		while ((blob_size = fread(blob, 1, MAXSIZE, fp)) > 0) {
			if (EVP_DigestUpdate(&mdctx, blob, blob_size) != 1) {
				TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestUpdate' failed");
				goto error;
			}
		}

		/* compute hash value */
		if (EVP_DigestFinal(&mdctx, value, &value_size) != 1) {
			TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestFinal' failed");
			goto error;
		}

		/* create measure */
		if (Tpa_Measure_Create(mod, type, value, value_size, label, label_len,
				measure) != TPA_SUCCESS) {
			TRACE_MSG_ERROR("Tpa_Measure_Create failed");
			goto error;
		}
	}

	result = TPA_SUCCESS;

	error: /* free */
	if (fp)
		fclose(fp);
	EVP_MD_CTX_cleanup(&mdctx);
	if (blob)
		free(blob);

	return result;
}

TPA_RESULT Tpa_Measure_Blob(TPA_MODULE_MEASURE *mod, int type,
		unsigned char *blob, unsigned int blob_size, char *label,
		unsigned int label_len, TPA_MEASURE **measure) {

	TPA_RESULT result = TPA_ERROR;

	EVP_MD_CTX mdctx = { 0 };
	unsigned char value[EVP_MAX_MD_SIZE] = { 0 };
	unsigned int value_size = 0;

	const EVP_MD *md = EVP_sha1();

#ifndef __NO_PARAM_CHECK
	if (mod == NULL) {
		TRACE_MSG_PARAM_ERROR("drv is NULL");
		return TPA_ERROR;
	}
	if (blob == NULL && blob_size > 0) {
		TRACE_MSG_PARAM_ERROR("blob_size > 0  but blob is NULL");
		return TPA_ERROR;
	}
	/* all other checks are left to Tpa_Measure_Create() */
#endif

	if (blob == NULL) {
		/* extend with all 0 */
		value_size = 20;
	} else {

		/* preprocessing */
		if (mod->preprocessing_callback != NULL) {
			if (mod->preprocessing_callback(mod, blob, blob_size)
					!= TPA_SUCCESS) {
				TRACE_MSG_ERROR("preprocessing_callback failed");
				goto error;
			}
		}

		/* compute hash value */
		EVP_MD_CTX_init(&mdctx);
		if (EVP_DigestInit(&mdctx, md) != 1) {
			TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestInit' failed");
			goto error;
		}
		if (EVP_DigestUpdate(&mdctx, blob, blob_size) != 1) {
			TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestUpdate' failed");
			goto error;
		}
		if (EVP_DigestFinal(&mdctx, value, &value_size) != 1) {
			TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestFinal' failed");
			goto error;
		}
	}

	/* create measure */
	if (Tpa_Measure_Create(mod, type, value, value_size, label, label_len,
			measure) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_Measure_Create failed");
		goto error;
	}

	result = TPA_SUCCESS;

	error: /* free */
	EVP_MD_CTX_cleanup(&mdctx);

	return result;
}

/*
 * TPA
 * create the measure object, save it in LS, extend the PCR
 * According to TSS specs, we serialize label and additional info into event,
 * and we extend the pcr with:
 *    digest = SHA-1( pcr || value || type || event )
 * Retrieve is only done with TPA internal LS, not with TSS (to discourage
 * using this module)
 *
 * TSS
 * similar to TPA, but create a TSS_PCR_EVENT and relies on TCS for store.
 * This is better because measures from distinct applications are all
 * stored in the same place, so all the history of the PCR is available
 * to all the applications.
 * Retrieve is also done with the TSS.
 *
 * IMA
 * it maybe make sense to avoid duplicate extension, so mod->targetPcr
 * maybe also accepted NULL.
 * Retrieve is done with the TSS module, as it support IMA.
 */
TPA_RESULT Tpa_Measure_Create(TPA_MODULE_MEASURE *mod, int type,
		unsigned char *value, unsigned int value_size, char *label,
		unsigned int label_len, TPA_MEASURE **measure) {

	TPA_RESULT result = TPA_ERROR;

	EVP_MD_CTX mdctx = { 0 };
	unsigned char blobPcr[EVP_MAX_MD_SIZE] = { 0 };
	unsigned int blobPcr_size = 0;
	unsigned char digest[EVP_MAX_MD_SIZE] = { 0 };
	unsigned int digest_size;
	unsigned char *event = NULL;
	unsigned int event_size = 0;
	int critical = 0;
	TPA_MEASURE *local_measure = NULL;
	uint32_t u32 = 0;

	const EVP_MD *md = EVP_sha1();

#ifndef __NO_PARAM_CHECK
	if (mod == NULL) {
		TRACE_MSG_PARAM_ERROR("drv is NULL");
		return TPA_ERROR;
	}
	if (mod->targetPcr == NULL) {
		TRACE_MSG_PARAM_ERROR("drv->targetPcr is NULL");
		return TPA_ERROR;
	}
	if (mod->ls == NULL) {
		TRACE_MSG_PARAM_ERROR("drv->ls is NULL");
		return TPA_ERROR;
	}
	if (value == NULL) {
		TRACE_MSG_PARAM_ERROR("value is NULL");
		return TPA_ERROR;
	}
	if (value_size == 0) {
		TRACE_MSG_PARAM_ERROR("value_size is 0");
		return TPA_ERROR;
	}
	if (value_size != 20) {
		TRACE_MSG_PARAM_ERROR("value_size is not 20");
		return TPA_ERROR;
	}
	if (label == NULL && label_len > 0) {
		TRACE_MSG_PARAM_ERROR("label_len > 0  but label is NULL");
		return TPA_ERROR;
	}
	/* measure can be NULL: the caller is not interested in having back the measure */
#endif

	if (measure != NULL) {
		local_measure = *measure;
	}

	/* allocate measure */
	if (Tpa_Measure_allocate(&local_measure) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_Measure_Allocate failed");
		goto error;
	}

	/* type */
	local_measure->type = type;

	/* value */
	local_measure->value_size = value_size;
	local_measure->value = malloc(local_measure->value_size);
	if (local_measure->value == NULL) {
		TRACE_MSG_ERROR("malloc failed (value)");
		goto error;
	}
	memcpy(local_measure->value, value, local_measure->value_size);

	/* label */
	local_measure->label_len = label_len;
	if (label_len > 0) {
		local_measure->label = strndup(label, label_len);
		if (local_measure->label == NULL) {
			TRACE_MSG_ERROR("strndup failed (label)");
			goto error;
		}
	} else {
		local_measure->label = NULL;
	}

	/* pcr (before) */
	if (Tpa_PCR_allocate(&local_measure->pcrBefore) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_PCR_allocate failed");
		goto error;
	}
	local_measure->pcrBefore->pcr = mod->targetPcr->pcr;
	if (Tpa_PCR_Read(local_measure->pcrBefore) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_PCR_Read failed");
		goto error;
	}
	//TODO we might check if drv->targetPcr is equal to pcrBefore

	/* timestamp is provided by LS */

	/* store measurement in local storage */
	if (Tpa_LS_Open(mod->ls) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_LS_Open failed");
		goto error;
	}

	if (Tpa_LS_MeasureStore(mod->ls, local_measure) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_LS_MeasureStore failed");
		goto error;
	}

	critical = 1;
	/* if something bad happens from now until PCR extend,
	 * we have to delete the measure from LS */

	if (Tpa_LS_Close(mod->ls) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_LS_Close failed");
		goto error;
	}

	/* create event = label + info */
	event_size = label_len;
	event = (unsigned char *) strndup(label, label_len);

	/* create measure digest */
	EVP_MD_CTX_init(&mdctx);
	if (EVP_DigestInit(&mdctx, md) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestInit' failed");
		goto error;
	}
	u32 = htonl(mod->targetPcr->pcr);
	if (EVP_DigestUpdate(&mdctx, &u32, sizeof(uint32_t)) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestUpdate' (PCR) failed");
		goto error;
	}
	if (EVP_DigestUpdate(&mdctx, value, value_size) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestUpdate' (value) failed");
		goto error;
	}
	u32 = TPA_MEASURE_TYPE_TPA | htonl(type);
	if (EVP_DigestUpdate(&mdctx, &u32, sizeof(uint32_t)) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestUpdate' (type) failed");
		goto error;
	}
	if (EVP_DigestUpdate(&mdctx, event, event_size) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestUpdate' (event) failed");
		goto error;
	}
	if (EVP_DigestFinal(&mdctx, digest, &digest_size) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestFinal' failed");
		goto error;
	}
	EVP_MD_CTX_cleanup(&mdctx);

	/* perform pcr_extend */
	if (Tpa_PCR_ExtendH(mod->targetPcr, digest, digest_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_PCR_Extend failed");
		goto error;
	}

	/* PCR is now extended, we must no longer delete the measure from LS */
	critical = 0;

	/* verify extension */
	EVP_MD_CTX_init(&mdctx);
	if (EVP_DigestInit(&mdctx, md) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestInit' failed");
		goto error;
	}
	if (EVP_DigestUpdate(&mdctx, local_measure->pcrBefore->value,
			local_measure->pcrBefore->pcr_size) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestUpdate' (pcrBefore->value) failed");
		goto error;
	}
	if (EVP_DigestUpdate(&mdctx, digest, digest_size) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestUpdate' (value) failed");
		goto error;
	}
	if (EVP_DigestFinal(&mdctx, blobPcr, &blobPcr_size) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestFinal' failed");
		goto error;
	}
	if ((blobPcr_size != mod->targetPcr->pcr_size) || (memcmp(blobPcr,
			mod->targetPcr->value, blobPcr_size) != 0)) {
		TRACE_MSG_ERROR("PCR Extension failed: value after extension is different than expected");
		goto error;
	}

	result = TPA_SUCCESS;

	error: /* free */
	EVP_MD_CTX_cleanup(&mdctx);
	if (event != NULL)
		free(event);
	/* free local_measure if it is not passed back to the caller */
	if (measure == NULL)
		Tpa_Measure_freeMemory(local_measure, 1);
	/* a critical error occured - delete measure from LS */
	if (critical)
		; //TODO

	return result;
}

/*
 * Delete all measures from Local Storage
 */
TPA_RESULT Tpa_Measure_DeleteAll(TPA_MODULE_MEASURE *mod) {

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (mod == NULL) {
		TRACE_MSG_PARAM_ERROR("drv is NULL");
		return TPA_ERROR;
	}
	if (mod->ls == NULL) {
		TRACE_MSG_PARAM_ERROR("drv->ls is NULL");
		return TPA_ERROR;
	}
#endif

	/* store measurement in local storage */
	if (Tpa_LS_Open(mod->ls) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_LS_Open failed");
		goto error;
	}

	if (Tpa_LS_MeasureDeleteAll(mod->ls) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_LS_MeasureDeleteAll failed");
		goto error;
	}

	if (Tpa_LS_Close(mod->ls) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_LS_Close failed");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* free */
	return result;
}

/*
 * callbacks
 */
TPA_RESULT Tpa_Measure_Callback_dlopen(TPA_MODULE_MEASURE *drv, char *filename) {
	drv->ctx_data = dlopen(filename, *((int *) drv->ctx_data));
	return TPA_SUCCESS;
}

TPA_RESULT Tpa_Measure_Callback_fopen(TPA_MODULE_MEASURE *drv, char *filename) {
	drv->ctx_data = fopen(filename, (const char *) drv->ctx_data);
	return TPA_SUCCESS;
}

TPA_RESULT Tpa_MeasureSet_VerifyPcrs(TPA_MEASURE_SET *measuresIn,
		TPA_PCR_SET *selectedPcrs, TPA_MEASURE_SET **measuresOut,
		TPA_VERIFICATION_RESULT *vres) {
	return TPA_NOT_IMPLEMENTED;
}
