/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include "TPA_PCR.h"
#include "TPA_TSS.h"
#include <string.h>
#include <arpa/inet.h>

/**
 * Allocates a ::TPA_PCR object.
 *
 * @param pcr a double pointer to a ::TPA_PCR object
 */
TPA_RESULT Tpa_PCR_allocate(TPA_PCR **pcr) {

#ifndef __NO_PARAM_CHECK
	if (pcr == NULL) {
		TRACE_MSG_PARAM_ERROR("pcr is NULL");
		return TPA_ERROR;
	}
	if (*pcr != NULL) {
		TRACE_MSG_PARAM_ERROR("*pcr is not NULL");
		return TPA_ERROR;
	}
#endif

	*pcr = malloc(sizeof(TPA_PCR));
	if (*pcr == NULL) {
		return TPA_ERROR;
	}

	(*pcr)->pcr = -1;
	(*pcr)->pcr_size = 0;
	(*pcr)->value = NULL;

	return TPA_SUCCESS;
}

/**
 * Frees a ::TPA_PCR object.
 *
 * @param measure a pointer to a ::TPA_PCR object
 * @param do_free a boolean to specify if invoke the free() on the object or not
 */
void Tpa_PCR_freeMemory(TPA_PCR *pcr, int do_free) {
#ifndef __NO_PARAM_CHECK
	if (pcr == NULL) {
		TRACE_MSG_PARAM_ERROR("pcr is NULL");
		return;
	}
#endif
	if (pcr->value)
		free(pcr->value);
	pcr->value = NULL;

	pcr->pcr = 0;
	pcr->pcr_size = 0;

	if (do_free)
		free(pcr);
}



TPA_RESULT Tpa_PCR_Read(TPA_PCR *pcr) {

	TPA_RESULT result = TPA_ERROR;
	TPA_TSS_ACCESS *tss = NULL;

#ifndef __NO_PARAM_CHECK
	if (pcr == NULL) {
		TRACE_MSG_PARAM_ERROR("pcr is NULL");
		return TPA_ERROR;
	}
	if (pcr->pcr < TPA_PCR_NUM_MIN || pcr->pcr > TPA_PCR_NUM_MAX) {
		TRACE_MSG_PARAM_ERROR("pcr->pcr is not a valid PCR number");
		return TPA_ERROR;
	}
#endif

#ifndef __NO_PARAM_CHECK
	if (pcr == NULL) {
		TRACE_MSG_PARAM_ERROR("pcr is NULL");
		return TPA_ERROR;
	}
	if (pcr->pcr < TPA_PCR_NUM_MIN || pcr->pcr > TPA_PCR_NUM_MAX) {
		TRACE_MSG_PARAM_ERROR("pcr->pcr is not a valid PCR number");
		return TPA_ERROR;
	}
#endif

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	if (Tpa_TSS_ReadPCR(tss, pcr->pcr, &pcr->value, &pcr->pcr_size)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_ReadPCR failed");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* Free */
	if (tss)
		Tpa_TSS_Cleanup(tss);

	return result;
}

/**
 * Extends a PCR described by a ::TPA_PCR object.
 *
 * @param pcr a pointer to the ::TPA_PCR object to extend
 * @param digest a digest containing the data to use for the extension
 * @param digest_size an integer containing the length of the digest
 */
TPA_RESULT Tpa_PCR_Extend(TPA_PCR *pcr, unsigned char *blob,
		unsigned int blob_size) {

	TPA_RESULT result = TPA_ERROR;
	TPA_TSS_ACCESS *tss = NULL;
	unsigned char *hash = NULL;
	unsigned int hash_size = 0;

#ifndef __NO_PARAM_CHECK
	if (pcr == NULL) {
		TRACE_MSG_PARAM_ERROR("pcr is NULL");
		goto error;
	}
	if (pcr->pcr < 0 || pcr->pcr > 23) {
		TRACE_MSG_PARAM_ERROR("pcr value not valid");
		goto error;
	}
	if (pcr->value != NULL) {
		TRACE_MSG_PARAM_ERROR("pcr->value must be null");
		goto error;
	}
	if (blob == NULL) {
		TRACE_MSG_PARAM_ERROR("blob is NULL");
		goto error;
	}
#endif

	if (Tpa_BLOB_Digest(blob, blob_size, &hash, &hash_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TpaHL_BLOB_Digest failed");
		goto error;
	}

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	if (Tpa_TSS_ExtendPCR(tss, pcr->pcr, hash, hash_size, &pcr->value,
			&pcr->pcr_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_ReadPCR failed");
		goto error;
	}

	result = TPA_SUCCESS;
	if (tss)
		Tpa_TSS_Cleanup(tss);
	if (hash)
		free(hash);
	return result;

	error: /* Free */
	if (tss)
		Tpa_TSS_Cleanup(tss);
	if (hash)
		free(hash);
	if (pcr->value)
		free(pcr->value);
	pcr->pcr_size = 0;

	return result;
}

TPA_RESULT Tpa_PCR_ExtendH(TPA_PCR *pcr, unsigned char *digest,
		unsigned int digest_size) {

	TPA_RESULT result = TPA_ERROR;
	TPA_TSS_ACCESS *tss = NULL;

#ifndef __NO_PARAM_CHECK
	if (pcr == NULL) {
		TRACE_MSG_PARAM_ERROR("pcr is NULL");
		goto error;
	}
	if (pcr->pcr < 0 || pcr->pcr > 23) {
		TRACE_MSG_PARAM_ERROR("pcr value not valid");
		goto error;
	}
	if (digest == NULL) {
		TRACE_MSG_PARAM_ERROR("blob is NULL");
		goto error;
	}
	if (digest_size != 20) {
		TRACE_MSG_PARAM_ERROR("digest_size must be 20 (only sha1 digest is supported)");
		goto error;
	}
#endif

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	if (Tpa_TSS_ExtendPCR(tss, pcr->pcr, digest, digest_size, &pcr->value,
			&pcr->pcr_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_ReadPCR failed");
		goto error;
	}

	result = TPA_SUCCESS;
	if (tss)
		Tpa_TSS_Cleanup(tss);
	return result;

	error: /* Free */
	if (tss)
		Tpa_TSS_Cleanup(tss);
	if (pcr->value)
		free(pcr->value);
	pcr->pcr_size = 0;

	return result;
}

TPA_RESULT Tpa_PCRSet_Serialize(TPA_PCR_SET *selectedPcrs,
		unsigned char **blob, unsigned int *blob_size) {

	TPA_RESULT result = TPA_ERROR;
	TPA_PCR_INFO selection = { 0 };
	uint8_t mask = 0x01;
	uint8_t tmask = 0x00;
	int i = 0;
	unsigned char *buffer = NULL;
	unsigned char *p = NULL;
	unsigned int buffer_size = 0;

#ifndef __NO_PARAM_CHECK
	if (selectedPcrs == NULL) {
		TRACE_MSG_PARAM_ERROR("selectedPcrs is NULL");
		goto error;
	}
	if (blob == NULL) {
		TRACE_MSG_PARAM_ERROR("blob is NULL");
		goto error;
	}
	if (blob_size == NULL) {
		TRACE_MSG_PARAM_ERROR("blob_size is NULL");
		goto error;
	}
#endif

	selection.sizeOfSelect = 2;
	for (i = 0; i < selectedPcrs->index; i++) {
		if (selectedPcrs->pcrList[i].pcr > 15) {
			selection.sizeOfSelect = 3;
			break;
		}
	}

	selection.pcrSelect = (unsigned char *) calloc(selection.sizeOfSelect,
			sizeof(unsigned char));

	for (i = 0; i < selectedPcrs->index; i++) {
		tmask = mask << ((selectedPcrs->pcrList[i].pcr) % 8);
		selection.pcrSelect[((selectedPcrs->pcrList[i].pcr) / 8)] |= tmask;
	}

	buffer_size = sizeof(uint16_t);
	buffer_size += selection.sizeOfSelect;
	buffer_size += sizeof(uint32_t);
	buffer_size += selectedPcrs->pcrList_size * selectedPcrs->pcr_size;

	buffer = (unsigned char *) calloc(buffer_size, sizeof(unsigned char));
	if (buffer == NULL) {
		TRACE_MSG_ERROR("calloc failed");
		goto error;
	}

	p = buffer;
	//	memcpy(p, &selection.sizeOfSelect, sizeof(selection.sizeOfSelect));
	*(uint16_t *) p = htons(selection.sizeOfSelect);
	p += sizeof(uint16_t);
	memcpy(p, selection.pcrSelect, selection.sizeOfSelect);
	p += selection.sizeOfSelect;

	*(uint32_t *) p = (uint32_t) htonl(selectedPcrs->index
			* selectedPcrs->pcr_size);
	p += sizeof(uint32_t);

	for (i = 0; i < selectedPcrs->index; i++) {
		memcpy(p, selectedPcrs->pcrList[i].value, selectedPcrs->pcr_size);
		p += selectedPcrs->pcr_size;
	}

	*blob = buffer;
	*blob_size = buffer_size;

	result = TPA_SUCCESS;
	error: /* free */
	if (selection.pcrSelect)
		free(selection.pcrSelect);

	return result;
}
