/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politectnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include <stdio.h>
#include <string.h>
#include <openssl/x509.h>

#include "TPA_RA.h"
#include "TPA_PCR.h"
#include "TPA_TSS.h"

TPA_RESULT Tpa_RA_Quote(TPA_TPM *tpm, TPA_RA *raResult) {

	TPA_TSS_ACCESS *tss = NULL;
	unsigned char *dataSigned = NULL;
	unsigned int dataSigned_size = 0;
	int getNonceBack = 0;

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (tpm->srkSecret == NULL) {
		TRACE_MSG_PARAM_ERROR("The TPM srk secret is not set");
		return TPA_ERROR;
	}

	if (raResult == NULL) {
		TRACE_MSG_PARAM_ERROR("raResult is NULL");
		return TPA_ERROR;
	}
	if (raResult->aik == NULL) {
		TRACE_MSG_PARAM_ERROR("raResult->aik is NULL");
		return TPA_ERROR;
	}
	if (raResult->pcrs == NULL) {
		TRACE_MSG_PARAM_ERROR("raResult->pcrs is NULL");
		return TPA_ERROR;
	}
	// raNonce can be NULL: generated by TSS
#endif

	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection to TSS */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	/* Load SRK Key */
	if (Tpa_TSS_LoadSRK(tss, tpm->srkSecret, tpm->srkSecret_len) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadSRK failed");
		goto error;
	}

	/* Load AIK from blob */
	if (Tpa_TSS_LoadAIKkey(tss, raResult->aik->aikSecret,
			raResult->aik->aikSecret_len, raResult->aik->keyBlob,
			raResult->aik->keyBlob_size) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_LoadAIKkey failed");
		goto error;
	}

	/* Set PCR Indexes */
	if (Tpa_TSS_SetPCRIndexes(tss, raResult->pcrs) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_SetPCRIndexes failed");
		goto error;
	}

	/* Set nonce */
	if (raResult->raNonce != NULL && raResult->raNonce_size == 20) {
		//TODO: hash nonce if bigger than 20 bytes
		if (Tpa_TSS_SetDataBlob(tss, raResult->raNonce, raResult->raNonce_size)
				!= TPA_SUCCESS) {
			TRACE_MSG_ERROR("Tpa_TSS_SetDataBlob failed");
			goto error;
		}
	} else {
		getNonceBack = 1;
	}

	/*
	 * Quote
	 */
	if (Tpa_TSS_Quote(tss, &dataSigned, &dataSigned_size, &(raResult->raSign),
			&(raResult->raSign_size)) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Quote failed");
		goto error;
	}

	if (getNonceBack) {
		/* allocated by Tpa_TSS_Quote, will be freed by raResult destructor */
		raResult->raNonce = tss->data;
		raResult->raNonce_size = tss->data_size;
	}

	/* Test correctness of the quote */
	//TODO

	result = TPA_SUCCESS;

	error: /* Free! */
	if (tss != NULL) {
		/* detach tss->data from nonce to avoid error in Tpa_TSS_Cleanup()
		 * no need to free here */
		if (tss->data)
			tss->data = NULL;

		Tpa_TSS_Cleanup(tss);
	}
	if (dataSigned != NULL)
		free(dataSigned);

	return result;
}

TPA_RESULT Tpa_RA_Verify(TPA_RA *ra, TPA_VERIFICATION_RESULT *res) {

	TPA_RESULT result = TPA_ERROR;

	unsigned int blobPcrs_size = 0;
	unsigned char *blobPcrs = NULL;
	unsigned char *tmp = NULL;

	const unsigned char qinfo_head[8] = { 1, 1, 0, 0, 'Q', 'U', 'O', 'T' };
	EVP_MD_CTX mdctx = { 0 };
	X509* aikCert = NULL;
	EVP_PKEY *aikKey = NULL;
	int ver = 0;

	const EVP_MD *md = EVP_sha1();

#ifndef __NO_PARAM_CHECK
	//TODO
#endif

	/* pcrs */
	if (Tpa_PCRSet_Serialize(ra->pcrs, &blobPcrs, &blobPcrs_size)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_PCRSet_Serialize failed");
		goto error;
	}

	EVP_MD_CTX_init(&mdctx);
	if (EVP_DigestInit(&mdctx, md) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestInit' failed");
		goto error;
	}
	if (EVP_DigestUpdate(&mdctx, blobPcrs, blobPcrs_size) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestUpdate' (blobPcrs) failed");
		goto error;
	}
	if (EVP_DigestFinal(&mdctx, blobPcrs, &blobPcrs_size) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_DigestFinal' failed");
		goto error;
	}
	EVP_MD_CTX_cleanup(&mdctx);

	/* nonce */
	if (ra->raNonce_size != 20) {
		//TODO: hash nonce if bigger than 20 bytes
	}

	/* AIK */
	if (ra->aik->pubKey != NULL) {
		// TODO
	} else if (ra->aik->cert != NULL) {
		tmp = ra->aik->cert;
		if (!d2i_X509(&aikCert, (const unsigned char **) &tmp,
				ra->aik->cert_size)) {
			TRACE_MSG_ERROR("OpenSSL call 'd2i_X509' failed");
			goto error;
		}
		aikKey = X509_get_pubkey(aikCert);
		if (aikKey == NULL) {
			TRACE_MSG_ERROR("OpenSSL call 'X509_get_pubkey' failed");
			goto error;
		}
	}

	EVP_MD_CTX_init(&mdctx);
	if (EVP_VerifyInit(&mdctx, md) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_VerifyInit' failed");
		goto error;
	}

	/* build the TPM_QUOTE_INFO data structure */
	/* qinfo header */
	if (EVP_VerifyUpdate(&mdctx, qinfo_head, sizeof(qinfo_head)) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_VerifyUpdate' (qinfo_head) failed");
		goto error;
	}
	/* hash of blobPcrs */
	if (EVP_VerifyUpdate(&mdctx, blobPcrs, blobPcrs_size) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_VerifyUpdate' (blobPcrs) failed");
		goto error;
	}
	/* nonce */
	if (EVP_VerifyUpdate(&mdctx, ra->raNonce, ra->raNonce_size) != 1) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_VerifyUpdate' (raNonce) failed");
		goto error;
	}

	/* verification */
	ver = EVP_VerifyFinal(&mdctx, ra->raSign, ra->raSign_size, aikKey);
	if (ver == 1)
		*res = TPA_VERIFICATION_OK;
	else if (ver == 0)
		*res = TPA_VERIFICATION_NOK;
	else {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_VerifyFinal' failed");
		goto error;
	}

	result = TPA_SUCCESS;

	error: /* Free! */
	EVP_MD_CTX_cleanup(&mdctx);
	if (blobPcrs != NULL)
		free(blobPcrs);
	if (aikKey != NULL)
		EVP_PKEY_free(aikKey);
	if (aikCert != NULL)
		X509_free(aikCert);
	return result;
}
