/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include "TPA_TPM.h"
#include "TPA_TSS.h"
#include <string.h>

TPA_RESULT Tpa_TPM_IsOwnershipTaken(TPA_TPM *tpm,
		TPA_TPM_OWNERSHIP_RESULT *isTaken) {
	TPA_TSS_ACCESS *tss = NULL;
	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (isTaken == NULL) {
		TRACE_MSG_PARAM_ERROR("isTaken is NULL");
		return TPA_ERROR;
	}

	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("TPA_TPM structure is NULL");
		return TPA_ERROR;
	}

#endif
	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	if (Tpa_TSS_IsOwnershipTaken(tss, isTaken) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_IsOwnershipTaken failed");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* Free! */
	if (!tss)
		Tpa_TSS_Cleanup(tss);
	return result;

}

TPA_RESULT Tpa_TPM_TakeOwnership(TPA_TPM *tpm) {
	TPA_TSS_ACCESS *tss = NULL;
	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (tpm == NULL) {
		TRACE_MSG_PARAM_ERROR("TPA_TPM structure is NULL");
		return TPA_ERROR;
	}
	if (tpm->ownSecret == NULL) {
		TRACE_MSG_PARAM_ERROR("The TPM owner secret is not set");
		return TPA_ERROR;
	}
	if (tpm->ownSecret_len != strlen(tpm->ownSecret)) {
		TRACE_MSG_PARAM_ERROR("ownSecret_len in TPA_TPM structure is not correct");
		return TPA_ERROR;
	}
	if (tpm->srkSecret == NULL) {
		TRACE_MSG_PARAM_ERROR("The TPM srk secret is not set");
		return TPA_ERROR;
	}
	if (tpm->srkSecret_len != strlen(tpm->srkSecret)) {
		TRACE_MSG_PARAM_ERROR("srkSecret_len in TPA_TPM structure is not correct");
		return TPA_ERROR;
	}
#endif
	/* Allocate TSS context */
	if (Tpa_TSS_Access_Allocate(&tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Access_Allocate failed");
		goto error;
	}

	/* Create connection */
	if (Tpa_TSS_Connect(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_Connect failed");
		goto error;
	}

	/* Set owner secret from TPM structure */
	if (Tpa_TSS_SetOwnerSecret(tss, tpm->ownSecret, tpm->ownSecret_len)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_SetOwnerSecret failed");
		goto error;
	}

	if (Tpa_TSS_CreateSRKObject(tss, tpm->srkSecret, tpm->srkSecret_len)
			!= TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_CreateSRKObject failed");
		goto error;
	}

	if (Tpa_TSS_TakeOwnership(tss) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_TakeOwnership failed");
		goto error;
	}

	result = TPA_SUCCESS;
	error: /* Free! */
	if (!tss)
		Tpa_TSS_Cleanup(tss);
	return result;

}
