/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include <string.h>
#include <stdio.h>

#include "TPA_TSS.h"

#include <netinet/in.h>
#include <openssl/bn.h>
#include <openssl/x509.h>

#define TPMSEAL_SECRET "default_secret"

/*
 * This code has been **strongly** revisited.
 * Don't add untrusted lines of code.
 *
 * DV stay away from this file!!
 */

TPA_RESULT Tpa_TSS_Access_Allocate(TPA_TSS_ACCESS **tss) {

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}
#endif

	*tss = (TPA_TSS_ACCESS *) malloc(sizeof(TPA_TSS_ACCESS));
	if (*tss == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto end;
	}

	(*tss)->hContext = 0;
	(*tss)->hTPM = 0;
	(*tss)->hSRK = 0;
	(*tss)->hIdentKey = 0;
	(*tss)->hPCAKey = 0;
	(*tss)->hPubEndKey = 0;
	(*tss)->hStorageMK = 0;
	(*tss)->hPcrs = 0;
	(*tss)->data = NULL;
	(*tss)->data_size = 0;

	result = TPA_SUCCESS;
	end: return result;
}

void Tpa_TSS_Cleanup(TPA_TSS_ACCESS *tss) {

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return;
	}
#endif

	if (tss->hContext != 0) {
		// so we have to free piece by piece...
		if (tss->hPubEndKey != 0) {
			Tspi_Context_CloseObject(tss->hContext, tss->hPubEndKey);
		}
		tss->hPubEndKey = 0;
		if (tss->hIdentKey != 0) {
			Tspi_Context_CloseObject(tss->hContext, tss->hIdentKey);
			tss->hIdentKey = 0;
		}
		if (tss->hPCAKey != 0) {
			Tspi_Context_CloseObject(tss->hContext, tss->hPCAKey);
			tss->hPCAKey = 0;
		}
		if (tss->hSRK != 0) {
			Tspi_Context_CloseObject(tss->hContext, tss->hSRK);
			tss->hSRK = 0;
		}
		if (tss->hTPM != 0) {
			Tspi_Context_CloseObject(tss->hContext, tss->hTPM);
			tss->hTPM = 0;
		}
		if (tss->hStorageMK != 0) {
			Tspi_Context_CloseObject(tss->hContext, tss->hStorageMK);
			tss->hStorageMK = 0;
		}
		if (tss->hPcrs != 0) {
			Tspi_Context_CloseObject(tss->hContext, tss->hPcrs);
			tss->hPcrs = 0;
		}

		Tspi_Context_Close(tss->hContext);
		tss->hContext = 0;
		tss->data_size = 0;
	}

	if (tss->data != NULL) {
		TRACE_MSG_PARAM_ERROR("tss->data is not NULL");
	}

	free(tss);
}

TPA_RESULT Tpa_TSS_Connect(TPA_TSS_ACCESS *tss) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}
#endif

	tss_result = Tspi_Context_Create(&(tss->hContext));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_Create failed", tss_result);
		goto end;
	}

	tss_result = Tspi_Context_Connect(tss->hContext, NULL);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_Connect failed", tss_result);
		tss->hContext = 0;
		goto end;
	}

	tss_result = Tspi_Context_GetTpmObject(tss->hContext, &(tss->hTPM));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_GetTpmObject failed", tss_result);
		tss->hTPM = 0;
		/* not sure it is actually needed */
		Tspi_Context_Close(tss->hContext);
		tss->hContext = 0;
		goto end;
	}

	result = TPA_SUCCESS;
	end: return result;
}

TPA_RESULT Tpa_TSS_SetOwnerSecret(TPA_TSS_ACCESS *tss, char *secret,
		unsigned int secret_len) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	TSS_HPOLICY hPolicy = 0;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hTPM is 0");
		goto end;
	}

	if (secret == NULL) {
		TRACE_MSG_PARAM_ERROR("secret is NULL");
		goto end;
	}
#endif

	tss_result = Tspi_GetPolicyObject(tss->hTPM, TSS_POLICY_USAGE, &hPolicy);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_GetPolicyObject failed", tss_result);
		goto end;
	}

	tss_result = Tspi_Policy_SetSecret(hPolicy, TSS_SECRET_MODE_PLAIN,
			secret_len, (BYTE*) secret);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
		goto end;
	}

	result = TPA_SUCCESS;
	end: return result;
}

TPA_RESULT Tpa_TSS_LoadSRK(TPA_TSS_ACCESS *tss, char *secret,
		unsigned int secret_len) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	TSS_HPOLICY hSrkPolicy = 0;
	TSS_UUID SRK_UUID = TSS_UUID_SRK;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (secret == NULL) {
		TRACE_MSG_PARAM_ERROR("secret is NULL");
		goto error;
	}
#endif

	tss_result = Tspi_Context_LoadKeyByUUID(tss->hContext, TSS_PS_TYPE_SYSTEM,
			SRK_UUID, &(tss->hSRK));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_LoadKeyByUUID failed", tss_result);
		goto error;
	}

	tss_result = Tspi_GetPolicyObject(tss->hSRK, TSS_POLICY_USAGE, &hSrkPolicy);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_GetPolicyObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_SetSecret(hSrkPolicy, TSS_SECRET_MODE_PLAIN,
			secret_len, (BYTE*) secret);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
		goto error;
	}

	return TPA_SUCCESS;
	error: if (tss->hSRK != 0) {
		Tspi_Context_CloseObject(tss->hContext, tss->hSRK);
		tss->hSRK = 0;
	}
	return result;
}

TPA_RESULT Tpa_TSS_CreateAIKkey(TPA_TSS_ACCESS *tss, char *secret,
		unsigned int secret_len) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	TSS_HPOLICY hIdentKeyPolicy = 0;
	UINT32 aikInitFlags = TSS_KEY_TYPE_IDENTITY | TSS_KEY_SIZE_2048
			| TSS_KEY_VOLATILE | TSS_KEY_AUTHORIZATION | TSS_KEY_NOT_MIGRATABLE;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (secret == NULL) {
		TRACE_MSG_PARAM_ERROR("secret is NULL");
		goto error;
	}
#endif

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_RSAKEY, aikInitFlags, &(tss->hIdentKey));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_POLICY, TSS_POLICY_USAGE, &hIdentKeyPolicy);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_AssignToObject(hIdentKeyPolicy, tss->hIdentKey);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_AssignToObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_SetSecret(hIdentKeyPolicy, TSS_SECRET_MODE_PLAIN,
			secret_len, (BYTE*) secret);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
		goto error;
	}

	return TPA_SUCCESS;

	error: if (hIdentKeyPolicy != 0)
		Tspi_Context_CloseObject(tss->hContext, hIdentKeyPolicy);
	if (tss->hIdentKey != 0) {
		Tspi_Context_CloseObject(tss->hContext, tss->hIdentKey);
		tss->hIdentKey = 0;
	}

	return result;
}

TPA_RESULT Tpa_TSS_LoadAIKkey(TPA_TSS_ACCESS *tss, char *secret,
		unsigned int secret_len, unsigned char *keyBlob,
		unsigned int keyBlob_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	TSS_HPOLICY hIdentKeyPolicy;
	UINT32 aikInitFlags = TSS_KEY_TYPE_IDENTITY | TSS_KEY_SIZE_2048
			| TSS_KEY_VOLATILE | TSS_KEY_AUTHORIZATION | TSS_KEY_NOT_MIGRATABLE;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (secret == NULL) {
		TRACE_MSG_PARAM_ERROR("secret is NULL");
		goto error;
	}

	if (keyBlob == NULL) {
		TRACE_MSG_PARAM_ERROR("keyBlob is NULL");
		goto error;
	}
#endif

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_RSAKEY, aikInitFlags, &(tss->hIdentKey));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Context_LoadKeyByBlob(tss->hContext, tss->hSRK,
			keyBlob_size, keyBlob, &(tss->hIdentKey));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_LoadKeyByBlob failed", tss_result);
		goto error;
	}

	tss_result = Tspi_GetPolicyObject(tss->hIdentKey, TSS_POLICY_USAGE,
			&hIdentKeyPolicy);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_GetPolicyObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_SetSecret(hIdentKeyPolicy, TSS_SECRET_MODE_PLAIN,
			secret_len, (BYTE*) secret);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
		goto error;
	}

	return TPA_SUCCESS;

	error: if (tss->hIdentKey != 0) {
		Tspi_Context_CloseObject(tss->hContext, tss->hIdentKey);
		tss->hIdentKey = 0;
	}

	return result;
}

static TPA_RESULT Tpa_TSS_SetPCRValues(TPA_TSS_ACCESS *tss,
		TPA_PCR_SET *selectedPcrs, TSS_HPCRS *hPcrs) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	TSS_FLAG initFlag = 0;
	UINT32 localityValue = 0;
	int i = 0;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (selectedPcrs == NULL) {
		TRACE_MSG_PARAM_ERROR("selectedPcrs is NULL");
		goto error;
	}

	if (hPcrs == NULL) {
		TRACE_MSG_PARAM_ERROR("hPcrs is NULL");
		goto error;
	}
#endif

	for (i = 0; i < selectedPcrs->index; i++) {
		if (selectedPcrs->pcrList[i].pcr > 15) {
			initFlag |= TSS_PCRS_STRUCT_INFO_LONG;
			break;
		}
	}

	tss_result = Tspi_Context_CreateObject(tss->hContext, TSS_OBJECT_TYPE_PCRS,
			initFlag, hPcrs);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject PCR Composite failed",
				tss_result);
		goto error;
	}

	for (i = 0; i < selectedPcrs->index; i++) {

		tss_result = Tspi_PcrComposite_SetPcrValue(*hPcrs,
				selectedPcrs->pcrList[i].pcr, selectedPcrs->pcr_size,
				selectedPcrs->pcrList[i].value);
		if (tss_result != TSS_SUCCESS) {
			TRACE_MSG_TSS_ERROR("Tspi_PcrComposite_SetPcrValue failed",
					tss_result);
			goto error;
		}
	}

	if (initFlag) {
		localityValue = TPM_LOC_ZERO | TPM_LOC_ONE | TPM_LOC_TWO
				| TPM_LOC_THREE | TPM_LOC_FOUR;
		tss_result = Tspi_PcrComposite_SetPcrLocality(*hPcrs, localityValue);
		if (tss_result != TSS_SUCCESS) {
			TRACE_MSG_TSS_ERROR("Tspi_PcrComposite_SetPcrLocality failed",
					tss_result);
			goto error;
		}
	}

	return TPA_SUCCESS;

	error: if (*hPcrs != 0) {
		Tspi_Context_CloseObject(tss->hContext, *hPcrs);
		*hPcrs = 0;
	}
	return result;
}

static TPA_RESULT TPA_TSS_GetPubKeyFromCertificate(unsigned char *cert_pointer,
		unsigned int cert_size, unsigned char **public_exponent,
		unsigned int *public_exponent_size, unsigned char **modulus,
		unsigned int *modulus_size, int *keysize) {
	X509* cert;
	EVP_PKEY *pcaKey;
	RSA *rsa = NULL;

	const unsigned char* temp_cert_pointer =
			(const unsigned char*) cert_pointer;
	cert = d2i_X509(NULL, &temp_cert_pointer, cert_size);

	if (!cert) {
		TRACE_MSG_ERROR("OpenSSL call 'd2i_X509' failed!");
		X509_free(cert);
		return TPA_ERROR;
	}

	pcaKey = X509_get_pubkey(cert);

	rsa = EVP_PKEY_get1_RSA(pcaKey);
	if (rsa == NULL) {
		TRACE_MSG_ERROR("OpenSSL call 'EVP_PKEY_get1_RSA' failed!");
		X509_free(cert);
		return TPA_ERROR;
	}

	EVP_PKEY_free(pcaKey);

	*keysize = RSA_size(rsa);

	*modulus = calloc(1, BN_num_bytes(rsa->n));
	*modulus_size = BN_bn2bin(rsa->n, (unsigned char*) *modulus);

	*public_exponent = calloc(1, BN_num_bytes(rsa->e));
	*public_exponent_size
			= BN_bn2bin(rsa->e, (unsigned char*) *public_exponent);

	RSA_free(rsa);

	if (*modulus_size <= 0) {
		TRACE_MSG_ERROR("OpenSSL call 'BN_bn2bin' failed!");
		X509_free(cert);
		return TPA_ERROR;
	}

	X509_free(cert);
	return TPA_SUCCESS;
}

TPA_RESULT Tpa_TSS_CreatePCAkey(TPA_TSS_ACCESS *tss, TPA_PCA *ca) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	UINT32 pcaInitFlags = 0;
	unsigned char *public_exponent = NULL;
	unsigned int public_exponent_size = 0;
	unsigned char *modulus = NULL;
	unsigned int modulus_size = 0;
	int keysize = 0;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (ca == NULL) {
		TRACE_MSG_PARAM_ERROR("ca is NULL");
		goto error;
	}
#endif

	if (TPA_TSS_GetPubKeyFromCertificate(ca->pcaCertificate,
			ca->pcaCertificate_size, &public_exponent, &public_exponent_size,
			&modulus, &modulus_size, &keysize) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("TPA_TSS_GetPubKeyFromCertificate failed");
		goto error;
	}

	/*
	 * NOTE:
	 * The full PrivacyCA public key blob is hashed into the AIK request sent
	 * to the PrivacyCA. The server side must reconstruct an identical blob
	 * to obtain the same hash. Thus, the following parameters of the key
	 * must be exactly the same. Unfortunately, there is no standard for this,
	 * so different AIK cycle and/or TSS implementations may be incompatible.
	 */
	switch (keysize) {
	case 512:
		pcaInitFlags = TSS_KEY_TYPE_LEGACY | TSS_KEY_SIZE_4096;
		break;
	case 256:
		pcaInitFlags = TSS_KEY_TYPE_LEGACY | TSS_KEY_SIZE_2048;
		break;
	case 128:
		pcaInitFlags = TSS_KEY_TYPE_LEGACY | TSS_KEY_SIZE_1024;
		break;
	case 64:
		pcaInitFlags = TSS_KEY_TYPE_LEGACY | TSS_KEY_SIZE_512;
		break;
	default:
		TRACE_MSG_ERROR("Unknown PCA key size");
		goto error;
		break;
	}

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_RSAKEY, pcaInitFlags, &(tss->hPCAKey));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_SetAttribData(tss->hPCAKey, TSS_TSPATTRIB_RSAKEY_INFO,
			TSS_TSPATTRIB_KEYINFO_RSA_MODULUS, modulus_size, (BYTE*) modulus);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_SetAttribData failed", tss_result);
		goto error;
	}

	tss_result = Tspi_SetAttribUint32(tss->hPCAKey, TSS_TSPATTRIB_KEY_INFO,
			TSS_TSPATTRIB_KEYINFO_ENCSCHEME, TSS_ES_RSAESPKCSV15);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_SetAttribUint32 failed", tss_result);
		goto error;
	}

	tss_result = Tspi_SetAttribUint32(tss->hPCAKey, TSS_TSPATTRIB_KEY_INFO,
			TSS_TSPATTRIB_KEYINFO_SIGSCHEME, TSS_SS_NONE);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_SetAttribUint32 failed", tss_result);
		goto error;
	}

	return TPA_SUCCESS;

	error: if (public_exponent)
		free(public_exponent);
	if (modulus)
		free(modulus);
	if (tss->hPCAKey != 0) {
		Tspi_Context_CloseObject(tss->hContext, tss->hPCAKey);
		tss->hPCAKey = 0;
	}

	return result;
}

TPA_RESULT Tpa_TSS_SetEKcert(TPA_TSS_ACCESS *tss, unsigned char *cert,
		unsigned int cert_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hTPM is 0");
		goto end;
	}

	if (cert == NULL) {
		TRACE_MSG_PARAM_ERROR("cert is NULL");
		goto end;
	}
#endif

	tss_result = Tspi_SetAttribData(tss->hTPM, TSS_TSPATTRIB_TPM_CREDENTIAL,
			TSS_TPMATTRIB_EKCERT, cert_size, (BYTE*) cert);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_SetAttribData failed", tss_result);
		goto end;
	}

	result = TPA_SUCCESS;
	end: return result;
}

TPA_RESULT Tpa_TSS_CollateIdentityRequest(TPA_TSS_ACCESS *tss, char *label,
		unsigned int label_len, unsigned char **request,
		unsigned int* request_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	BYTE* pca_request = NULL;
	UINT32 pca_request_size = 0;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hTPM is 0");
		goto error;
	}

	if (tss->hSRK == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hSRK is 0");
		goto error;
	}

	if (tss->hPCAKey == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hPCAKey is 0");
		goto error;
	}

	if (label == NULL) {
		TRACE_MSG_PARAM_ERROR("label is NULL");
		goto error;
	}

	if (request == NULL) {
		TRACE_MSG_PARAM_ERROR("request is NULL");
		goto error;
	}

	if (request_size == NULL) {
		TRACE_MSG_PARAM_ERROR("request_size is NULL");
		goto error;
	}
#endif

	tss_result = Tspi_TPM_CollateIdentityRequest(tss->hTPM, tss->hSRK,
			tss->hPCAKey, label_len, (unsigned char *) label, tss->hIdentKey,
			TSS_ALG_AES, &pca_request_size, &pca_request);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_TPM_CollateIdentityRequest failed",
				tss_result);
		goto error;
	}

	*request = malloc(pca_request_size);
	if (*request == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}

	memcpy(*request, pca_request, pca_request_size);
	*request_size = pca_request_size;

	Tspi_Context_FreeMemory(tss->hContext, pca_request);

	return TPA_SUCCESS;

	error: if (pca_request != NULL) {
		Tspi_Context_FreeMemory(tss->hContext, pca_request);
	}
	if (*request != NULL) {
		free(*request);
		*request = NULL;
		*request_size = 0;
	}

	return result;
}

TPA_RESULT Tpa_TSS_ActivateIdentity(TPA_TSS_ACCESS *tss,
		unsigned char* asymBlob, unsigned int asymBlob_size,
		unsigned char* symBlob, unsigned int symBlob_size,
		unsigned char **credential, unsigned int* credential_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	BYTE* tss_credential = NULL;
	UINT32 tss_credential_size = 0;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hTPM is 0");
		goto error;
	}

	if (tss->hIdentKey == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hIdentKey is 0");
		goto error;
	}

	if (tss->hSRK == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hSRK is 0");
		goto error;
	}

	if (asymBlob == NULL) {
		TRACE_MSG_PARAM_ERROR("asymBlob is NULL");
		goto error;
	}

	if (symBlob == NULL) {
		TRACE_MSG_PARAM_ERROR("symBlob is NULL");
		goto error;
	}

	if (credential == NULL) {
		TRACE_MSG_PARAM_ERROR("credential is NULL");
		goto error;
	}

	if (credential_size == NULL) {
		TRACE_MSG_PARAM_ERROR("credential_size is NULL");
		goto error;
	}
#endif

	tss_result = Tspi_Key_LoadKey(tss->hIdentKey, tss->hSRK);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Key_LoadKey failed", tss_result);
		goto error;
	}

	tss_result = Tspi_TPM_ActivateIdentity(tss->hTPM, tss->hIdentKey,
			asymBlob_size, asymBlob, symBlob_size, symBlob,
			&tss_credential_size, &tss_credential);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_TPM_ActivateIdentity failed", tss_result);
		goto error;
	}

	*credential = malloc(tss_credential_size);
	if (*credential == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}

	memcpy(*credential, tss_credential, tss_credential_size);
	*credential_size = tss_credential_size;

	Tspi_Context_FreeMemory(tss->hContext, tss_credential);

	return TPA_SUCCESS;

	error: if (tss_credential != NULL) {
		Tspi_Context_FreeMemory(tss->hContext, tss_credential);
	}
	if (*credential != NULL) {
		free(*credential);
		*credential = NULL;
		*credential_size = 0;
	}

	return result;
}

TPA_RESULT Tpa_TSS_GetPubEndKey(TPA_TSS_ACCESS *tss) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hTPM is 0");
		goto end;
	}
#endif

	/* PubEndKey already loaded */
	if (tss->hPubEndKey != 0) {
		result = TPA_SUCCESS;
		goto end;
	}

	tss_result = Tspi_TPM_GetPubEndorsementKey(tss->hTPM, TRUE, NULL,
			&(tss->hPubEndKey));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_TPM_GetPubEndorsementKey failed", tss_result);
		goto end;
	}

	result = TPA_SUCCESS;
	end: return result;
}

TPA_RESULT Tpa_TSS_GetKeyBlob(TPA_TSS_ACCESS *tss, TSS_HKEY hKey,
		unsigned char **blob, unsigned int* blob_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	UINT32 data_size = 0;
	BYTE* data = NULL;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (hKey == 0) {
		TRACE_MSG_PARAM_ERROR("hKey is 0");
		goto error;
	}

	if (blob == NULL) {
		TRACE_MSG_PARAM_ERROR("blob is NULL");
		goto error;
	}

	if (blob_size == 0) {
		TRACE_MSG_PARAM_ERROR("blob_size is NULL");
		goto error;
	}
#endif

	tss_result = Tspi_GetAttribData(hKey, TSS_TSPATTRIB_KEY_BLOB,
			TSS_TSPATTRIB_KEYBLOB_BLOB, &data_size, &data);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_GetAttribData failed", tss_result);
		goto error;
	}

	*blob = malloc(data_size);
	if (*blob == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	memcpy(*blob, data, data_size);
	*blob_size = data_size;

	Tspi_Context_FreeMemory(tss->hContext, data);

	return TPA_SUCCESS;

	error: if (data != NULL) {
		Tspi_Context_FreeMemory(tss->hContext, data);
	}
	if (*blob != NULL) {
		free(*blob);
		*blob = NULL;
		*blob_size = 0;
	}

	return result;
}

TPA_RESULT Tpa_TSS_GetKeyModulus(TPA_TSS_ACCESS *tss, TSS_HKEY hKey,
		unsigned char **blob, unsigned int *blob_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	UINT32 data_size = 0;
	BYTE* data = NULL;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (hKey == 0) {
		TRACE_MSG_PARAM_ERROR("hKey is 0");
		goto error;
	}

	if (blob == NULL) {
		TRACE_MSG_PARAM_ERROR("blob is NULL");
		goto error;
	}

	if (blob_size == 0) {
		TRACE_MSG_PARAM_ERROR("blob_size is NULL");
		goto error;
	}
#endif

	tss_result = Tspi_GetAttribData(hKey, TSS_TSPATTRIB_RSAKEY_INFO,
			TSS_TSPATTRIB_KEYINFO_RSA_MODULUS, &data_size, &data);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_GetAttribData failed", tss_result);
		goto error;
	}

	*blob = malloc(data_size);
	if (*blob == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	memcpy(*blob, data, data_size);
	*blob_size = data_size;

	Tspi_Context_FreeMemory(tss->hContext, data);

	return TPA_SUCCESS;

	error: if (data != NULL) {
		Tspi_Context_FreeMemory(tss->hContext, data);
	}
	if (*blob != NULL) {
		free(*blob);
		*blob = NULL;
		*blob_size = 0;
	}

	return result;
}

TPA_RESULT Tpa_TSS_GetKeyExponent(TPA_TSS_ACCESS *tss, TSS_HKEY hKey,
		unsigned char **blob, unsigned int *blob_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	UINT32 data_size = 0;
	BYTE* data = NULL;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (hKey == 0) {
		TRACE_MSG_PARAM_ERROR("hKey is 0");
		goto error;
	}

	if (blob == NULL) {
		TRACE_MSG_PARAM_ERROR("blob is NULL");
		goto error;
	}

	if (blob_size == 0) {
		TRACE_MSG_PARAM_ERROR("blob_size is NULL");
		goto error;
	}
#endif

	tss_result = Tspi_GetAttribData(hKey, TSS_TSPATTRIB_RSAKEY_INFO,
			TSS_TSPATTRIB_KEYINFO_RSA_EXPONENT, &data_size, &data);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_GetAttribData failed", tss_result);
		goto error;
	}

	*blob = malloc(data_size);
	if (*blob == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	memcpy(*blob, data, data_size);
	*blob_size = data_size;

	Tspi_Context_FreeMemory(tss->hContext, data);

	return TPA_SUCCESS;

	error: if (data != NULL) {
		Tspi_Context_FreeMemory(tss->hContext, data);
	}
	if (*blob != NULL) {
		free(*blob);
		*blob = NULL;
		*blob_size = 0;
	}

	return result;
}

TPA_RESULT Tpa_TSS_GetRandom(TPA_TSS_ACCESS *tss, unsigned int blob_size,
		unsigned char **blob) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	BYTE* data = NULL;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hTPM is 0");
		goto error;
	}

	if (blob == NULL) {
		TRACE_MSG_PARAM_ERROR("blob is NULL");
		goto error;
	}
#endif

	tss_result = Tspi_TPM_GetRandom(tss->hTPM, blob_size, &data);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_TPM_GetRandom failed", tss_result);
		goto error;
	}

	*blob = malloc(blob_size);
	if (*blob == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	memcpy(*blob, data, blob_size);

	Tspi_Context_FreeMemory(tss->hContext, data);

	return TPA_SUCCESS;

	error: if (data != NULL) {
		Tspi_Context_FreeMemory(tss->hContext, data);
	}
	return result;
}

TPA_RESULT Tpa_TSS_CreateStorageMasterKey(TPA_TSS_ACCESS *tss) {

	TPA_RESULT result = TPA_ERROR;
	TSS_HPOLICY hPolicy;
	TSS_RESULT tss_result;
	TSS_FLAG keyFlags = TSS_KEY_TYPE_STORAGE | TSS_KEY_SIZE_2048
			| TSS_KEY_VOLATILE | TSS_KEY_AUTHORIZATION | TSS_KEY_NOT_MIGRATABLE;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (tss->hSRK == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hSRK is 0");
		goto error;
	}
#endif

	/* Build an RSA key object that will be created by the TPM
	 (this will encrypt and protect the symmetric key) */
	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_RSAKEY, keyFlags, &(tss->hStorageMK));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_POLICY, TSS_POLICY_USAGE, &hPolicy);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_SetSecret(hPolicy, TSS_SECRET_MODE_PLAIN, strlen(
			TPMSEAL_SECRET), (BYTE *) TPMSEAL_SECRET);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_AssignToObject(hPolicy, tss->hStorageMK);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_AssignToObject failed", tss_result);
		goto error;
	}

	/* Create the RSA key (under the SRK) */
	tss_result = Tspi_Key_CreateKey(tss->hStorageMK, tss->hSRK, 0);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Key_CreateKey failed", tss_result);
		goto error;
	}

	/* Load the newly created RSA key */
	tss_result = Tspi_Key_LoadKey(tss->hStorageMK, tss->hSRK);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Key_LoadKey for hStorageMK failed", tss_result);
		goto error;
	}

	return TPA_SUCCESS;

	error: if (hPolicy != 0)
		Tspi_Context_CloseObject(tss->hContext, hPolicy);
	if (tss->hStorageMK != 0) {
		Tspi_Context_CloseObject(tss->hContext, tss->hStorageMK);
		tss->hStorageMK = 0;
	}

	return result;
}

/* TODO: here we create a new policy and assign it to the key,
 * while in other Load functions we retrieve the policy. Why?
 */
TPA_RESULT Tpa_TSS_LoadStorageMasterKey(TPA_TSS_ACCESS *tss,
		unsigned char *keyBlob, unsigned int keyBlob_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_HPOLICY hPolicy;
	TSS_RESULT tss_result;
	TSS_FLAG keyFlags = TSS_KEY_TYPE_STORAGE | TSS_KEY_SIZE_2048
			| TSS_KEY_VOLATILE | TSS_KEY_AUTHORIZATION | TSS_KEY_NOT_MIGRATABLE;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (keyBlob == NULL) {
		TRACE_MSG_PARAM_ERROR("keyBlob is NULL");
		goto error;
	}

	if (tss->hSRK == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hSRK is 0");
		goto error;
	}
#endif

	/* Build an RSA key object that will be created by the TPM
	 (this will encrypt and protect the symmetric key) */
	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_RSAKEY, keyFlags, &(tss->hStorageMK));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Context_LoadKeyByBlob(tss->hContext, tss->hSRK,
			keyBlob_size, keyBlob, &(tss->hStorageMK));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_LoadKeyByBlob failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_POLICY, TSS_POLICY_USAGE, &hPolicy);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_SetSecret(hPolicy, TSS_SECRET_MODE_PLAIN, strlen(
			TPMSEAL_SECRET), (BYTE *) TPMSEAL_SECRET);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_AssignToObject(hPolicy, tss->hStorageMK);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_AssignToObject failed", tss_result);
		goto error;
	}

	//Tspi_Context_CloseObject(tss->hContext, hPolicy);

	return TPA_SUCCESS;

	error: if (hPolicy != 0)
		Tspi_Context_CloseObject(tss->hContext, hPolicy);
	if (tss->hStorageMK != 0) {
		Tspi_Context_CloseObject(tss->hContext, tss->hStorageMK);
		tss->hStorageMK = 0;
	}

	return result;
}

TPA_RESULT Tpa_TSS_SealBlobWithMK(TPA_TSS_ACCESS *tss,
		TPA_PCR_SET *selectedPcrs, unsigned char *plainblob,
		unsigned int plainblob_size, unsigned char **sealedblob,
		unsigned int *sealedblob_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result = 0;
	TSS_HPCRS hPcrs = 0;
	TSS_HPOLICY hPolicy = 0;
	TSS_HENCDATA hEncData = 0;
	unsigned char *tmp_sealedblob = NULL;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (tss->hStorageMK == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hStorageMK is 0");
		goto error;
	}

	if (selectedPcrs == NULL) {
		TRACE_MSG_PARAM_ERROR("selectedPcrs is NULL");
		goto error;
	}

	if (plainblob == NULL) {
		TRACE_MSG_PARAM_ERROR("plainblob is NULL");
		goto error;
	}

	if (sealedblob == NULL) {
		TRACE_MSG_PARAM_ERROR("sealedblob is NULL");
		goto error;
	}

	if (sealedblob_size == NULL) {
		TRACE_MSG_PARAM_ERROR("sealedblob_size is NULL");
		goto error;
	}
#endif

	if (Tpa_TSS_SetPCRValues(tss, selectedPcrs, &hPcrs) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_SetPCRValues failed");
		goto error;
	}

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_ENCDATA, TSS_ENCDATA_SEAL, &hEncData);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_POLICY, TSS_POLICY_USAGE, &hPolicy);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_SetSecret(hPolicy, TSS_SECRET_MODE_PLAIN, strlen(
			TPMSEAL_SECRET), (BYTE *) TPMSEAL_SECRET);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_AssignToObject(hPolicy, hEncData);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_AssignToObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Data_Seal(hEncData, tss->hStorageMK, plainblob_size,
			plainblob, hPcrs);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Data_Seal failed", tss_result);
		goto error;
	}

	tss_result = Tspi_GetAttribData(hEncData, TSS_TSPATTRIB_ENCDATA_BLOB,
			TSS_TSPATTRIB_ENCDATABLOB_BLOB, sealedblob_size, &tmp_sealedblob)
			!= TSS_SUCCESS;
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_GetAttribData failed", tss_result);
		goto error;
	}

	*sealedblob = (unsigned char *) malloc(*sealedblob_size);
	if (*sealedblob == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	memcpy(*sealedblob, tmp_sealedblob, *sealedblob_size);

	Tspi_Context_CloseObject(tss->hContext, hPcrs);
	Tspi_Context_CloseObject(tss->hContext, hEncData);
	Tspi_Context_CloseObject(tss->hContext, hPolicy);
	Tspi_Context_FreeMemory(tss->hContext, tmp_sealedblob);

	return TPA_SUCCESS;

	error: if (hPcrs != 0)
		Tspi_Context_CloseObject(tss->hContext, hPcrs);
	if (hEncData != 0)
		Tspi_Context_CloseObject(tss->hContext, hEncData);
	if (hPolicy != 0)
		Tspi_Context_CloseObject(tss->hContext, hPolicy);
	if (tmp_sealedblob != NULL)
		Tspi_Context_FreeMemory(tss->hContext, tmp_sealedblob);
	if (*sealedblob != NULL) {
		free(*sealedblob);
		*sealedblob = NULL;
		*sealedblob_size = 0;
	}
	return result;
}

TPA_RESULT Tpa_TSS_UnsealBlobWithMK(TPA_TSS_ACCESS *tss,
		unsigned char *sealedblob, unsigned int sealedblob_size,
		unsigned char **plainblob, unsigned int *plainblob_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	TSS_HPOLICY hPolicy = 0;
	TSS_HENCDATA hEncData = 0;
	unsigned char *tmp_plainblob = NULL;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (tss->hStorageMK == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hStorageMK is 0");
		goto error;
	}

	if (sealedblob == NULL) {
		TRACE_MSG_PARAM_ERROR("sealedblob is NULL");
		goto error;
	}

	if (plainblob == NULL) {
		TRACE_MSG_PARAM_ERROR("plainblob is NULL");
		goto error;
	}

	if (plainblob_size == NULL) {
		TRACE_MSG_PARAM_ERROR("sealedblob_size is NULL");
		goto error;
	}
#endif

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_ENCDATA, TSS_ENCDATA_SEAL, &hEncData);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_SetAttribData(hEncData, TSS_TSPATTRIB_ENCDATA_BLOB,
			TSS_TSPATTRIB_ENCDATABLOB_BLOB, sealedblob_size, sealedblob);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_SetAttribData failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_POLICY, TSS_POLICY_USAGE, &hPolicy);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_SetSecret(hPolicy, TSS_SECRET_MODE_PLAIN, strlen(
			TPMSEAL_SECRET), (BYTE *) TPMSEAL_SECRET);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_AssignToObject(hPolicy, hEncData);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_AssignToObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Data_Unseal(hEncData, tss->hStorageMK, plainblob_size,
			&tmp_plainblob);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Data_Unseal failed", tss_result);
		goto error;
	}

	*plainblob = (unsigned char *) malloc(*plainblob_size);
	if (*plainblob == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	memcpy(*plainblob, tmp_plainblob, *plainblob_size);

	Tspi_Context_CloseObject(tss->hContext, hEncData);
	Tspi_Context_CloseObject(tss->hContext, hPolicy);
	Tspi_Context_FreeMemory(tss->hContext, tmp_plainblob);

	return TPA_SUCCESS;

	error: if (hEncData != 0)
		Tspi_Context_CloseObject(tss->hContext, hEncData);
	if (hPolicy != 0)
		Tspi_Context_CloseObject(tss->hContext, hPolicy);
	if (tmp_plainblob != NULL)
		Tspi_Context_FreeMemory(tss->hContext, tmp_plainblob);
	if (*plainblob != NULL) {
		free(*plainblob);
		*plainblob = NULL;
		*plainblob_size = 0;
	}
	return result;
}

void Tpa_TSS_GetPcrInfo(BYTE *cInfo, BYTE **pcrInfo, UINT32 *pcrInfo_size) {
	off_t offset = 0;
	UINT32 parmSize = 0;
	UINT32 pcrinfosize = 0;

#ifndef __NO_PARAM_CHECK
	if (cInfo == NULL) {
		TRACE_MSG_PARAM_ERROR("cInfo is NULL");
		return;
	}

	if (pcrInfo == NULL) {
		TRACE_MSG_PARAM_ERROR("pcrInfo is NULL");
		return;
	}

	if (pcrInfo_size == NULL) {
		TRACE_MSG_PARAM_ERROR("pcrInfo_size is NULL");
		return;
	}
#endif

	offset += sizeof(TPM_STRUCT_VER);
	offset += sizeof(TPM_KEY_USAGE);
	offset += sizeof(TPM_KEY_FLAGS);
	offset += sizeof(TPM_AUTH_DATA_USAGE);
	offset += sizeof(TPM_ALGORITHM_ID);
	offset += sizeof(TPM_ENC_SCHEME);
	offset += sizeof(TPM_SIG_SCHEME);
	memcpy(&parmSize, cInfo + offset, sizeof(UINT32));
	parmSize = ntohl(parmSize);
	offset += sizeof(UINT32);
	offset += parmSize;
	offset += sizeof(TPM_DIGEST);
	offset += sizeof(TPM_NONCE);
	offset += sizeof(TSS_BOOL);

	memcpy(&pcrinfosize, cInfo + offset, sizeof(UINT32));
	*pcrInfo_size = ntohl(pcrinfosize);

	offset += sizeof(UINT32);

	*pcrInfo = malloc(*pcrInfo_size);
	if (*pcrInfo == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		return;
	}
	memcpy(*pcrInfo, cInfo + offset, *pcrInfo_size);
}

void Tpa_TSS_GetPcrBitmap(BYTE *pcrInfo, BYTE **bitmap, UINT32 *bitmap_size) {
	UINT16 sizeOfSelect;
	*bitmap_size = 0;

#ifndef __NO_PARAM_CHECK
	if (pcrInfo == NULL) {
		TRACE_MSG_PARAM_ERROR("pcrInfo is NULL");
		return;
	}

	if (bitmap == NULL) {
		TRACE_MSG_PARAM_ERROR("bitmap is NULL");
		return;
	}

	if (bitmap_size == NULL) {
		TRACE_MSG_PARAM_ERROR("bitmap_size is NULL");
		return;
	}
#endif

	memcpy(&sizeOfSelect, pcrInfo, sizeof(UINT16));
	sizeOfSelect = ntohs(sizeOfSelect);
	*bitmap_size = sizeOfSelect;
	*bitmap = malloc(*bitmap_size);
	if (*bitmap == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		return;
	}
	memcpy(*bitmap, pcrInfo + sizeof(UINT16), *bitmap_size);
}

void Tpa_TSS_GetDigestAtRelease(BYTE *pcrInfo, BYTE **digest,
		UINT32 *digest_size) {
	off_t offset = 0;
	UINT16 sizeOfSelect;

#ifndef __NO_PARAM_CHECK
	if (pcrInfo == NULL) {
		TRACE_MSG_PARAM_ERROR("pcrInfo is NULL");
		return;
	}

	if (digest == NULL) {
		TRACE_MSG_PARAM_ERROR("bitmap is NULL");
		return;
	}

	if (digest_size == NULL) {
		TRACE_MSG_PARAM_ERROR("bitmap_size is NULL");
		return;
	}
#endif

	memcpy(&sizeOfSelect, pcrInfo, sizeof(UINT16));
	offset += (sizeof(UINT16) + ntohs(sizeOfSelect));

	*digest_size = sizeof(TPM_COMPOSITE_HASH);
	*digest = malloc(*digest_size);
	if (*digest == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		return;
	}
	memcpy(*digest, pcrInfo + offset, *digest_size);
}

TPA_RESULT Tpa_TSS_GetTPMPublicEK(TPA_TSS_ACCESS *tss, unsigned char **modulus,
		unsigned int *modulus_size, unsigned char **exponent,
		unsigned int *exponent_size) {

	TSS_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (modulus == NULL) {
		TRACE_MSG_PARAM_ERROR("modulus is NULL");
		goto end;
	}

	if (modulus_size == NULL) {
		TRACE_MSG_PARAM_ERROR("modulus_size is NULL");
		goto end;
	}

	if (exponent == NULL) {
		TRACE_MSG_PARAM_ERROR("exponent is NULL");
		goto end;
	}

	if (exponent_size == NULL) {
		TRACE_MSG_PARAM_ERROR("exponent_size is NULL");
		goto end;
	}
#endif

	if ((result = Tpa_TSS_GetPubEndKey(tss)) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetPubEndKey failed");
		goto end;
	}

	if ((result = Tpa_TSS_GetKeyExponent(tss, tss->hPubEndKey, exponent,
			exponent_size)) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetKeyExponent failed");
		goto end;
	}
	if ((result = Tpa_TSS_GetKeyModulus(tss, tss->hPubEndKey, modulus,
			modulus_size)) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_GetKeyModule failed");
		goto end;
	}

	result = TPA_SUCCESS;
	end: /* end */
	return result;

}

TPA_RESULT Tpa_TSS_ReadPCR(TPA_TSS_ACCESS *tss, int pcr_index,
		unsigned char **buffer, unsigned int *buffer_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	BYTE *data = NULL;
	UINT32 data_size = 0;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hTPM is 0");
		goto error;
	}

	if (pcr_index < 0 || pcr_index > 23) {
		TRACE_MSG_PARAM_ERROR("pcr_index is not in 0..23");
		goto error;
	}

	if (buffer == NULL) {
		TRACE_MSG_PARAM_ERROR("buffer is NULL");
		goto error;
	}

	if (buffer_size == NULL) {
		TRACE_MSG_PARAM_ERROR("buffer_size is NULL");
		goto error;
	}

#endif

	tss_result = Tspi_TPM_PcrRead(tss->hTPM, pcr_index, &data_size, &data);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_TPM_PcrRead failed", tss_result);
		goto error;
	}

	*buffer = (unsigned char *) malloc(data_size);
	if (*buffer == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	memcpy(*buffer, data, data_size);
	*buffer_size = data_size;

	Tspi_Context_FreeMemory(tss->hContext, data);

	return TPA_SUCCESS;

	error: if (data != NULL)
		Tspi_Context_FreeMemory(tss->hContext, data);
	if (*buffer != NULL) {
		free(*buffer);
		*buffer = NULL;
		*buffer_size = 0;
	}
	return result;
}

TPA_RESULT Tpa_TSS_ExtendPCR(TPA_TSS_ACCESS *tss, int pcr_index,
		unsigned char *buffer, unsigned int buffer_size,
		unsigned char **extended, unsigned int *extended_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	unsigned char *data = NULL;
	unsigned int data_size = 0;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hTPM is 0");
		goto error;
	}

	if (pcr_index < 0 || pcr_index > 23) {
		TRACE_MSG_PARAM_ERROR("pcr_index is not in 0..23");
		goto error;
	}

	if (buffer == NULL) {
		TRACE_MSG_PARAM_ERROR("digest is NULL");
		goto error;
	}

	if (buffer_size == 0) {
		TRACE_MSG_PARAM_ERROR("digest is 0");
		goto error;
	}
#endif

	tss_result = Tspi_TPM_PcrExtend(tss->hTPM, pcr_index, buffer_size, buffer,
			NULL, &data_size, &data);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_TPM_PcrRead failed", tss_result);
		goto error;
	}

	if (*extended_size != data_size) {
		*extended = (unsigned char *) realloc(*extended, data_size);
	}
	if (*extended == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	memcpy(*extended, data, data_size);
	*extended_size = data_size;

	Tspi_Context_FreeMemory(tss->hContext, data);

	return TPA_SUCCESS;

	error: /* free memory */
	if (data != NULL)
		Tspi_Context_FreeMemory(tss->hContext, data);
	if (*extended != NULL) {
		free(*extended);
		*extended = NULL;
		*extended_size = 0;
	}
	return result;
}

TPA_RESULT Tpa_TSS_ReadEKCert(TPA_TSS_ACCESS *tss, unsigned char **buffer,
		unsigned int *buffer_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	TSS_HNVSTORE hNV;
	UINT32 nvIndex = TSS_NV_DEFINED | TPM_NV_INDEX_EKCert;
	UINT32 blobLen = 0;
	UINT32 offset = 0;
	UINT32 ekOffset = 0;
	UINT32 tag = 0;
	UINT32 certType = 0;
	BYTE *blob = NULL;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hTPM is 0");
		goto error;
	}

	if (buffer == NULL) {
		TRACE_MSG_PARAM_ERROR("buffer is NULL");
		goto error;
	}

	if (buffer_size == NULL) {
		TRACE_MSG_PARAM_ERROR("buffer_size is NULL");
		goto error;
	}

#endif

	tss_result = Tspi_Context_CreateObject(tss->hContext, TSS_OBJECT_TYPE_NV,
			0, &hNV);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_SetAttribUint32(hNV, TSS_TSPATTRIB_NV_INDEX, 0, nvIndex);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_SetAttribUint32 failed", tss_result);
		goto error;
	}

	/* Try reading certificate header from NV memory */
	blobLen = 5;
	tss_result = Tspi_NV_ReadValue(hNV, 0, &blobLen, &blob);
	if (tss_result != TSS_SUCCESS) {
		/* Try again with authorization */
		TSS_HPOLICY hNVPolicy;

		tss_result = Tspi_Context_CreateObject(tss->hContext,
				TSS_OBJECT_TYPE_POLICY, TSS_POLICY_USAGE, &hNVPolicy);
		if (tss_result != TSS_SUCCESS) {
			TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
			goto error;
		}

		tss_result = Tspi_Policy_SetSecret(hNVPolicy, TSS_SECRET_MODE_POPUP, 0,
				NULL);
		if (tss_result != TSS_SUCCESS) {
			TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
			goto error;
		}

		tss_result = Tspi_Policy_AssignToObject(hNVPolicy, hNV);
		if (tss_result != TSS_SUCCESS) {
			TRACE_MSG_TSS_ERROR("Tspi_Policy_AssignToObject failed", tss_result);
			goto error;
		}

		blobLen = 5;
		tss_result = Tspi_NV_ReadValue(hNV, 0, &blobLen, &blob);
		if (tss_result != TSS_SUCCESS) {
			TRACE_MSG_TSS_ERROR("Tspi_NV_ReadValue failed", tss_result);
			goto error;
		}
	}
	if (tss_result != TSS_SUCCESS) {
		fprintf(stderr, "Unable to read EK Certificate from TPM\n");
		goto error;
	}

	if (blobLen < 5)
		goto parseerr;

	tag = (blob[0] << 8) | blob[1];
	if (tag != TCG_TAG_PCCLIENT_STORED_CERT)
		goto parseerr;

	certType = blob[2];
	if (certType != TCG_FULL_CERT)
		goto parseerr;

	*buffer_size = (blob[3] << 8) | blob[4];
	/*	result = Tspi_Context_FreeMemory (hContext, blob); CKERR; */
	offset = 5;
	blobLen = 2;
	tss_result = Tspi_NV_ReadValue(hNV, offset, &blobLen, &blob);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_NV_ReadValue failed", tss_result);
		goto error;
	}

	if (blobLen < 2)
		goto parseerr;

	tag = (blob[0] << 8) | blob[1];
	if (tag == TCG_TAG_PCCLIENT_FULL_CERT) {
		offset += 2;
		*buffer_size -= 2;
	} else if (blob[0] != 0x30) /* Marker of cert structure */
		goto parseerr;

	/* Read cert from chip in pieces - too large requests may fail */
	*buffer = malloc(*buffer_size);
	if (*buffer == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}

	ekOffset = 0;
	while (ekOffset < *buffer_size) {
		blobLen = *buffer_size - ekOffset;
		if (blobLen > BSIZE)
			blobLen = BSIZE;
		tss_result = Tspi_NV_ReadValue(hNV, offset, &blobLen, &blob);
		if (tss_result != TSS_SUCCESS) {
			TRACE_MSG_TSS_ERROR("Tspi_NV_ReadValue failed", tss_result);
			goto error;
		}
		memcpy(*buffer + ekOffset, blob, blobLen);
		/*		result = Tspi_Context_FreeMemory (hContext, blob); CKERR; */
		offset += blobLen;
		ekOffset += blobLen;
	}

	result = TPA_SUCCESS;
	goto end;
	error: TRACE_MSG_ERROR("Failure, error code");
	parseerr: TRACE_MSG_ERROR(
			"Failure, unable to parse certificate store structure");
	end: return result;

}

/*
 * New functions
 * */
static TPA_RESULT Tpa_TSS_KeyAllocate(TPA_TSS_KEY **key) {

#ifndef __NO_PARAM_CHECK
	if (key == NULL) {
		TRACE_MSG_PARAM_ERROR("key is NULL");
		return TPA_ERROR;
	}
#endif

	*key = (TPA_TSS_KEY *) malloc(sizeof(TPA_TSS_KEY));
	if (*key == NULL) {
		return TPA_ERROR;
	}

	(*key)->hKey = 0;
	(*key)->hKeyParent = 0;
	(*key)->keyFlags = 0;
	(*key)->blob = NULL;
	(*key)->secret = NULL;

	return TPA_SUCCESS;

}

void Tpa_TSS_KeyFreeMemory(TPA_TSS_ACCESS *tss, TPA_TSS_KEY *key) {

	if (!key) {
		TRACE_MSG_PARAM_ERROR("key is NULL");
		return;
	}
	if (!tss) {
		TRACE_MSG_PARAM_ERROR("key is NULL");
		return;
	}

	if (key->hKey != 0) {
		Tspi_Context_CloseObject(tss->hContext, key->hKey);
	}

	//	if (key->blob)
	//		free(key->blob);
	//
	//	if (key->secret)
	//		free(key->secret);

	key->hKey = 0;
	key->hKeyParent = 0;
	key->keyFlags = 0;
	key->blob = NULL;
	key->blob_size = 0;
	key->secret = NULL;
	key->secret_size = 0;

	free(key);
}

static TPA_RESULT Tpa_TSS_CreateKey(TPA_TSS_ACCESS *tss, TPA_TSS_KEY *key) {

	TPA_RESULT result = TPA_ERROR;
	TSS_HPOLICY hPolicy = 0;
	TSS_RESULT tss_result;
	TSS_FLAG keyFlags = 0;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (key == NULL) {
		TRACE_MSG_PARAM_ERROR("key is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (key->hKeyParent == 0) {
		TRACE_MSG_PARAM_ERROR("key->hKeyParent is 0");
		goto error;
	}

	if (key->secret == NULL || key->secret_size == 0) {
		TRACE_MSG_PARAM_ERROR("key->secret is 0");
		goto error;
	}
#endif

	keyFlags = key->keyFlags;

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_RSAKEY, keyFlags, &(key->hKey));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_POLICY, TSS_POLICY_USAGE, &hPolicy);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_SetSecret(hPolicy, TSS_SECRET_MODE_PLAIN,
			key->secret_size, (BYTE *) key->secret);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_AssignToObject(hPolicy, key->hKey);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_AssignToObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Key_CreateKey(key->hKey, key->hKeyParent, 0);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Key_CreateKey failed", tss_result);
		goto error;
	}

	/* Load the newly created RSA key */
	tss_result = Tspi_Key_LoadKey(key->hKey, key->hKeyParent);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Key_LoadKey for hStorageMK failed", tss_result);
		goto error;
	}

	return TPA_SUCCESS;

	error: if (hPolicy != 0)
		Tspi_Context_CloseObject(tss->hContext, hPolicy);
	if (tss->hStorageMK != 0) {
		Tspi_Context_CloseObject(tss->hContext, key->hKey);
		tss->hStorageMK = 0;
	}

	return result;
}

static TPA_RESULT Tpa_TSS_LoadKey(TPA_TSS_ACCESS *tss, TPA_TSS_KEY *key) {

	TPA_RESULT result = TPA_ERROR;
	TSS_HPOLICY hPolicy = 0;
	TSS_RESULT tss_result;
	TSS_FLAG keyFlags = 0;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (key == NULL) {
		TRACE_MSG_PARAM_ERROR("key is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (key->blob == NULL) {
		TRACE_MSG_PARAM_ERROR("key->blob is NULL");
		goto error;
	}

	if (key->blob_size == 0) {
		TRACE_MSG_PARAM_ERROR("key->blob_size is NULL");
		goto error;
	}

	if (key->hKeyParent == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hSRK is 0");
		goto error;
	}

	if (key->secret == NULL || key->secret_size == 0) {
		TRACE_MSG_PARAM_ERROR("key->secret is 0");
		goto error;
	}
#endif

	keyFlags = key->keyFlags;

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_RSAKEY, keyFlags, &(key->hKey));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Context_LoadKeyByBlob(tss->hContext, key->hKeyParent,
			key->blob_size, key->blob, &(key->hKey));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_LoadKeyByBlob failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_POLICY, TSS_POLICY_USAGE, &hPolicy);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_SetSecret(hPolicy, TSS_SECRET_MODE_PLAIN,
			key->secret_size, (BYTE *) key->secret);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_AssignToObject(hPolicy, key->hKey);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_AssignToObject failed", tss_result);
		goto error;
	}

	return TPA_SUCCESS;

	error: if (hPolicy != 0)
		Tspi_Context_CloseObject(tss->hContext, hPolicy);
	if (key->hKey != 0) {
		Tspi_Context_CloseObject(tss->hContext, key->hKey);
		key->hKey = 0;
	}

	return result;
}

TPA_RESULT Tpa_TSS_CreateBindingKey(TPA_TSS_ACCESS *tss,
		TPA_TSS_KEY *parentKey, TPA_TSS_KEYPWD secret,
		unsigned int secret_size, TPA_TSS_KEY **key) {

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (secret == NULL) {
		TRACE_MSG_PARAM_ERROR ("secret is NULL");
		goto error;
	}

	if (secret_size == 0) {
		TRACE_MSG_PARAM_ERROR("secret_size is 0");
		goto error;
	}

	if (key == NULL) {
		TRACE_MSG_PARAM_ERROR("key is NULL");
		goto error;
	}
#endif

	if (Tpa_TSS_KeyAllocate(key) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_KeyAllocate failed");
		goto error;
	}

	if (parentKey == NULL) {
		if (tss->hStorageMK == 0) {
			TRACE_MSG_ERROR("tss->hStorageMK is 0");
			goto error;
		}
		(*key)->hKeyParent = tss->hStorageMK;
	} else {
		if (parentKey->hKey == 0) {
			TRACE_MSG_ERROR("parentKey->hKeyParent is 0");
			goto error;
		}
		(*key)->hKeyParent = parentKey->hKey;
	}

	(*key)->keyFlags = TPA_KEY_TSSFLAGS_BINDING;
	(*key)->secret = secret;
	(*key)->secret_size = secret_size;

	if (Tpa_TSS_CreateKey(tss, *key) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_CreateKey failed");
		goto error;
	}

	return TPA_SUCCESS;

	error: /* Error label */
	Tpa_TSS_KeyFreeMemory(tss, *key);
	*key = NULL;
	return result;
}

TPA_RESULT Tpa_TSS_LoadBindingKey(TPA_TSS_ACCESS *tss, TPA_TSS_KEYBLOB keyblob,
		unsigned int keyblob_size, TPA_TSS_KEYPWD secret,
		unsigned int secret_size, TPA_TSS_KEY *parentKey, TPA_TSS_KEY **key) {

	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (keyblob == NULL) {
		TRACE_MSG_PARAM_ERROR ("keyblob is NULL");
		goto error;
	}

	if (keyblob_size == 0) {
		TRACE_MSG_PARAM_ERROR("keyblob_size is 0");
		goto error;
	}

	if (secret == NULL) {
		TRACE_MSG_PARAM_ERROR ("secret is NULL");
		goto error;
	}

	if (secret_size == 0) {
		TRACE_MSG_PARAM_ERROR("secret_size is 0");
		goto error;
	}

	if (key == NULL) {
		TRACE_MSG_PARAM_ERROR("key is NULL");
		goto error;
	}
#endif

	if (Tpa_TSS_KeyAllocate(key) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_KeyAllocate failed");
		goto error;
	}

	if (parentKey == NULL) {
		if (tss->hStorageMK == 0) {
			TRACE_MSG_ERROR("tss->hStorageMK is 0");
			goto error;
		}
		(*key)->hKeyParent = tss->hStorageMK;
	} else {
		if (parentKey->hKey == 0) {
			TRACE_MSG_ERROR("parentKey->hKeyParent is 0");
			goto error;
		}
		(*key)->hKeyParent = parentKey->hKey;
	}

	(*key)->blob = keyblob;
	(*key)->blob_size = keyblob_size;
	(*key)->secret = secret;
	(*key)->secret_size = secret_size;
	(*key)->keyFlags = TPA_KEY_TSSFLAGS_BINDING;

	if (Tpa_TSS_LoadKey(tss, *key) != TPA_SUCCESS) {
		TRACE_MSG_ERROR("Tpa_TSS_CreateKey failed");
		goto error;
	}

	return TPA_SUCCESS;

	error: /* Error label */
	Tpa_TSS_KeyFreeMemory(tss, *key);
	*key = NULL;
	return result;
}

TPA_RESULT Tpa_TSS_Bind(TPA_TSS_ACCESS *tss, TPA_TSS_KEY *key,
		unsigned char *plainblob, unsigned int plainblob_size,
		unsigned char **boundblob, unsigned int *boundblob_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result = 0;
	TSS_HPOLICY hPolicy = 0;
	TSS_HENCDATA hEncData = 0;
	unsigned char *tmp_boundblob = NULL;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (key->hKey == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hKey is 0");
		goto error;
	}

	if (plainblob == NULL) {
		TRACE_MSG_PARAM_ERROR("plainblob is NULL");
		goto error;
	}

	if (plainblob_size == 0) {
		TRACE_MSG_PARAM_ERROR("plainblob_size is 0");
		goto error;
	}

	if (boundblob == NULL) {
		TRACE_MSG_PARAM_ERROR("boundblob is NULL");
		goto error;
	}

	if (boundblob_size == 0) {
		TRACE_MSG_PARAM_ERROR("boundblob_size is 0");
		goto error;
	}
#endif

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_ENCDATA, TSS_ENCDATA_BIND, &hEncData);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Data_Bind(hEncData, key->hKey, plainblob_size, plainblob);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Data_Bind failed", tss_result);
		goto error;
	}

	tss_result = Tspi_GetAttribData(hEncData, TSS_TSPATTRIB_ENCDATA_BLOB,
			TSS_TSPATTRIB_ENCDATABLOB_BLOB, boundblob_size, &tmp_boundblob)
			!= TSS_SUCCESS;
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_GetAttribData failed", tss_result);
		goto error;
	}

	*boundblob = (unsigned char *) malloc(*boundblob_size);
	if (*boundblob == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	memcpy(*boundblob, tmp_boundblob, *boundblob_size);

	Tspi_Context_CloseObject(tss->hContext, hEncData);
	Tspi_Context_CloseObject(tss->hContext, hPolicy);
	Tspi_Context_FreeMemory(tss->hContext, tmp_boundblob);

	return TPA_SUCCESS;

	error: if (hEncData != 0)
		Tspi_Context_CloseObject(tss->hContext, hEncData);
	if (hPolicy != 0)
		Tspi_Context_CloseObject(tss->hContext, hPolicy);
	if (tmp_boundblob != NULL)
		Tspi_Context_FreeMemory(tss->hContext, tmp_boundblob);
	if (*boundblob != NULL) {
		free(*boundblob);
		*boundblob = NULL;
		*boundblob_size = 0;
	}
	return result;
}

TPA_RESULT Tpa_TSS_Unbind(TPA_TSS_ACCESS *tss, TPA_TSS_KEY *key,
		unsigned char *boundblob, unsigned int boundblob_size,
		unsigned char **plainblob, unsigned int *plainblob_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	TSS_HPOLICY hPolicy = 0;
	TSS_HENCDATA hEncData = 0;
	unsigned char *tmp_plainblob = NULL;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (key->hKey == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hKey is 0");
		goto error;
	}

	if (boundblob == NULL) {
		TRACE_MSG_PARAM_ERROR("boundblob is NULL");
		goto error;
	}

	if (boundblob_size == 0) {
		TRACE_MSG_PARAM_ERROR("boundblob_size is 0");
		goto error;
	}

	if (plainblob == NULL) {
		TRACE_MSG_PARAM_ERROR("plainblob is NULL");
		goto error;
	}

	if (plainblob_size == 0) {
		TRACE_MSG_PARAM_ERROR("plainblob_size is NULL");
		goto error;
	}
#endif

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_ENCDATA, TSS_ENCDATA_BIND, &hEncData);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_SetAttribData(hEncData, TSS_TSPATTRIB_ENCDATA_BLOB,
			TSS_TSPATTRIB_ENCDATABLOB_BLOB, boundblob_size, boundblob);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_SetAttribData failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Data_Unbind(hEncData, key->hKey, plainblob_size,
			&tmp_plainblob);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Data_Unseal failed", tss_result);
		goto error;
	}

	*plainblob = (unsigned char *) malloc(*plainblob_size);
	if (*plainblob == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	memcpy(*plainblob, tmp_plainblob, *plainblob_size);

	Tspi_Context_CloseObject(tss->hContext, hEncData);
	Tspi_Context_CloseObject(tss->hContext, hPolicy);
	Tspi_Context_FreeMemory(tss->hContext, tmp_plainblob);

	return TPA_SUCCESS;

	error: if (hEncData != 0)
		Tspi_Context_CloseObject(tss->hContext, hEncData);
	if (hPolicy != 0)
		Tspi_Context_CloseObject(tss->hContext, hPolicy);
	if (tmp_plainblob != NULL)
		Tspi_Context_FreeMemory(tss->hContext, tmp_plainblob);
	if (*plainblob != NULL) {
		free(*plainblob);
		*plainblob = NULL;
		*plainblob_size = 0;
	}
	return result;
}

TPA_RESULT Tpa_TSS_SetSRKSecret(TPA_TSS_ACCESS *tss, char *secret,
		unsigned int secret_len) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	TSS_HPOLICY hPolicy = 0;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hTPM is 0");
		goto end;
	}

	if (secret == NULL) {
		TRACE_MSG_PARAM_ERROR("secret is NULL");
		goto end;
	}
#endif

	tss_result = Tspi_GetPolicyObject(tss->hSRK, TSS_POLICY_USAGE, &hPolicy);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_GetPolicyObject failed", tss_result);
		goto end;
	}

	tss_result = Tspi_Policy_SetSecret(hPolicy, TSS_SECRET_MODE_PLAIN,
			secret_len, (BYTE*) secret);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
		goto end;
	}

	result = TPA_SUCCESS;
	end: return result;
}

TPA_RESULT Tpa_TSS_IsOwnershipTaken(TPA_TSS_ACCESS *tss,
		TPA_TPM_OWNERSHIP_RESULT *isTaken) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	//TSS_HPOLICY hPolicy = 0;
	unsigned int subcap = TSS_TPMCAP_PROP_OWNER;
	unsigned char *owner = NULL;
	unsigned int owner_size = 0;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (isTaken == NULL) {
		TRACE_MSG_PARAM_ERROR("isTaken is NULL");
		goto error;
	}

#endif

	tss_result = Tspi_TPM_GetCapability(tss->hTPM, TSS_TPMCAP_PROPERTY,
			sizeof(subcap), (unsigned char *) &subcap, &owner_size, &owner);

	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_TPM_GetCapability failed", tss_result);
		goto error;
	}

	if ((int) *owner == TRUE)
		*isTaken = TPM_IS_OWNED;
	else
		*isTaken = TPM_IS_NOT_OWNED;

	result = TPA_SUCCESS;
	error: if (owner != NULL)
		Tspi_Context_FreeMemory(tss->hContext, owner);
	return result;
}

TPA_RESULT Tpa_TSS_CreateSRKObject(TPA_TSS_ACCESS *tss, char *secret,
		unsigned int secret_len) {
	TPA_RESULT result = TPA_ERROR;
	TSS_HPOLICY hPolicy;
	TSS_RESULT tss_result;
	TSS_FLAG keyFlags = TSS_KEY_TSP_SRK | TSS_KEY_AUTHORIZATION;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

#endif

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_RSAKEY, keyFlags, &(tss->hSRK));
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Context_CreateObject(tss->hContext,
			TSS_OBJECT_TYPE_POLICY, TSS_POLICY_USAGE, &hPolicy);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_SetSecret(hPolicy, TSS_SECRET_MODE_PLAIN,
			secret_len, (BYTE *) secret);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_SetSecret failed", tss_result);
		goto error;
	}

	tss_result = Tspi_Policy_AssignToObject(hPolicy, tss->hSRK);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Policy_AssignToObject failed", tss_result);
		goto error;
	}

	return TPA_SUCCESS;

	error: if (hPolicy != 0)
		Tspi_Context_CloseObject(tss->hContext, hPolicy);
	if (tss->hSRK != 0) {
		Tspi_Context_CloseObject(tss->hContext, tss->hSRK);
		tss->hSRK = 0;
	}

	return result;
}

TPA_RESULT Tpa_TSS_TakeOwnership(TPA_TSS_ACCESS *tss) {
	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;

#ifndef __NO_PARAM_CHECK
	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hTPM is 0");
		goto end;
	}

	if (tss->hSRK == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hSRK is 0");
		goto end;
	}

#endif

	tss_result = Tspi_TPM_TakeOwnership(tss->hTPM, tss->hSRK, (TSS_HKEY) 0);

	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_TPM_TakeOwnership failed", tss_result);
		goto end;
	}

	result = TPA_SUCCESS;
	end: return result;
}

/*********************************************************************/
/* CODE TOBE CHECKED                                                 */
/*********************************************************************/

TPA_RESULT Tpa_TSS_SetPCRIndexes(TPA_TSS_ACCESS *tss, TPA_PCR_SET *selectedPcrs) {
	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result;
	TSS_FLAG initFlag = 0;
	UINT32 localityValue = 0;
	int i = 0;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}

	if (selectedPcrs == NULL) {
		TRACE_MSG_PARAM_ERROR("selectedPcrs is NULL");
		goto error;
	}
#endif

	for (i = 0; i < selectedPcrs->index; i++) {
		if (selectedPcrs->pcrList[i].pcr > 15) {
			initFlag |= TSS_PCRS_STRUCT_INFO_LONG;
			break;
		}
	}

	tss_result = Tspi_Context_CreateObject(tss->hContext, TSS_OBJECT_TYPE_PCRS,
			initFlag, &tss->hPcrs);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_Context_CreateObject PCR Composite failed",
				tss_result);
		goto error;
	}

	for (i = 0; i < selectedPcrs->index; i++) {

		tss_result = Tspi_PcrComposite_SelectPcrIndex(tss->hPcrs,
				selectedPcrs->pcrList[i].pcr);
		if (tss_result != TSS_SUCCESS) {
			TRACE_MSG_TSS_ERROR("Tspi_PcrComposite_SetPcrIndex failed",
					tss_result);
			goto error;
		}
	}

	if (initFlag) {
		localityValue = TPM_LOC_ZERO | TPM_LOC_ONE | TPM_LOC_TWO
				| TPM_LOC_THREE | TPM_LOC_FOUR;
		tss_result
				= Tspi_PcrComposite_SetPcrLocality(tss->hPcrs, localityValue);
		if (tss_result != TSS_SUCCESS) {
			TRACE_MSG_TSS_ERROR("Tspi_PcrComposite_SetPcrLocality failed",
					tss_result);
			goto error;
		}
	}

	return TPA_SUCCESS;

	error: if (tss->hPcrs != 0) {
		Tspi_Context_CloseObject(tss->hContext, tss->hPcrs);
		tss->hPcrs = 0;
	}

	return result;
}

TPA_RESULT Tpa_TSS_SetData(TPA_TSS_ACCESS *tss, void *data,
		unsigned int data_size) {
	TPA_RESULT result = TPA_ERROR;

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}

	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto end;
	}

	if (data == NULL) {
		TRACE_MSG_PARAM_ERROR("data is NULL");
		goto end;
	}
#endif

	if (tss->data != NULL) {
		TRACE_MSG_ERROR("tss->data is already assigned");
		goto end;
	}

	tss->data = data;
	tss->data_size = data_size;

	result = TPA_SUCCESS;
	end: return result;
}

TPA_RESULT Tpa_TSS_SetDataBlob(TPA_TSS_ACCESS *tss, unsigned char *data,
		unsigned int data_size) {
	return Tpa_TSS_SetData(tss, (void *) data, data_size);
}

TPA_RESULT Tpa_TSS_Quote(TPA_TSS_ACCESS *tss, unsigned char ** dataSigned,
		unsigned int * dataSigned_size, unsigned char ** raSignature,
		unsigned int * raSignature_size) {

	TPA_RESULT result = TPA_ERROR;
	TSS_RESULT tss_result = 0;
	TSS_VALIDATION valid = { { 0 } };

#ifndef __NO_PARAM_CHECK
	if (tss == NULL) {
		TRACE_MSG_PARAM_ERROR("tss is NULL");
		return TPA_ERROR;
	}
	if (tss->hContext == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hContext is 0");
		goto error;
	}
	if (tss->hTPM == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hTPM is 0");
		goto error;
	}
	if (tss->hIdentKey == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hIdentKey is 0");
		goto error;
	}
	if (tss->hPcrs == 0) {
		TRACE_MSG_PARAM_ERROR("tss->hPcrs is 0");
		goto error;
	}
	if (tss->data != NULL && tss->data_size != sizeof(TCPA_NONCE)) {
		TRACE_MSG_PARAM_ERROR("tss->data_size MUST be sizeof(TCPA_NONCE)");
		goto error;
	}

	if (dataSigned == NULL) {
		TRACE_MSG_PARAM_ERROR("dataSigned is NULL");
		goto error;
	}
	if (dataSigned_size == 0) {
		TRACE_MSG_PARAM_ERROR("dataSigned_size is 0");
		goto error;
	}
	if (raSignature == NULL) {
		TRACE_MSG_PARAM_ERROR("raSignature is NULL");
		goto error;
	}
	if (raSignature_size == 0) {
		TRACE_MSG_PARAM_ERROR("raSignature_size is 0");
		goto error;
	}
#endif

	/* get nonce from tss->data or generate it */
	if (tss->data != NULL) {
		valid.ulExternalDataLength = (UINT32) tss->data_size;
		valid.rgbExternalData = (BYTE*) tss->data;
	} else {
		tss_result = Tspi_TPM_GetRandom(tss->hTPM, sizeof(TCPA_NONCE),
				&valid.rgbExternalData);
		if (tss_result != TSS_SUCCESS) {
			TRACE_MSG_TSS_ERROR("Tspi_TPM_GetRandom failed", tss_result);
			goto error;
		}
		valid.ulExternalDataLength = sizeof(TCPA_NONCE);

		/* up to the caller to free tss->data */
		tss->data_size = valid.ulExternalDataLength;
		tss->data = valid.rgbExternalData;
		/* don't copy to avoid inconsistency with the case where nonce is passed
		 tss->data = (unsigned char *) malloc(tss->data_size);
		 if (tss->data == NULL) {
		 TRACE_MSG_ERROR("malloc failed");
		 goto error;
		 }
		 memcpy(tss->data, valid.rgbExternalData, tss->data_size);
		 */
	}

	/* Quote */
	tss_result = Tspi_TPM_Quote(tss->hTPM, tss->hIdentKey, tss->hPcrs, &valid);
	if (tss_result != TSS_SUCCESS) {
		TRACE_MSG_TSS_ERROR("Tspi_TPM_Quote failed", tss_result);
		goto error;
	}

	/* copy dataSigned */
	*dataSigned_size = sizeof(TPM_QUOTE_INFO);
	*dataSigned = (unsigned char *) malloc(*dataSigned_size);
	if (*dataSigned == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	memcpy(*dataSigned, valid.rgbData, *dataSigned_size);

	/* copy raSignature */
	*raSignature_size = valid.ulValidationDataLength;
	*raSignature = (unsigned char *) malloc(*raSignature_size);
	if (*raSignature == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		goto error;
	}
	memcpy(*raSignature, valid.rgbValidationData, *raSignature_size);

	Tspi_Context_FreeMemory(tss->hContext, valid.rgbData);
	Tspi_Context_FreeMemory(tss->hContext, valid.rgbValidationData);

	return TPA_SUCCESS;

	error: if (*dataSigned != NULL) {
		free(*dataSigned);
		*dataSigned = NULL;
		*dataSigned_size = 0;
	}
	if (*raSignature != NULL) {
		free(*raSignature);
		*raSignature = NULL;
		*raSignature_size = 0;
	}
	if (valid.rgbData)
		Tspi_Context_FreeMemory(tss->hContext, valid.rgbData);
	if (valid.rgbValidationData)
		Tspi_Context_FreeMemory(tss->hContext, valid.rgbValidationData);
	return result;
}
