/* ==========================================================================
 * This file is part of the Trusted Platform Agent (TPA)
 *
 * Copyright (C) 2007-2010 Politecnico di Torino, Italy
 * TORSEC group -- http://security.polito.it
 *
 * Trusted Platform Agent (TPA) is copyrighted free software by Politecnico
 * di Torino (TORSEC group).  You can redistribute it and/or modify it under
 * either the terms of the LGPL (see COPYING.LESSER file), or the GPL (see
 * COPYING file).
 *
 * DISCLAIMER of WARRANTY
 *
 * The following software "Trusted Platform Agent (TPA)" developed for
 * Open_TC project (http://www.opentc.net), whilst this version is not
 * endorsed by Open_TC, is experimental and is provided "as is", and no
 * guarantee or warranty is given by Politecnico di Torino (TORSEC group),
 * that the software is fit for any particular purpose.
 * The user thereof uses the software at its sole risk and liability.
 * Politecnico di Torino (TORSEC group) shall have no obligation to maintain
 * or support this software.  Politecnico di Torino (TORSEC group) MAKES NO
 * EXPRESS OR IMPLIED WARRANTY OF ANY KIND REGARDING THIS SOFTWARE.
 * Politecnico di Torino SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT,
 * SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES, WHETHER BASED ON
 * CONTRACT, TORT OR ANY OTHER LEGAL THEORY, IN CONNECTION WITH OR
 * ARISING OUT OF THE FURNISHING, PERFORMANCE OR USE OF THIS SOFTWARE.
 * ==========================================================================
 */
#include "TPA_Utils.h"

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <string.h>
#include <openssl/bio.h>
#include <openssl/evp.h>
#include <openssl/buffer.h>

/****************************************************************************/
/* BLOB                                                                     */
/****************************************************************************/

/**
 * Reads a blob from a file.
 *
 * @param filename a string containing the file name
 * @param blob a pointer to a buffer which is allocated and filled with the blob by the function
 * @param blob_size a pointer to an integer which is allocated and filled with the length of the blob by the function
 */
TPA_RESULT Tpa_BLOB_readFromFile(char * filename, unsigned char ** blob,
		unsigned int *blob_size) {

	FILE *in;
	TPA_RESULT res = TPA_ERROR;
	struct stat file_stats;

#ifndef __NO_PARAM_CHECK
	/*
	 * Parameter check
	 */
	if (strlen(filename) == 0) {
		return TPA_ERROR;
	}
	if (blob_size == NULL) {
		return TPA_ERROR;
	}
	if (blob == NULL) {
		return TPA_ERROR;
	}
#endif

	/*
	 * Actual code
	 */
	if (stat(filename, &file_stats) != 0) {
		return TPA_ERROR;
	}

	*blob
			= (unsigned char *) malloc(sizeof(unsigned char)
					* file_stats.st_size);
	if (blob == NULL) {
		return TPA_ERROR;
	}

	in = fopen(filename, "rb");
	if (in == NULL) {
		return TPA_ERROR;
	}

	*blob_size = fread(*blob, 1, file_stats.st_size, in);
	if (*blob_size == file_stats.st_size) {
		res = TPA_SUCCESS;
	}

	fclose(in);

	return res;
}

/**
 * Writes a blob to a file.
 *
 * @param filename a string containing the file name
 * @param blob a buffer that contains the blob
 * @param blob_size an integer that contains the length of the blob
 */
TPA_RESULT Tpa_BLOB_writeToFile(char *filename, unsigned char *blob,
		unsigned int blob_size) {

	FILE *fp = NULL;
	TPA_RESULT res = TPA_ERROR;
	int w_len = 0;

	/*
	 * Parameter control
	 */
#ifndef __NO_PARAM_CHECK
	if (filename == NULL || blob == NULL || blob_size <= 0) {
		return res;
	}
#endif
	fp = fopen(filename, "wb");
	if (fp == NULL) {
		return res;
	}

	w_len = fwrite(blob, 1, blob_size, fp);
	if (w_len == blob_size) {
		res = TPA_SUCCESS;
	}

	fclose(fp);

	return res;
}

/**
 * Computes the SHA-1 digest of a blob (relying on OpenSSL).
 *
 * @param blob a buffer that contains the blob
 * @param blob_len an integer that contains the length of the blob
 * @param digest a pointer to a buffer which is allocated and filled with the digest by the function
 * @param digest_len a pointer to an integer which is allocated and filled with the length of the digest by the function
 */
TPA_RESULT Tpa_BLOB_Digest(unsigned char *blob, unsigned int blob_len,
		unsigned char **digest, unsigned int *digest_len) {
	TPA_RESULT result = TPA_ERROR;
	EVP_MD_CTX *mdctx = NULL;

	*digest = (unsigned char *) calloc(EVP_MAX_MD_SIZE, sizeof(unsigned char));
	if (*digest == NULL) {
		TRACE_MSG_ERROR("calloc failed.");
		goto error;
	}

	mdctx = EVP_MD_CTX_create();
	EVP_DigestInit(mdctx, EVP_sha1());
	EVP_DigestUpdate(mdctx, blob, blob_len);
	EVP_DigestFinal(mdctx, *digest, digest_len);
	EVP_MD_CTX_destroy(mdctx);

	result = TPA_SUCCESS;

	error: /* Free values */

	return result;
}

TPA_RESULT Tpa_BLOB_copy(unsigned char *blobIn, unsigned int blobIn_size,
		unsigned char **blobOut, unsigned int *blobOut_size) {
#ifndef __NO_PARAM_CHECK
	if (blobIn == NULL) {
		TRACE_MSG_PARAM_ERROR("blobIn is NULL");
		return TPA_ERROR;
	}
	if (blobIn_size == 0) {
		TRACE_MSG_PARAM_ERROR("blobIn_size is 0");
		return TPA_ERROR;
	}
	if (blobOut == NULL) {
		TRACE_MSG_PARAM_ERROR("blobOut is NULL");
		return TPA_ERROR;
	}
	if (blobOut_size == NULL) {
		TRACE_MSG_PARAM_ERROR("blobOut_size is NULL");
		return TPA_ERROR;
	}
#endif
	*blobOut_size = 0;
	*blobOut = NULL;

	*blobOut = malloc(blobIn_size);
	if (*blobOut == NULL) {
		TRACE_MSG_ERROR("malloc failed");
		return TPA_ERROR;
	}

	*blobOut_size = blobIn_size;
	memcpy(*blobOut, blobIn, blobIn_size);

	return TPA_SUCCESS;
}

/*
 * Older and deprecated
 */
TPA_RESULT Tpa_setMember(void ** member, long len, void * data) {
#ifndef __NO_PARAM_CHECK
	if (member == NULL) {
		return TPA_ERROR;
	}
	if (len <= 0) {
		return TPA_ERROR;
	}
	if (data == NULL) {
		return TPA_ERROR;
	}
#endif

	*member = (void *) malloc(sizeof(void) * len);
	if (member == NULL) {
		return TPA_ERROR;
	}

	memmove(*member, data, len);
	if (memcmp(*member, data, len) != 0) {
		return TPA_ERROR;
	}

	return TPA_SUCCESS;
}

TPA_RESULT Tpa_setString(void ** member, long len, void * data) {
#ifndef __NO_PARAM_CHECK
	if (member == NULL) {
		return TPA_ERROR;
	}
	if (len == 0) {
		return TPA_ERROR;
	}
	if (data == NULL) {
		return TPA_ERROR;
	}
#endif

	*member = (void *) calloc(len + 1, sizeof(void));
	if (member == NULL) {
		return TPA_ERROR;
	}

	strncpy(*member, data, len);
	if (strcmp(*member, data) != 0) {
		return TPA_ERROR;
	}

	return TPA_SUCCESS;
}

#if 0
char* Tpa_BASE64_BinToBase64(char* input_buffer, int input_length, int* output_length) {
	BIO *bmem, *b64;
	BUF_MEM *bptr;

	b64 = BIO_new(BIO_f_base64());
	bmem = BIO_new(BIO_s_mem());
	BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
	b64 = BIO_push(b64, bmem);
	BIO_write(b64, input_buffer, input_length);
	BIO_flush(b64);
	BIO_get_mem_ptr(b64, &bptr);

	char *buffer = (char *) malloc(bptr->length);
	if (buffer == NULL) {
		BIO_free_all(b64);
		*output_length = 0;
		return NULL;
	}
	memcpy(buffer, bptr->data, bptr->length);
	*output_length = bptr->length;
	BIO_free_all(b64);

	return buffer;
}

char* Tpa_BASE64_Base64ToBin(char* input_buffer, int input_length, int* output_length) {
	BIO *bmem, *b64;

	char *buffer = calloc(input_length, sizeof(char));
	if (buffer == NULL) {
		*output_length = 0;
		return NULL;
	}
	memset(buffer, 0, input_length);
	b64 = BIO_new(BIO_f_base64());
	bmem = BIO_new_mem_buf(input_buffer, input_length);
	bmem = BIO_push(b64, bmem);
	BIO_set_flags(b64, BIO_FLAGS_BASE64_NO_NL);
	*output_length = BIO_read(bmem, buffer, input_length);
	BIO_free_all(bmem);
	return buffer;
}
#endif
